<head>
  <link rel="preload" href="/static/fonts/ubuntu-v15-latin-regular.woff2" as="font" type="font/woff2" crossorigin>
  <link rel="preload" href="/static/fonts/ubuntu-v15-subset-500.woff2" as="font" type="font/woff2" crossorigin>
  <link rel="preload" href="/static/files/feather.png" as="image" type="image/svg+xml">
  <meta charset="utf-8">
  <meta content="IE=edge" http-equiv="X-UA-Compatible">
  <meta content="width=device-width, initial-scale=1" name="viewport">
  {%- if page.robots -%}
  <meta name="robots" content="{{ page.robots }}">
  {%- endif -%}
  {%- seo -%}
  <link rel="icon" type="image/svg+xml" href="/static/files/feather.png">
  <!-- Stylesheets -->
  <style>{%- include modern-normalize.css -%}</style>
  <style>{%- include inline-style.css -%}</style>
  {% if include.with_aside %}
  <style>{%- include inline-aside-style.css -%}</style>
  {% endif %}
  <link rel="stylesheet" href="/static/css/additional-styles.css">
</head>
