@font-face {
  /* ubuntu-mono-regular - latin */
  font-family: 'Ubuntu Mono';
  src: local(''), url('/static/fonts/ubuntu-mono-v10-latin-regular.woff2') format('woff2'), url('/static/fonts/ubuntu-mono-v10-latin-regular.woff') format('woff'), url('/static/fonts/ubuntu-mono-v10-latin-regular.ttf') format('truetype');
  font-style: normal;
  font-weight: 400;
  font-display: swap;
}

@font-face {
  /* ubuntu-500 - latin */
  font-family: 'Ubuntu';
  src: local(''), url('/static/fonts/ubuntu-v15-latin-500.woff2') format('woff2'), url('/static/fonts/ubuntu-v15-latin-500.woff') format('woff'), url('/static/fonts/ubuntu-v15-latin-500.ttf') format('truetype');
  font-style: normal;
  font-weight: 500;
  font-display: swap;
}

@font-face {
  /* ubuntu-italic - latin */
  font-family: 'Ubuntu';
  src: local(''), url('/static/fonts/ubuntu-v15-latin-italic.woff2') format('woff2'), url('/static/fonts/ubuntu-v15-latin-italic.woff') format('woff'), url('/static/fonts/ubuntu-v15-latin-italic.ttf') format('truetype');
  font-style: italic;
  font-weight: 400;
  font-display: swap;
}

code, pre, kbd {
  font-family: 'Ubuntu Mono', monospace;
  background-color: var(--c-bg-secondary);
  overflow-x: auto;
  border-radius: .3rem;
}

code, kbd {
  padding: .1rem .3rem;
  border: 1px solid var(--c-border);
}

pre code {
  padding: 0;
  border: none;
  background-color: transparent;
}

pre {
  padding: .3rem 1rem;
  border: 1px solid var(--c-border);
}

div.table {
  overflow-x: auto;
  max-width: 100%;
}

div.table p:first-child, div.table p:last-child {
  margin: 0;
}

table {
  border-collapse: collapse;
}

table thead {
  border-block-end: 3px solid var(--c-border);
}

table tbody tr:nth-child(even) {
  background-color: var(--c-bg-secondary);
}

table tbody tr:nth-child(even) kbd, table tbody tr:nth-child(even) code {
  background-color: var(--c-bg);
}

table th {
  font-weight: 500;
  background-color: var(--c-bg-secondary);
}

table th, table td {
  padding: .3rem;
  vertical-align: top;
  text-align: start;
  border: 1px solid var(--c-border);
}

table td p:last-child {
  margin-block-end: 0;
}

table td p:first-child {
  margin-block-start: 0;
}

blockquote {
  border-inline-start: .25rem solid var(--c-border);
  margin: 0;
  padding-inline-start: var(--content-inline-padding);
  font-style: italic;
}

hr {
  border: none;
  border-block-end: 1px solid var(--c-border);
  margin: 2rem 10%;
}

hr+* {
  margin-block-start: 0;
}

main img {
  display: block;
  margin: 1.5rem auto;
  max-width: var(--content-max-width);
  max-height: 40vh;
}

main .sponsors {
  display: grid;
  grid-template-columns: 1fr;
}

main .sponsors a {
  width: 100%;
  min-height: 5rem;
  font-size: 1.25rem;
  text-decoration: none;
  display: flex;
  align-items: center;
  justify-content: center;
  text-align: center;
  padding: .5rem;
}

main .sponsors a:hover {
  background-color: var(--c-bg-secondary);
}

main .sponsors a img {
  height: unset;
  max-height: 4rem;
  max-width: 100%;
  margin: 0;
}

@media screen and (min-width: 800px) {
  main .sponsors {
    grid-template-columns: 1fr 1fr;
  }
  main .sponsors a {
    height: 5rem;
  }
}
