<!DOCTYPE html>
<html lang="en">
{%- include head.html -%}

<body>
  {%- include nav.html -%}

  <main id="main">
    {% if page.title %}
    <h1>{{ page.title }}</h1>
    {% endif %}
    {{ content }}
  </main>

  {%- include footer.html -%}
