#ifdef __APPLE__
  #define MONERO_LOCAL_SYMBOL(x) .private_extern x
#else
  #define MONERO_LOCAL_SYMBOL(x) .hidden x
#endif

.data

.globl crypto_sign_ed25519_amd64_51_30k_batch_REDMASK51
.globl crypto_sign_ed25519_amd64_51_30k_batch_121666_213
.globl crypto_sign_ed25519_amd64_51_30k_batch_2P0
.globl crypto_sign_ed25519_amd64_51_30k_batch_2P1234
.globl crypto_sign_ed25519_amd64_51_30k_batch_4P0
.globl crypto_sign_ed25519_amd64_51_30k_batch_4P1234
.globl crypto_sign_ed25519_amd64_51_30k_batch_MU0
.globl crypto_sign_ed25519_amd64_51_30k_batch_MU1
.globl crypto_sign_ed25519_amd64_51_30k_batch_MU2
.globl crypto_sign_ed25519_amd64_51_30k_batch_MU3
.globl crypto_sign_ed25519_amd64_51_30k_batch_MU4
.globl crypto_sign_ed25519_amd64_51_30k_batch_ORDER0
.globl crypto_sign_ed25519_amd64_51_30k_batch_ORDER1
.globl crypto_sign_ed25519_amd64_51_30k_batch_ORDER2
.globl crypto_sign_ed25519_amd64_51_30k_batch_ORDER3
.globl crypto_sign_ed25519_amd64_51_30k_batch_EC2D0
.globl crypto_sign_ed25519_amd64_51_30k_batch_EC2D1
.globl crypto_sign_ed25519_amd64_51_30k_batch_EC2D2
.globl crypto_sign_ed25519_amd64_51_30k_batch_EC2D3
.globl crypto_sign_ed25519_amd64_51_30k_batch_EC2D4
.globl crypto_sign_ed25519_amd64_51_30k_batch__38

MONERO_LOCAL_SYMBOL(crypto_sign_ed25519_amd64_51_30k_batch_REDMASK51)
MONERO_LOCAL_SYMBOL(crypto_sign_ed25519_amd64_51_30k_batch_121666_213)
MONERO_LOCAL_SYMBOL(crypto_sign_ed25519_amd64_51_30k_batch_2P0)
MONERO_LOCAL_SYMBOL(crypto_sign_ed25519_amd64_51_30k_batch_2P1234)
MONERO_LOCAL_SYMBOL(crypto_sign_ed25519_amd64_51_30k_batch_4P0)
MONERO_LOCAL_SYMBOL(crypto_sign_ed25519_amd64_51_30k_batch_4P1234)
MONERO_LOCAL_SYMBOL(crypto_sign_ed25519_amd64_51_30k_batch_MU0)
MONERO_LOCAL_SYMBOL(crypto_sign_ed25519_amd64_51_30k_batch_MU1)
MONERO_LOCAL_SYMBOL(crypto_sign_ed25519_amd64_51_30k_batch_MU2)
MONERO_LOCAL_SYMBOL(crypto_sign_ed25519_amd64_51_30k_batch_MU3)
MONERO_LOCAL_SYMBOL(crypto_sign_ed25519_amd64_51_30k_batch_MU4)
MONERO_LOCAL_SYMBOL(crypto_sign_ed25519_amd64_51_30k_batch_ORDER0)
MONERO_LOCAL_SYMBOL(crypto_sign_ed25519_amd64_51_30k_batch_ORDER1)
MONERO_LOCAL_SYMBOL(crypto_sign_ed25519_amd64_51_30k_batch_ORDER2)
MONERO_LOCAL_SYMBOL(crypto_sign_ed25519_amd64_51_30k_batch_ORDER3)
MONERO_LOCAL_SYMBOL(crypto_sign_ed25519_amd64_51_30k_batch_EC2D0)
MONERO_LOCAL_SYMBOL(crypto_sign_ed25519_amd64_51_30k_batch_EC2D1)
MONERO_LOCAL_SYMBOL(crypto_sign_ed25519_amd64_51_30k_batch_EC2D2)
MONERO_LOCAL_SYMBOL(crypto_sign_ed25519_amd64_51_30k_batch_EC2D3)
MONERO_LOCAL_SYMBOL(crypto_sign_ed25519_amd64_51_30k_batch_EC2D4)
MONERO_LOCAL_SYMBOL(crypto_sign_ed25519_amd64_51_30k_batch__38)

.p2align 4

crypto_sign_ed25519_amd64_51_30k_batch_REDMASK51:   .quad 0x0007FFFFFFFFFFFF
crypto_sign_ed25519_amd64_51_30k_batch_121666_213: .quad 996687872
crypto_sign_ed25519_amd64_51_30k_batch_2P0:        .quad 0xFFFFFFFFFFFDA
crypto_sign_ed25519_amd64_51_30k_batch_2P1234:     .quad 0xFFFFFFFFFFFFE
crypto_sign_ed25519_amd64_51_30k_batch_4P0:        .quad 0x1FFFFFFFFFFFB4
crypto_sign_ed25519_amd64_51_30k_batch_4P1234:     .quad 0x1FFFFFFFFFFFFC
crypto_sign_ed25519_amd64_51_30k_batch_MU0:         .quad 0xED9CE5A30A2C131B
crypto_sign_ed25519_amd64_51_30k_batch_MU1:         .quad 0x2106215D086329A7
crypto_sign_ed25519_amd64_51_30k_batch_MU2:         .quad 0xFFFFFFFFFFFFFFEB
crypto_sign_ed25519_amd64_51_30k_batch_MU3:         .quad 0xFFFFFFFFFFFFFFFF
crypto_sign_ed25519_amd64_51_30k_batch_MU4:         .quad 0x000000000000000F
crypto_sign_ed25519_amd64_51_30k_batch_ORDER0:      .quad 0x5812631A5CF5D3ED
crypto_sign_ed25519_amd64_51_30k_batch_ORDER1:      .quad 0x14DEF9DEA2F79CD6
crypto_sign_ed25519_amd64_51_30k_batch_ORDER2:      .quad 0x0000000000000000
crypto_sign_ed25519_amd64_51_30k_batch_ORDER3:      .quad 0x1000000000000000
crypto_sign_ed25519_amd64_51_30k_batch_EC2D0:       .quad 1859910466990425
crypto_sign_ed25519_amd64_51_30k_batch_EC2D1:       .quad 932731440258426
crypto_sign_ed25519_amd64_51_30k_batch_EC2D2:       .quad 1072319116312658
crypto_sign_ed25519_amd64_51_30k_batch_EC2D3:       .quad 1815898335770999
crypto_sign_ed25519_amd64_51_30k_batch_EC2D4:       .quad 633789495995903
crypto_sign_ed25519_amd64_51_30k_batch__38:         .quad 38
