<footer>
  <div>
  </div>
  <div>
    <a href="https://featherwallet.org" rel="noopener noreferrer">Main site</a>
    /
    <a href="https://github.com/feather-wallet/feather" rel="noopener noreferrer">Git repository</a>
  </div>
</footer>
