<a class="skiplink" href="#main">Skip to content</a>
<header>
  <a class="logo" href="{{ site.url }}">
    <h1>Feather Wallet Documentation</h1>
  </a>

  <nav class="" id="main-nav" aria-labelledby="nav-toggle" role="menubar">
    <a id="nav-close" class="nav-close hidden" aria-controls="main-nav"><span>Close</span></a>
    {% for navgroup in site.data.nav %}
    {% if navgroup.items %}
    <details>
      <summary
        {% if navgroup.highlight %}class="highlight"{% endif %}
      >{{ navgroup.name }}</summary>

      <div>
        {% for link in navgroup.items %}
        <a
        role="menuitem"
        href="{{ link.link }}"
        {% if link.external %} rel="noopener noreferrer" {% endif %}
        ><span>{{ link.name }}</span></a>
        {% endfor %}
      </div>
    </details>
    {% else %}
      <a role="menuitem" href="{{ navgroup.link }}"><span>{{ navgroup.name }}</span></a>
    {% endif %}
    {% endfor %}
  </nav>
</header>
