// SPDX-License-Identifier: BSD-3-Clause
// SPDX-FileCopyrightText: The Monero Project

#include "constants.h"

namespace constants {
    NetworkType::Type networkType = NetworkType::Type::MAINNET;
};
