/* This file was generated by tinfo/MKfallback.sh */

/*
 * DO NOT EDIT THIS FILE BY HAND!
 */

#include <curses.priv.h>

#include <tic.h>

/* fallback entries for: linux rxvt vt100 xterm xterm-256color screen screen.linux screen.rxvt screen.xterm-new screen.xterm-256color */
/* linux */

static char linux_alias_data[] = "linux|linux console";

static char linux_s_bel         [] = "\007";
static char linux_s_cr          [] = "\015";
static char linux_s_csr         [] = "\033[%i%p1%d;%p2%dr";
static char linux_s_tbc         [] = "\033[3g";
static char linux_s_clear       [] = "\033[H\033[J";
static char linux_s_el          [] = "\033[K";
static char linux_s_ed          [] = "\033[J";
static char linux_s_hpa         [] = "\033[%i%p1%dG";
static char linux_s_cup         [] = "\033[%i%p1%d;%p2%dH";
static char linux_s_cud1        [] = "\012";
static char linux_s_home        [] = "\033[H";
static char linux_s_civis       [] = "\033[?25l\033[?1c";
static char linux_s_cub1        [] = "\010";
static char linux_s_cnorm       [] = "\033[?25h\033[?0c";
static char linux_s_cuf1        [] = "\033[C";
static char linux_s_cuu1        [] = "\033[A";
static char linux_s_cvvis       [] = "\033[?25h\033[?8c";
static char linux_s_dch1        [] = "\033[P";
static char linux_s_dl1         [] = "\033[M";
static char linux_s_smacs       [] = "\016";
static char linux_s_blink       [] = "\033[5m";
static char linux_s_bold        [] = "\033[1m";
static char linux_s_dim         [] = "\033[2m";
static char linux_s_smir        [] = "\033[4h";
static char linux_s_rev         [] = "\033[7m";
static char linux_s_smso        [] = "\033[7m";
static char linux_s_smul        [] = "\033[4m";
static char linux_s_ech         [] = "\033[%p1%dX";
static char linux_s_rmacs       [] = "\017";
static char linux_s_sgr0        [] = "\033[m\017";
static char linux_s_rmir        [] = "\033[4l";
static char linux_s_rmso        [] = "\033[27m";
static char linux_s_rmul        [] = "\033[24m";
static char linux_s_flash       [] = "\033[?5h$<200/>\033[?5l";
static char linux_s_ich1        [] = "\033[@";
static char linux_s_il1         [] = "\033[L";
static char linux_s_kbs         [] = "\177";
static char linux_s_kdch1       [] = "\033[3~";
static char linux_s_kcud1       [] = "\033[B";
static char linux_s_kf1         [] = "\033[[A";
static char linux_s_kf10        [] = "\033[21~";
static char linux_s_kf2         [] = "\033[[B";
static char linux_s_kf3         [] = "\033[[C";
static char linux_s_kf4         [] = "\033[[D";
static char linux_s_kf5         [] = "\033[[E";
static char linux_s_kf6         [] = "\033[17~";
static char linux_s_kf7         [] = "\033[18~";
static char linux_s_kf8         [] = "\033[19~";
static char linux_s_kf9         [] = "\033[20~";
static char linux_s_khome       [] = "\033[1~";
static char linux_s_kich1       [] = "\033[2~";
static char linux_s_kcub1       [] = "\033[D";
static char linux_s_knp         [] = "\033[6~";
static char linux_s_kpp         [] = "\033[5~";
static char linux_s_kcuf1       [] = "\033[C";
static char linux_s_kcuu1       [] = "\033[A";
static char linux_s_nel         [] = "\015\012";
static char linux_s_dch         [] = "\033[%p1%dP";
static char linux_s_dl          [] = "\033[%p1%dM";
static char linux_s_cud         [] = "\033[%p1%dB";
static char linux_s_ich         [] = "\033[%p1%d@";
static char linux_s_il          [] = "\033[%p1%dL";
static char linux_s_cub         [] = "\033[%p1%dD";
static char linux_s_cuf         [] = "\033[%p1%dC";
static char linux_s_cuu         [] = "\033[%p1%dA";
static char linux_s_rs1         [] = "\033c\033]R";
static char linux_s_rc          [] = "\0338";
static char linux_s_vpa         [] = "\033[%i%p1%dd";
static char linux_s_sc          [] = "\0337";
static char linux_s_ind         [] = "\012";
static char linux_s_ri          [] = "\033M";
static char linux_s_sgr         [] = "\033[0;10%?%p1%t;7%;%?%p2%t;4%;%?%p3%t;7%;%?%p4%t;5%;%?%p5%t;2%;%?%p6%t;1%;m%?%p9%t\016%e\017%;";
static char linux_s_hts         [] = "\033H";
static char linux_s_ht          [] = "\011";
static char linux_s_kb2         [] = "\033[G";
static char linux_s_acsc        [] = "++,,--..00__``aaffgghhiijjkkllmmnnooppqqrrssttuuvvwwxxyyzz{{||}c~~";
static char linux_s_kcbt        [] = "\033[Z";
static char linux_s_smam        [] = "\033[?7h";
static char linux_s_rmam        [] = "\033[?7l";
static char linux_s_enacs       [] = "\033)0";
static char linux_s_kend        [] = "\033[4~";
static char linux_s_kspd        [] = "\032";
static char linux_s_kf11        [] = "\033[23~";
static char linux_s_kf12        [] = "\033[24~";
static char linux_s_kf13        [] = "\033[25~";
static char linux_s_kf14        [] = "\033[26~";
static char linux_s_kf15        [] = "\033[28~";
static char linux_s_kf16        [] = "\033[29~";
static char linux_s_kf17        [] = "\033[31~";
static char linux_s_kf18        [] = "\033[32~";
static char linux_s_kf19        [] = "\033[33~";
static char linux_s_kf20        [] = "\033[34~";
static char linux_s_el1         [] = "\033[1K";
static char linux_s_u6          [] = "\033[%i%d;%dR";
static char linux_s_u7          [] = "\033[6n";
static char linux_s_u8          [] = "\033[?6c";
static char linux_s_u9          [] = "\033[c";
static char linux_s_op          [] = "\033[39;49m";
static char linux_s_oc          [] = "\033]R";
static char linux_s_initc       [] = "\033]P%p1%x%p2%{255}%*%{1000}%/%02x%p3%{255}%*%{1000}%/%02x%p4%{255}%*%{1000}%/%02x";
static char linux_s_kmous       [] = "\033[M";
static char linux_s_setaf       [] = "\033[3%p1%dm";
static char linux_s_setab       [] = "\033[4%p1%dm";
static char linux_s_smpch       [] = "\033[11m";
static char linux_s_rmpch       [] = "\033[10m";

static char linux_bool_data[] = {
	/*   0: bw       */	FALSE,
	/*   1: am       */	TRUE,
	/*   2: xsb      */	FALSE,
	/*   3: xhp      */	FALSE,
	/*   4: xenl     */	TRUE,
	/*   5: eo       */	TRUE,
	/*   6: gn       */	FALSE,
	/*   7: hc       */	FALSE,
	/*   8: km       */	FALSE,
	/*   9: hs       */	FALSE,
	/*  10: in       */	FALSE,
	/*  11: da       */	FALSE,
	/*  12: db       */	FALSE,
	/*  13: mir      */	TRUE,
	/*  14: msgr     */	TRUE,
	/*  15: os       */	FALSE,
	/*  16: eslok    */	FALSE,
	/*  17: xt       */	FALSE,
	/*  18: hz       */	FALSE,
	/*  19: ul       */	FALSE,
	/*  20: xon      */	TRUE,
	/*  21: nxon     */	FALSE,
	/*  22: mc5i     */	FALSE,
	/*  23: chts     */	FALSE,
	/*  24: nrrmc    */	FALSE,
	/*  25: npc      */	FALSE,
	/*  26: ndscr    */	FALSE,
	/*  27: ccc      */	TRUE,
	/*  28: bce      */	TRUE,
	/*  29: hls      */	FALSE,
	/*  30: xhpa     */	FALSE,
	/*  31: crxm     */	FALSE,
	/*  32: daisy    */	FALSE,
	/*  33: xvpa     */	FALSE,
	/*  34: sam      */	FALSE,
	/*  35: cpix     */	FALSE,
	/*  36: lpix     */	FALSE,
	/*  37: OTbs     */	FALSE,
	/*  38: OTns     */	FALSE,
	/*  39: OTnc     */	FALSE,
	/*  40: OTMT     */	FALSE,
	/*  41: OTNL     */	FALSE,
	/*  42: OTpt     */	FALSE,
	/*  43: OTxr     */	FALSE,
};
static NCURSES_INT2 linux_number_data[] = {
	/*   0: cols     */	ABSENT_NUMERIC,
	/*   1: it       */	8,
	/*   2: lines    */	ABSENT_NUMERIC,
	/*   3: lm       */	ABSENT_NUMERIC,
	/*   4: xmc      */	ABSENT_NUMERIC,
	/*   5: pb       */	ABSENT_NUMERIC,
	/*   6: vt       */	ABSENT_NUMERIC,
	/*   7: wsl      */	ABSENT_NUMERIC,
	/*   8: nlab     */	ABSENT_NUMERIC,
	/*   9: lh       */	ABSENT_NUMERIC,
	/*  10: lw       */	ABSENT_NUMERIC,
	/*  11: ma       */	ABSENT_NUMERIC,
	/*  12: wnum     */	ABSENT_NUMERIC,
	/*  13: colors   */	8,
	/*  14: pairs    */	64,
	/*  15: ncv      */	18,
	/*  16: bufsz    */	ABSENT_NUMERIC,
	/*  17: spinv    */	ABSENT_NUMERIC,
	/*  18: spinh    */	ABSENT_NUMERIC,
	/*  19: maddr    */	ABSENT_NUMERIC,
	/*  20: mjump    */	ABSENT_NUMERIC,
	/*  21: mcs      */	ABSENT_NUMERIC,
	/*  22: mls      */	ABSENT_NUMERIC,
	/*  23: npins    */	ABSENT_NUMERIC,
	/*  24: orc      */	ABSENT_NUMERIC,
	/*  25: orl      */	ABSENT_NUMERIC,
	/*  26: orhi     */	ABSENT_NUMERIC,
	/*  27: orvi     */	ABSENT_NUMERIC,
	/*  28: cps      */	ABSENT_NUMERIC,
	/*  29: widcs    */	ABSENT_NUMERIC,
	/*  30: btns     */	ABSENT_NUMERIC,
	/*  31: bitwin   */	ABSENT_NUMERIC,
	/*  32: bitype   */	ABSENT_NUMERIC,
	/*  33: OTug     */	ABSENT_NUMERIC,
	/*  34: OTdC     */	ABSENT_NUMERIC,
	/*  35: OTdN     */	ABSENT_NUMERIC,
	/*  36: OTdB     */	ABSENT_NUMERIC,
	/*  37: OTdT     */	ABSENT_NUMERIC,
	/*  38: OTkn     */	ABSENT_NUMERIC,
};
static char * linux_string_data[] = {
	/*   0: cbt      */	ABSENT_STRING,
	/*   1: bel      */	linux_s_bel,
	/*   2: cr       */	linux_s_cr,
	/*   3: csr      */	linux_s_csr,
	/*   4: tbc      */	linux_s_tbc,
	/*   5: clear    */	linux_s_clear,
	/*   6: el       */	linux_s_el,
	/*   7: ed       */	linux_s_ed,
	/*   8: hpa      */	linux_s_hpa,
	/*   9: cmdch    */	ABSENT_STRING,
	/*  10: cup      */	linux_s_cup,
	/*  11: cud1     */	linux_s_cud1,
	/*  12: home     */	linux_s_home,
	/*  13: civis    */	linux_s_civis,
	/*  14: cub1     */	linux_s_cub1,
	/*  15: mrcup    */	ABSENT_STRING,
	/*  16: cnorm    */	linux_s_cnorm,
	/*  17: cuf1     */	linux_s_cuf1,
	/*  18: ll       */	ABSENT_STRING,
	/*  19: cuu1     */	linux_s_cuu1,
	/*  20: cvvis    */	linux_s_cvvis,
	/*  21: dch1     */	linux_s_dch1,
	/*  22: dl1      */	linux_s_dl1,
	/*  23: dsl      */	ABSENT_STRING,
	/*  24: hd       */	ABSENT_STRING,
	/*  25: smacs    */	linux_s_smacs,
	/*  26: blink    */	linux_s_blink,
	/*  27: bold     */	linux_s_bold,
	/*  28: smcup    */	ABSENT_STRING,
	/*  29: smdc     */	ABSENT_STRING,
	/*  30: dim      */	linux_s_dim,
	/*  31: smir     */	linux_s_smir,
	/*  32: invis    */	ABSENT_STRING,
	/*  33: prot     */	ABSENT_STRING,
	/*  34: rev      */	linux_s_rev,
	/*  35: smso     */	linux_s_smso,
	/*  36: smul     */	linux_s_smul,
	/*  37: ech      */	linux_s_ech,
	/*  38: rmacs    */	linux_s_rmacs,
	/*  39: sgr0     */	linux_s_sgr0,
	/*  40: rmcup    */	ABSENT_STRING,
	/*  41: rmdc     */	ABSENT_STRING,
	/*  42: rmir     */	linux_s_rmir,
	/*  43: rmso     */	linux_s_rmso,
	/*  44: rmul     */	linux_s_rmul,
	/*  45: flash    */	linux_s_flash,
	/*  46: ff       */	ABSENT_STRING,
	/*  47: fsl      */	ABSENT_STRING,
	/*  48: is1      */	ABSENT_STRING,
	/*  49: is2      */	ABSENT_STRING,
	/*  50: is3      */	ABSENT_STRING,
	/*  51: if       */	ABSENT_STRING,
	/*  52: ich1     */	linux_s_ich1,
	/*  53: il1      */	linux_s_il1,
	/*  54: ip       */	ABSENT_STRING,
	/*  55: kbs      */	linux_s_kbs,
	/*  56: ktbc     */	ABSENT_STRING,
	/*  57: kclr     */	ABSENT_STRING,
	/*  58: kctab    */	ABSENT_STRING,
	/*  59: kdch1    */	linux_s_kdch1,
	/*  60: kdl1     */	ABSENT_STRING,
	/*  61: kcud1    */	linux_s_kcud1,
	/*  62: krmir    */	ABSENT_STRING,
	/*  63: kel      */	ABSENT_STRING,
	/*  64: ked      */	ABSENT_STRING,
	/*  65: kf0      */	ABSENT_STRING,
	/*  66: kf1      */	linux_s_kf1,
	/*  67: kf10     */	linux_s_kf10,
	/*  68: kf2      */	linux_s_kf2,
	/*  69: kf3      */	linux_s_kf3,
	/*  70: kf4      */	linux_s_kf4,
	/*  71: kf5      */	linux_s_kf5,
	/*  72: kf6      */	linux_s_kf6,
	/*  73: kf7      */	linux_s_kf7,
	/*  74: kf8      */	linux_s_kf8,
	/*  75: kf9      */	linux_s_kf9,
	/*  76: khome    */	linux_s_khome,
	/*  77: kich1    */	linux_s_kich1,
	/*  78: kil1     */	ABSENT_STRING,
	/*  79: kcub1    */	linux_s_kcub1,
	/*  80: kll      */	ABSENT_STRING,
	/*  81: knp      */	linux_s_knp,
	/*  82: kpp      */	linux_s_kpp,
	/*  83: kcuf1    */	linux_s_kcuf1,
	/*  84: kind     */	ABSENT_STRING,
	/*  85: kri      */	ABSENT_STRING,
	/*  86: khts     */	ABSENT_STRING,
	/*  87: kcuu1    */	linux_s_kcuu1,
	/*  88: rmkx     */	ABSENT_STRING,
	/*  89: smkx     */	ABSENT_STRING,
	/*  90: lf0      */	ABSENT_STRING,
	/*  91: lf1      */	ABSENT_STRING,
	/*  92: lf10     */	ABSENT_STRING,
	/*  93: lf2      */	ABSENT_STRING,
	/*  94: lf3      */	ABSENT_STRING,
	/*  95: lf4      */	ABSENT_STRING,
	/*  96: lf5      */	ABSENT_STRING,
	/*  97: lf6      */	ABSENT_STRING,
	/*  98: lf7      */	ABSENT_STRING,
	/*  99: lf8      */	ABSENT_STRING,
	/* 100: lf9      */	ABSENT_STRING,
	/* 101: rmm      */	ABSENT_STRING,
	/* 102: smm      */	ABSENT_STRING,
	/* 103: nel      */	linux_s_nel,
	/* 104: pad      */	ABSENT_STRING,
	/* 105: dch      */	linux_s_dch,
	/* 106: dl       */	linux_s_dl,
	/* 107: cud      */	linux_s_cud,
	/* 108: ich      */	linux_s_ich,
	/* 109: indn     */	ABSENT_STRING,
	/* 110: il       */	linux_s_il,
	/* 111: cub      */	linux_s_cub,
	/* 112: cuf      */	linux_s_cuf,
	/* 113: rin      */	ABSENT_STRING,
	/* 114: cuu      */	linux_s_cuu,
	/* 115: pfkey    */	ABSENT_STRING,
	/* 116: pfloc    */	ABSENT_STRING,
	/* 117: pfx      */	ABSENT_STRING,
	/* 118: mc0      */	ABSENT_STRING,
	/* 119: mc4      */	ABSENT_STRING,
	/* 120: mc5      */	ABSENT_STRING,
	/* 121: rep      */	ABSENT_STRING,
	/* 122: rs1      */	linux_s_rs1,
	/* 123: rs2      */	ABSENT_STRING,
	/* 124: rs3      */	ABSENT_STRING,
	/* 125: rf       */	ABSENT_STRING,
	/* 126: rc       */	linux_s_rc,
	/* 127: vpa      */	linux_s_vpa,
	/* 128: sc       */	linux_s_sc,
	/* 129: ind      */	linux_s_ind,
	/* 130: ri       */	linux_s_ri,
	/* 131: sgr      */	linux_s_sgr,
	/* 132: hts      */	linux_s_hts,
	/* 133: wind     */	ABSENT_STRING,
	/* 134: ht       */	linux_s_ht,
	/* 135: tsl      */	ABSENT_STRING,
	/* 136: uc       */	ABSENT_STRING,
	/* 137: hu       */	ABSENT_STRING,
	/* 138: iprog    */	ABSENT_STRING,
	/* 139: ka1      */	ABSENT_STRING,
	/* 140: ka3      */	ABSENT_STRING,
	/* 141: kb2      */	linux_s_kb2,
	/* 142: kc1      */	ABSENT_STRING,
	/* 143: kc3      */	ABSENT_STRING,
	/* 144: mc5p     */	ABSENT_STRING,
	/* 145: rmp      */	ABSENT_STRING,
	/* 146: acsc     */	linux_s_acsc,
	/* 147: pln      */	ABSENT_STRING,
	/* 148: kcbt     */	linux_s_kcbt,
	/* 149: smxon    */	ABSENT_STRING,
	/* 150: rmxon    */	ABSENT_STRING,
	/* 151: smam     */	linux_s_smam,
	/* 152: rmam     */	linux_s_rmam,
	/* 153: xonc     */	ABSENT_STRING,
	/* 154: xoffc    */	ABSENT_STRING,
	/* 155: enacs    */	linux_s_enacs,
	/* 156: smln     */	ABSENT_STRING,
	/* 157: rmln     */	ABSENT_STRING,
	/* 158: kbeg     */	ABSENT_STRING,
	/* 159: kcan     */	ABSENT_STRING,
	/* 160: kclo     */	ABSENT_STRING,
	/* 161: kcmd     */	ABSENT_STRING,
	/* 162: kcpy     */	ABSENT_STRING,
	/* 163: kcrt     */	ABSENT_STRING,
	/* 164: kend     */	linux_s_kend,
	/* 165: kent     */	ABSENT_STRING,
	/* 166: kext     */	ABSENT_STRING,
	/* 167: kfnd     */	ABSENT_STRING,
	/* 168: khlp     */	ABSENT_STRING,
	/* 169: kmrk     */	ABSENT_STRING,
	/* 170: kmsg     */	ABSENT_STRING,
	/* 171: kmov     */	ABSENT_STRING,
	/* 172: knxt     */	ABSENT_STRING,
	/* 173: kopn     */	ABSENT_STRING,
	/* 174: kopt     */	ABSENT_STRING,
	/* 175: kprv     */	ABSENT_STRING,
	/* 176: kprt     */	ABSENT_STRING,
	/* 177: krdo     */	ABSENT_STRING,
	/* 178: kref     */	ABSENT_STRING,
	/* 179: krfr     */	ABSENT_STRING,
	/* 180: krpl     */	ABSENT_STRING,
	/* 181: krst     */	ABSENT_STRING,
	/* 182: kres     */	ABSENT_STRING,
	/* 183: ksav     */	ABSENT_STRING,
	/* 184: kspd     */	linux_s_kspd,
	/* 185: kund     */	ABSENT_STRING,
	/* 186: kBEG     */	ABSENT_STRING,
	/* 187: kCAN     */	ABSENT_STRING,
	/* 188: kCMD     */	ABSENT_STRING,
	/* 189: kCPY     */	ABSENT_STRING,
	/* 190: kCRT     */	ABSENT_STRING,
	/* 191: kDC      */	ABSENT_STRING,
	/* 192: kDL      */	ABSENT_STRING,
	/* 193: kslt     */	ABSENT_STRING,
	/* 194: kEND     */	ABSENT_STRING,
	/* 195: kEOL     */	ABSENT_STRING,
	/* 196: kEXT     */	ABSENT_STRING,
	/* 197: kFND     */	ABSENT_STRING,
	/* 198: kHLP     */	ABSENT_STRING,
	/* 199: kHOM     */	ABSENT_STRING,
	/* 200: kIC      */	ABSENT_STRING,
	/* 201: kLFT     */	ABSENT_STRING,
	/* 202: kMSG     */	ABSENT_STRING,
	/* 203: kMOV     */	ABSENT_STRING,
	/* 204: kNXT     */	ABSENT_STRING,
	/* 205: kOPT     */	ABSENT_STRING,
	/* 206: kPRV     */	ABSENT_STRING,
	/* 207: kPRT     */	ABSENT_STRING,
	/* 208: kRDO     */	ABSENT_STRING,
	/* 209: kRPL     */	ABSENT_STRING,
	/* 210: kRIT     */	ABSENT_STRING,
	/* 211: kRES     */	ABSENT_STRING,
	/* 212: kSAV     */	ABSENT_STRING,
	/* 213: kSPD     */	ABSENT_STRING,
	/* 214: kUND     */	ABSENT_STRING,
	/* 215: rfi      */	ABSENT_STRING,
	/* 216: kf11     */	linux_s_kf11,
	/* 217: kf12     */	linux_s_kf12,
	/* 218: kf13     */	linux_s_kf13,
	/* 219: kf14     */	linux_s_kf14,
	/* 220: kf15     */	linux_s_kf15,
	/* 221: kf16     */	linux_s_kf16,
	/* 222: kf17     */	linux_s_kf17,
	/* 223: kf18     */	linux_s_kf18,
	/* 224: kf19     */	linux_s_kf19,
	/* 225: kf20     */	linux_s_kf20,
	/* 226: kf21     */	ABSENT_STRING,
	/* 227: kf22     */	ABSENT_STRING,
	/* 228: kf23     */	ABSENT_STRING,
	/* 229: kf24     */	ABSENT_STRING,
	/* 230: kf25     */	ABSENT_STRING,
	/* 231: kf26     */	ABSENT_STRING,
	/* 232: kf27     */	ABSENT_STRING,
	/* 233: kf28     */	ABSENT_STRING,
	/* 234: kf29     */	ABSENT_STRING,
	/* 235: kf30     */	ABSENT_STRING,
	/* 236: kf31     */	ABSENT_STRING,
	/* 237: kf32     */	ABSENT_STRING,
	/* 238: kf33     */	ABSENT_STRING,
	/* 239: kf34     */	ABSENT_STRING,
	/* 240: kf35     */	ABSENT_STRING,
	/* 241: kf36     */	ABSENT_STRING,
	/* 242: kf37     */	ABSENT_STRING,
	/* 243: kf38     */	ABSENT_STRING,
	/* 244: kf39     */	ABSENT_STRING,
	/* 245: kf40     */	ABSENT_STRING,
	/* 246: kf41     */	ABSENT_STRING,
	/* 247: kf42     */	ABSENT_STRING,
	/* 248: kf43     */	ABSENT_STRING,
	/* 249: kf44     */	ABSENT_STRING,
	/* 250: kf45     */	ABSENT_STRING,
	/* 251: kf46     */	ABSENT_STRING,
	/* 252: kf47     */	ABSENT_STRING,
	/* 253: kf48     */	ABSENT_STRING,
	/* 254: kf49     */	ABSENT_STRING,
	/* 255: kf50     */	ABSENT_STRING,
	/* 256: kf51     */	ABSENT_STRING,
	/* 257: kf52     */	ABSENT_STRING,
	/* 258: kf53     */	ABSENT_STRING,
	/* 259: kf54     */	ABSENT_STRING,
	/* 260: kf55     */	ABSENT_STRING,
	/* 261: kf56     */	ABSENT_STRING,
	/* 262: kf57     */	ABSENT_STRING,
	/* 263: kf58     */	ABSENT_STRING,
	/* 264: kf59     */	ABSENT_STRING,
	/* 265: kf60     */	ABSENT_STRING,
	/* 266: kf61     */	ABSENT_STRING,
	/* 267: kf62     */	ABSENT_STRING,
	/* 268: kf63     */	ABSENT_STRING,
	/* 269: el1      */	linux_s_el1,
	/* 270: mgc      */	ABSENT_STRING,
	/* 271: smgl     */	ABSENT_STRING,
	/* 272: smgr     */	ABSENT_STRING,
	/* 273: fln      */	ABSENT_STRING,
	/* 274: sclk     */	ABSENT_STRING,
	/* 275: dclk     */	ABSENT_STRING,
	/* 276: rmclk    */	ABSENT_STRING,
	/* 277: cwin     */	ABSENT_STRING,
	/* 278: wingo    */	ABSENT_STRING,
	/* 279: hup      */	ABSENT_STRING,
	/* 280: dial     */	ABSENT_STRING,
	/* 281: qdial    */	ABSENT_STRING,
	/* 282: tone     */	ABSENT_STRING,
	/* 283: pulse    */	ABSENT_STRING,
	/* 284: hook     */	ABSENT_STRING,
	/* 285: pause    */	ABSENT_STRING,
	/* 286: wait     */	ABSENT_STRING,
	/* 287: u0       */	ABSENT_STRING,
	/* 288: u1       */	ABSENT_STRING,
	/* 289: u2       */	ABSENT_STRING,
	/* 290: u3       */	ABSENT_STRING,
	/* 291: u4       */	ABSENT_STRING,
	/* 292: u5       */	ABSENT_STRING,
	/* 293: u6       */	linux_s_u6,
	/* 294: u7       */	linux_s_u7,
	/* 295: u8       */	linux_s_u8,
	/* 296: u9       */	linux_s_u9,
	/* 297: op       */	linux_s_op,
	/* 298: oc       */	linux_s_oc,
	/* 299: initc    */	linux_s_initc,
	/* 300: initp    */	ABSENT_STRING,
	/* 301: scp      */	ABSENT_STRING,
	/* 302: setf     */	ABSENT_STRING,
	/* 303: setb     */	ABSENT_STRING,
	/* 304: cpi      */	ABSENT_STRING,
	/* 305: lpi      */	ABSENT_STRING,
	/* 306: chr      */	ABSENT_STRING,
	/* 307: cvr      */	ABSENT_STRING,
	/* 308: defc     */	ABSENT_STRING,
	/* 309: swidm    */	ABSENT_STRING,
	/* 310: sdrfq    */	ABSENT_STRING,
	/* 311: sitm     */	ABSENT_STRING,
	/* 312: slm      */	ABSENT_STRING,
	/* 313: smicm    */	ABSENT_STRING,
	/* 314: snlq     */	ABSENT_STRING,
	/* 315: snrmq    */	ABSENT_STRING,
	/* 316: sshm     */	ABSENT_STRING,
	/* 317: ssubm    */	ABSENT_STRING,
	/* 318: ssupm    */	ABSENT_STRING,
	/* 319: sum      */	ABSENT_STRING,
	/* 320: rwidm    */	ABSENT_STRING,
	/* 321: ritm     */	ABSENT_STRING,
	/* 322: rlm      */	ABSENT_STRING,
	/* 323: rmicm    */	ABSENT_STRING,
	/* 324: rshm     */	ABSENT_STRING,
	/* 325: rsubm    */	ABSENT_STRING,
	/* 326: rsupm    */	ABSENT_STRING,
	/* 327: rum      */	ABSENT_STRING,
	/* 328: mhpa     */	ABSENT_STRING,
	/* 329: mcud1    */	ABSENT_STRING,
	/* 330: mcub1    */	ABSENT_STRING,
	/* 331: mcuf1    */	ABSENT_STRING,
	/* 332: mvpa     */	ABSENT_STRING,
	/* 333: mcuu1    */	ABSENT_STRING,
	/* 334: porder   */	ABSENT_STRING,
	/* 335: mcud     */	ABSENT_STRING,
	/* 336: mcub     */	ABSENT_STRING,
	/* 337: mcuf     */	ABSENT_STRING,
	/* 338: mcuu     */	ABSENT_STRING,
	/* 339: scs      */	ABSENT_STRING,
	/* 340: smgb     */	ABSENT_STRING,
	/* 341: smgbp    */	ABSENT_STRING,
	/* 342: smglp    */	ABSENT_STRING,
	/* 343: smgrp    */	ABSENT_STRING,
	/* 344: smgt     */	ABSENT_STRING,
	/* 345: smgtp    */	ABSENT_STRING,
	/* 346: sbim     */	ABSENT_STRING,
	/* 347: scsd     */	ABSENT_STRING,
	/* 348: rbim     */	ABSENT_STRING,
	/* 349: rcsd     */	ABSENT_STRING,
	/* 350: subcs    */	ABSENT_STRING,
	/* 351: supcs    */	ABSENT_STRING,
	/* 352: docr     */	ABSENT_STRING,
	/* 353: zerom    */	ABSENT_STRING,
	/* 354: csnm     */	ABSENT_STRING,
	/* 355: kmous    */	linux_s_kmous,
	/* 356: minfo    */	ABSENT_STRING,
	/* 357: reqmp    */	ABSENT_STRING,
	/* 358: getm     */	ABSENT_STRING,
	/* 359: setaf    */	linux_s_setaf,
	/* 360: setab    */	linux_s_setab,
	/* 361: pfxl     */	ABSENT_STRING,
	/* 362: devt     */	ABSENT_STRING,
	/* 363: csin     */	ABSENT_STRING,
	/* 364: s0ds     */	ABSENT_STRING,
	/* 365: s1ds     */	ABSENT_STRING,
	/* 366: s2ds     */	ABSENT_STRING,
	/* 367: s3ds     */	ABSENT_STRING,
	/* 368: smglr    */	ABSENT_STRING,
	/* 369: smgtb    */	ABSENT_STRING,
	/* 370: birep    */	ABSENT_STRING,
	/* 371: binel    */	ABSENT_STRING,
	/* 372: bicr     */	ABSENT_STRING,
	/* 373: colornm  */	ABSENT_STRING,
	/* 374: defbi    */	ABSENT_STRING,
	/* 375: endbi    */	ABSENT_STRING,
	/* 376: setcolor */	ABSENT_STRING,
	/* 377: slines   */	ABSENT_STRING,
	/* 378: dispc    */	ABSENT_STRING,
	/* 379: smpch    */	linux_s_smpch,
	/* 380: rmpch    */	linux_s_rmpch,
	/* 381: smsc     */	ABSENT_STRING,
	/* 382: rmsc     */	ABSENT_STRING,
	/* 383: pctrm    */	ABSENT_STRING,
	/* 384: scesc    */	ABSENT_STRING,
	/* 385: scesa    */	ABSENT_STRING,
	/* 386: ehhlm    */	ABSENT_STRING,
	/* 387: elhlm    */	ABSENT_STRING,
	/* 388: elohlm   */	ABSENT_STRING,
	/* 389: erhlm    */	ABSENT_STRING,
	/* 390: ethlm    */	ABSENT_STRING,
	/* 391: evhlm    */	ABSENT_STRING,
	/* 392: sgr1     */	ABSENT_STRING,
	/* 393: slength  */	ABSENT_STRING,
	/* 394: OTi2     */	ABSENT_STRING,
	/* 395: OTrs     */	ABSENT_STRING,
	/* 396: OTnl     */	ABSENT_STRING,
	/* 397: OTbc     */	ABSENT_STRING,
	/* 398: OTko     */	ABSENT_STRING,
	/* 399: OTma     */	ABSENT_STRING,
	/* 400: OTG2     */	ABSENT_STRING,
	/* 401: OTG3     */	ABSENT_STRING,
	/* 402: OTG1     */	ABSENT_STRING,
	/* 403: OTG4     */	ABSENT_STRING,
	/* 404: OTGR     */	ABSENT_STRING,
	/* 405: OTGL     */	ABSENT_STRING,
	/* 406: OTGU     */	ABSENT_STRING,
	/* 407: OTGD     */	ABSENT_STRING,
	/* 408: OTGH     */	ABSENT_STRING,
	/* 409: OTGV     */	ABSENT_STRING,
	/* 410: OTGC     */	ABSENT_STRING,
	/* 411: meml     */	ABSENT_STRING,
	/* 412: memu     */	ABSENT_STRING,
	/* 413: box1     */	ABSENT_STRING,
};
/* rxvt */

static char rxvt_alias_data[] = "rxvt|rxvt terminal emulator (X Window System)";

static char rxvt_s_bel          [] = "\007";
static char rxvt_s_cr           [] = "\015";
static char rxvt_s_csr          [] = "\033[%i%p1%d;%p2%dr";
static char rxvt_s_tbc          [] = "\033[3g";
static char rxvt_s_clear        [] = "\033[H\033[2J";
static char rxvt_s_el           [] = "\033[K";
static char rxvt_s_ed           [] = "\033[J";
static char rxvt_s_hpa          [] = "\033[%i%p1%dG";
static char rxvt_s_cup          [] = "\033[%i%p1%d;%p2%dH";
static char rxvt_s_cud1         [] = "\012";
static char rxvt_s_home         [] = "\033[H";
static char rxvt_s_civis        [] = "\033[?25l";
static char rxvt_s_cub1         [] = "\010";
static char rxvt_s_cnorm        [] = "\033[?25h";
static char rxvt_s_cuf1         [] = "\033[C";
static char rxvt_s_cuu1         [] = "\033[A";
static char rxvt_s_dl1          [] = "\033[M";
static char rxvt_s_smacs        [] = "\016";
static char rxvt_s_blink        [] = "\033[5m";
static char rxvt_s_bold         [] = "\033[1m";
static char rxvt_s_smcup        [] = "\0337\033[?47h";
static char rxvt_s_smir         [] = "\033[4h";
static char rxvt_s_rev          [] = "\033[7m";
static char rxvt_s_smso         [] = "\033[7m";
static char rxvt_s_smul         [] = "\033[4m";
static char rxvt_s_rmacs        [] = "\017";
static char rxvt_s_sgr0         [] = "\033[m\017";
static char rxvt_s_rmcup        [] = "\033[2J\033[?47l\0338";
static char rxvt_s_rmir         [] = "\033[4l";
static char rxvt_s_rmso         [] = "\033[27m";
static char rxvt_s_rmul         [] = "\033[24m";
static char rxvt_s_flash        [] = "\033[?5h$<100/>\033[?5l";
static char rxvt_s_is1          [] = "\033[?47l\033=\033[?1l";
static char rxvt_s_is2          [] = "\033[r\033[m\033[2J\033[H\033[?7h\033[?1;3;4;6l\033[4l";
static char rxvt_s_ich1         [] = "\033[@";
static char rxvt_s_il1          [] = "\033[L";
static char rxvt_s_kbs          [] = "\010";
static char rxvt_s_kdch1        [] = "\033[3~";
static char rxvt_s_kcud1        [] = "\033[B";
static char rxvt_s_kel          [] = "\033[8^";
static char rxvt_s_kf0          [] = "\033[21~";
static char rxvt_s_kf1          [] = "\033[11~";
static char rxvt_s_kf10         [] = "\033[21~";
static char rxvt_s_kf2          [] = "\033[12~";
static char rxvt_s_kf3          [] = "\033[13~";
static char rxvt_s_kf4          [] = "\033[14~";
static char rxvt_s_kf5          [] = "\033[15~";
static char rxvt_s_kf6          [] = "\033[17~";
static char rxvt_s_kf7          [] = "\033[18~";
static char rxvt_s_kf8          [] = "\033[19~";
static char rxvt_s_kf9          [] = "\033[20~";
static char rxvt_s_khome        [] = "\033[7~";
static char rxvt_s_kich1        [] = "\033[2~";
static char rxvt_s_kcub1        [] = "\033[D";
static char rxvt_s_knp          [] = "\033[6~";
static char rxvt_s_kpp          [] = "\033[5~";
static char rxvt_s_kcuf1        [] = "\033[C";
static char rxvt_s_kind         [] = "\033[a";
static char rxvt_s_kri          [] = "\033[b";
static char rxvt_s_kcuu1        [] = "\033[A";
static char rxvt_s_rmkx         [] = "\033>";
static char rxvt_s_smkx         [] = "\033=";
static char rxvt_s_dl           [] = "\033[%p1%dM";
static char rxvt_s_cud          [] = "\033[%p1%dB";
static char rxvt_s_ich          [] = "\033[%p1%d@";
static char rxvt_s_il           [] = "\033[%p1%dL";
static char rxvt_s_cub          [] = "\033[%p1%dD";
static char rxvt_s_cuf          [] = "\033[%p1%dC";
static char rxvt_s_cuu          [] = "\033[%p1%dA";
static char rxvt_s_rs1          [] = "\033>\033[1;3;4;5;6l\033[?7h\033[m\033[r\033[2J\033[H";
static char rxvt_s_rs2          [] = "\033[r\033[m\033[2J\033[H\033[?7h\033[?1;3;4;6l\033[4l\033>\033[?1000l\033[?25h";
static char rxvt_s_rc           [] = "\0338";
static char rxvt_s_vpa          [] = "\033[%i%p1%dd";
static char rxvt_s_sc           [] = "\0337";
static char rxvt_s_ind          [] = "\012";
static char rxvt_s_ri           [] = "\033M";
static char rxvt_s_sgr          [] = "\033[0%?%p6%t;1%;%?%p2%t;4%;%?%p1%p3%|%t;7%;%?%p4%t;5%;m%?%p9%t\016%e\017%;";
static char rxvt_s_hts          [] = "\033H";
static char rxvt_s_ht           [] = "\011";
static char rxvt_s_ka1          [] = "\033Ow";
static char rxvt_s_ka3          [] = "\033Oy";
static char rxvt_s_kb2          [] = "\033Ou";
static char rxvt_s_kc1          [] = "\033Oq";
static char rxvt_s_kc3          [] = "\033Os";
static char rxvt_s_acsc         [] = "``aaffggjjkkllmmnnooppqqrrssttuuvvwwxxyyzz{{||}}~~";
static char rxvt_s_kcbt         [] = "\033[Z";
static char rxvt_s_enacs        [] = "\033(B\033)0";
static char rxvt_s_kend         [] = "\033[8~";
static char rxvt_s_kent         [] = "\033OM";
static char rxvt_s_kfnd         [] = "\033[1~";
static char rxvt_s_kDC          [] = "\033[3$";
static char rxvt_s_kslt         [] = "\033[4~";
static char rxvt_s_kEND         [] = "\033[8$";
static char rxvt_s_kHOM         [] = "\033[7$";
static char rxvt_s_kIC          [] = "\033[2$";
static char rxvt_s_kLFT         [] = "\033[d";
static char rxvt_s_kNXT         [] = "\033[6$";
static char rxvt_s_kPRV         [] = "\033[5$";
static char rxvt_s_kRIT         [] = "\033[c";
static char rxvt_s_kf11         [] = "\033[23~";
static char rxvt_s_kf12         [] = "\033[24~";
static char rxvt_s_kf13         [] = "\033[25~";
static char rxvt_s_kf14         [] = "\033[26~";
static char rxvt_s_kf15         [] = "\033[28~";
static char rxvt_s_kf16         [] = "\033[29~";
static char rxvt_s_kf17         [] = "\033[31~";
static char rxvt_s_kf18         [] = "\033[32~";
static char rxvt_s_kf19         [] = "\033[33~";
static char rxvt_s_kf20         [] = "\033[34~";
static char rxvt_s_kf21         [] = "\033[23$";
static char rxvt_s_kf22         [] = "\033[24$";
static char rxvt_s_kf23         [] = "\033[11^";
static char rxvt_s_kf24         [] = "\033[12^";
static char rxvt_s_kf25         [] = "\033[13^";
static char rxvt_s_kf26         [] = "\033[14^";
static char rxvt_s_kf27         [] = "\033[15^";
static char rxvt_s_kf28         [] = "\033[17^";
static char rxvt_s_kf29         [] = "\033[18^";
static char rxvt_s_kf30         [] = "\033[19^";
static char rxvt_s_kf31         [] = "\033[20^";
static char rxvt_s_kf32         [] = "\033[21^";
static char rxvt_s_kf33         [] = "\033[23^";
static char rxvt_s_kf34         [] = "\033[24^";
static char rxvt_s_kf35         [] = "\033[25^";
static char rxvt_s_kf36         [] = "\033[26^";
static char rxvt_s_kf37         [] = "\033[28^";
static char rxvt_s_kf38         [] = "\033[29^";
static char rxvt_s_kf39         [] = "\033[31^";
static char rxvt_s_kf40         [] = "\033[32^";
static char rxvt_s_kf41         [] = "\033[33^";
static char rxvt_s_kf42         [] = "\033[34^";
static char rxvt_s_kf43         [] = "\033[23@";
static char rxvt_s_kf44         [] = "\033[24@";
static char rxvt_s_el1          [] = "\033[1K";
static char rxvt_s_u6           [] = "\033[%i%d;%dR";
static char rxvt_s_u7           [] = "\033[6n";
static char rxvt_s_u8           [] = "\033[?1;2c";
static char rxvt_s_u9           [] = "\033[c";
static char rxvt_s_op           [] = "\033[39;49m";
static char rxvt_s_kmous        [] = "\033[M";
static char rxvt_s_setaf        [] = "\033[3%p1%dm";
static char rxvt_s_setab        [] = "\033[4%p1%dm";
static char rxvt_s_s0ds         [] = "\033(B";
static char rxvt_s_s1ds         [] = "\033(0";

static char rxvt_bool_data[] = {
	/*   0: bw       */	FALSE,
	/*   1: am       */	TRUE,
	/*   2: xsb      */	FALSE,
	/*   3: xhp      */	FALSE,
	/*   4: xenl     */	TRUE,
	/*   5: eo       */	TRUE,
	/*   6: gn       */	FALSE,
	/*   7: hc       */	FALSE,
	/*   8: km       */	FALSE,
	/*   9: hs       */	FALSE,
	/*  10: in       */	FALSE,
	/*  11: da       */	FALSE,
	/*  12: db       */	FALSE,
	/*  13: mir      */	TRUE,
	/*  14: msgr     */	TRUE,
	/*  15: os       */	FALSE,
	/*  16: eslok    */	FALSE,
	/*  17: xt       */	FALSE,
	/*  18: hz       */	FALSE,
	/*  19: ul       */	FALSE,
	/*  20: xon      */	TRUE,
	/*  21: nxon     */	FALSE,
	/*  22: mc5i     */	FALSE,
	/*  23: chts     */	FALSE,
	/*  24: nrrmc    */	FALSE,
	/*  25: npc      */	FALSE,
	/*  26: ndscr    */	FALSE,
	/*  27: ccc      */	FALSE,
	/*  28: bce      */	TRUE,
	/*  29: hls      */	FALSE,
	/*  30: xhpa     */	FALSE,
	/*  31: crxm     */	FALSE,
	/*  32: daisy    */	FALSE,
	/*  33: xvpa     */	FALSE,
	/*  34: sam      */	FALSE,
	/*  35: cpix     */	FALSE,
	/*  36: lpix     */	FALSE,
	/*  37: OTbs     */	TRUE,
	/*  38: OTns     */	FALSE,
	/*  39: OTnc     */	FALSE,
	/*  40: OTMT     */	FALSE,
	/*  41: OTNL     */	FALSE,
	/*  42: OTpt     */	FALSE,
	/*  43: OTxr     */	FALSE,
};
static NCURSES_INT2 rxvt_number_data[] = {
	/*   0: cols     */	80,
	/*   1: it       */	8,
	/*   2: lines    */	24,
	/*   3: lm       */	ABSENT_NUMERIC,
	/*   4: xmc      */	ABSENT_NUMERIC,
	/*   5: pb       */	ABSENT_NUMERIC,
	/*   6: vt       */	ABSENT_NUMERIC,
	/*   7: wsl      */	ABSENT_NUMERIC,
	/*   8: nlab     */	ABSENT_NUMERIC,
	/*   9: lh       */	ABSENT_NUMERIC,
	/*  10: lw       */	ABSENT_NUMERIC,
	/*  11: ma       */	ABSENT_NUMERIC,
	/*  12: wnum     */	ABSENT_NUMERIC,
	/*  13: colors   */	8,
	/*  14: pairs    */	64,
	/*  15: ncv      */	CANCELLED_NUMERIC,
	/*  16: bufsz    */	ABSENT_NUMERIC,
	/*  17: spinv    */	ABSENT_NUMERIC,
	/*  18: spinh    */	ABSENT_NUMERIC,
	/*  19: maddr    */	ABSENT_NUMERIC,
	/*  20: mjump    */	ABSENT_NUMERIC,
	/*  21: mcs      */	ABSENT_NUMERIC,
	/*  22: mls      */	ABSENT_NUMERIC,
	/*  23: npins    */	ABSENT_NUMERIC,
	/*  24: orc      */	ABSENT_NUMERIC,
	/*  25: orl      */	ABSENT_NUMERIC,
	/*  26: orhi     */	ABSENT_NUMERIC,
	/*  27: orvi     */	ABSENT_NUMERIC,
	/*  28: cps      */	ABSENT_NUMERIC,
	/*  29: widcs    */	ABSENT_NUMERIC,
	/*  30: btns     */	ABSENT_NUMERIC,
	/*  31: bitwin   */	ABSENT_NUMERIC,
	/*  32: bitype   */	ABSENT_NUMERIC,
	/*  33: OTug     */	ABSENT_NUMERIC,
	/*  34: OTdC     */	ABSENT_NUMERIC,
	/*  35: OTdN     */	ABSENT_NUMERIC,
	/*  36: OTdB     */	ABSENT_NUMERIC,
	/*  37: OTdT     */	ABSENT_NUMERIC,
	/*  38: OTkn     */	ABSENT_NUMERIC,
};
static char * rxvt_string_data[] = {
	/*   0: cbt      */	ABSENT_STRING,
	/*   1: bel      */	rxvt_s_bel,
	/*   2: cr       */	rxvt_s_cr,
	/*   3: csr      */	rxvt_s_csr,
	/*   4: tbc      */	rxvt_s_tbc,
	/*   5: clear    */	rxvt_s_clear,
	/*   6: el       */	rxvt_s_el,
	/*   7: ed       */	rxvt_s_ed,
	/*   8: hpa      */	rxvt_s_hpa,
	/*   9: cmdch    */	ABSENT_STRING,
	/*  10: cup      */	rxvt_s_cup,
	/*  11: cud1     */	rxvt_s_cud1,
	/*  12: home     */	rxvt_s_home,
	/*  13: civis    */	rxvt_s_civis,
	/*  14: cub1     */	rxvt_s_cub1,
	/*  15: mrcup    */	ABSENT_STRING,
	/*  16: cnorm    */	rxvt_s_cnorm,
	/*  17: cuf1     */	rxvt_s_cuf1,
	/*  18: ll       */	ABSENT_STRING,
	/*  19: cuu1     */	rxvt_s_cuu1,
	/*  20: cvvis    */	ABSENT_STRING,
	/*  21: dch1     */	ABSENT_STRING,
	/*  22: dl1      */	rxvt_s_dl1,
	/*  23: dsl      */	ABSENT_STRING,
	/*  24: hd       */	ABSENT_STRING,
	/*  25: smacs    */	rxvt_s_smacs,
	/*  26: blink    */	rxvt_s_blink,
	/*  27: bold     */	rxvt_s_bold,
	/*  28: smcup    */	rxvt_s_smcup,
	/*  29: smdc     */	ABSENT_STRING,
	/*  30: dim      */	ABSENT_STRING,
	/*  31: smir     */	rxvt_s_smir,
	/*  32: invis    */	ABSENT_STRING,
	/*  33: prot     */	ABSENT_STRING,
	/*  34: rev      */	rxvt_s_rev,
	/*  35: smso     */	rxvt_s_smso,
	/*  36: smul     */	rxvt_s_smul,
	/*  37: ech      */	ABSENT_STRING,
	/*  38: rmacs    */	rxvt_s_rmacs,
	/*  39: sgr0     */	rxvt_s_sgr0,
	/*  40: rmcup    */	rxvt_s_rmcup,
	/*  41: rmdc     */	ABSENT_STRING,
	/*  42: rmir     */	rxvt_s_rmir,
	/*  43: rmso     */	rxvt_s_rmso,
	/*  44: rmul     */	rxvt_s_rmul,
	/*  45: flash    */	rxvt_s_flash,
	/*  46: ff       */	ABSENT_STRING,
	/*  47: fsl      */	ABSENT_STRING,
	/*  48: is1      */	rxvt_s_is1,
	/*  49: is2      */	rxvt_s_is2,
	/*  50: is3      */	ABSENT_STRING,
	/*  51: if       */	ABSENT_STRING,
	/*  52: ich1     */	rxvt_s_ich1,
	/*  53: il1      */	rxvt_s_il1,
	/*  54: ip       */	ABSENT_STRING,
	/*  55: kbs      */	rxvt_s_kbs,
	/*  56: ktbc     */	ABSENT_STRING,
	/*  57: kclr     */	ABSENT_STRING,
	/*  58: kctab    */	ABSENT_STRING,
	/*  59: kdch1    */	rxvt_s_kdch1,
	/*  60: kdl1     */	ABSENT_STRING,
	/*  61: kcud1    */	rxvt_s_kcud1,
	/*  62: krmir    */	ABSENT_STRING,
	/*  63: kel      */	rxvt_s_kel,
	/*  64: ked      */	ABSENT_STRING,
	/*  65: kf0      */	rxvt_s_kf0,
	/*  66: kf1      */	rxvt_s_kf1,
	/*  67: kf10     */	rxvt_s_kf10,
	/*  68: kf2      */	rxvt_s_kf2,
	/*  69: kf3      */	rxvt_s_kf3,
	/*  70: kf4      */	rxvt_s_kf4,
	/*  71: kf5      */	rxvt_s_kf5,
	/*  72: kf6      */	rxvt_s_kf6,
	/*  73: kf7      */	rxvt_s_kf7,
	/*  74: kf8      */	rxvt_s_kf8,
	/*  75: kf9      */	rxvt_s_kf9,
	/*  76: khome    */	rxvt_s_khome,
	/*  77: kich1    */	rxvt_s_kich1,
	/*  78: kil1     */	ABSENT_STRING,
	/*  79: kcub1    */	rxvt_s_kcub1,
	/*  80: kll      */	ABSENT_STRING,
	/*  81: knp      */	rxvt_s_knp,
	/*  82: kpp      */	rxvt_s_kpp,
	/*  83: kcuf1    */	rxvt_s_kcuf1,
	/*  84: kind     */	rxvt_s_kind,
	/*  85: kri      */	rxvt_s_kri,
	/*  86: khts     */	ABSENT_STRING,
	/*  87: kcuu1    */	rxvt_s_kcuu1,
	/*  88: rmkx     */	rxvt_s_rmkx,
	/*  89: smkx     */	rxvt_s_smkx,
	/*  90: lf0      */	ABSENT_STRING,
	/*  91: lf1      */	ABSENT_STRING,
	/*  92: lf10     */	ABSENT_STRING,
	/*  93: lf2      */	ABSENT_STRING,
	/*  94: lf3      */	ABSENT_STRING,
	/*  95: lf4      */	ABSENT_STRING,
	/*  96: lf5      */	ABSENT_STRING,
	/*  97: lf6      */	ABSENT_STRING,
	/*  98: lf7      */	ABSENT_STRING,
	/*  99: lf8      */	ABSENT_STRING,
	/* 100: lf9      */	ABSENT_STRING,
	/* 101: rmm      */	ABSENT_STRING,
	/* 102: smm      */	ABSENT_STRING,
	/* 103: nel      */	ABSENT_STRING,
	/* 104: pad      */	ABSENT_STRING,
	/* 105: dch      */	ABSENT_STRING,
	/* 106: dl       */	rxvt_s_dl,
	/* 107: cud      */	rxvt_s_cud,
	/* 108: ich      */	rxvt_s_ich,
	/* 109: indn     */	ABSENT_STRING,
	/* 110: il       */	rxvt_s_il,
	/* 111: cub      */	rxvt_s_cub,
	/* 112: cuf      */	rxvt_s_cuf,
	/* 113: rin      */	ABSENT_STRING,
	/* 114: cuu      */	rxvt_s_cuu,
	/* 115: pfkey    */	ABSENT_STRING,
	/* 116: pfloc    */	ABSENT_STRING,
	/* 117: pfx      */	ABSENT_STRING,
	/* 118: mc0      */	ABSENT_STRING,
	/* 119: mc4      */	ABSENT_STRING,
	/* 120: mc5      */	ABSENT_STRING,
	/* 121: rep      */	ABSENT_STRING,
	/* 122: rs1      */	rxvt_s_rs1,
	/* 123: rs2      */	rxvt_s_rs2,
	/* 124: rs3      */	ABSENT_STRING,
	/* 125: rf       */	ABSENT_STRING,
	/* 126: rc       */	rxvt_s_rc,
	/* 127: vpa      */	rxvt_s_vpa,
	/* 128: sc       */	rxvt_s_sc,
	/* 129: ind      */	rxvt_s_ind,
	/* 130: ri       */	rxvt_s_ri,
	/* 131: sgr      */	rxvt_s_sgr,
	/* 132: hts      */	rxvt_s_hts,
	/* 133: wind     */	ABSENT_STRING,
	/* 134: ht       */	rxvt_s_ht,
	/* 135: tsl      */	ABSENT_STRING,
	/* 136: uc       */	ABSENT_STRING,
	/* 137: hu       */	ABSENT_STRING,
	/* 138: iprog    */	ABSENT_STRING,
	/* 139: ka1      */	rxvt_s_ka1,
	/* 140: ka3      */	rxvt_s_ka3,
	/* 141: kb2      */	rxvt_s_kb2,
	/* 142: kc1      */	rxvt_s_kc1,
	/* 143: kc3      */	rxvt_s_kc3,
	/* 144: mc5p     */	ABSENT_STRING,
	/* 145: rmp      */	ABSENT_STRING,
	/* 146: acsc     */	rxvt_s_acsc,
	/* 147: pln      */	ABSENT_STRING,
	/* 148: kcbt     */	rxvt_s_kcbt,
	/* 149: smxon    */	ABSENT_STRING,
	/* 150: rmxon    */	ABSENT_STRING,
	/* 151: smam     */	ABSENT_STRING,
	/* 152: rmam     */	ABSENT_STRING,
	/* 153: xonc     */	ABSENT_STRING,
	/* 154: xoffc    */	ABSENT_STRING,
	/* 155: enacs    */	rxvt_s_enacs,
	/* 156: smln     */	ABSENT_STRING,
	/* 157: rmln     */	ABSENT_STRING,
	/* 158: kbeg     */	ABSENT_STRING,
	/* 159: kcan     */	ABSENT_STRING,
	/* 160: kclo     */	ABSENT_STRING,
	/* 161: kcmd     */	ABSENT_STRING,
	/* 162: kcpy     */	ABSENT_STRING,
	/* 163: kcrt     */	ABSENT_STRING,
	/* 164: kend     */	rxvt_s_kend,
	/* 165: kent     */	rxvt_s_kent,
	/* 166: kext     */	ABSENT_STRING,
	/* 167: kfnd     */	rxvt_s_kfnd,
	/* 168: khlp     */	ABSENT_STRING,
	/* 169: kmrk     */	ABSENT_STRING,
	/* 170: kmsg     */	ABSENT_STRING,
	/* 171: kmov     */	ABSENT_STRING,
	/* 172: knxt     */	ABSENT_STRING,
	/* 173: kopn     */	ABSENT_STRING,
	/* 174: kopt     */	ABSENT_STRING,
	/* 175: kprv     */	ABSENT_STRING,
	/* 176: kprt     */	ABSENT_STRING,
	/* 177: krdo     */	ABSENT_STRING,
	/* 178: kref     */	ABSENT_STRING,
	/* 179: krfr     */	ABSENT_STRING,
	/* 180: krpl     */	ABSENT_STRING,
	/* 181: krst     */	ABSENT_STRING,
	/* 182: kres     */	ABSENT_STRING,
	/* 183: ksav     */	ABSENT_STRING,
	/* 184: kspd     */	ABSENT_STRING,
	/* 185: kund     */	ABSENT_STRING,
	/* 186: kBEG     */	ABSENT_STRING,
	/* 187: kCAN     */	ABSENT_STRING,
	/* 188: kCMD     */	ABSENT_STRING,
	/* 189: kCPY     */	ABSENT_STRING,
	/* 190: kCRT     */	ABSENT_STRING,
	/* 191: kDC      */	rxvt_s_kDC,
	/* 192: kDL      */	ABSENT_STRING,
	/* 193: kslt     */	rxvt_s_kslt,
	/* 194: kEND     */	rxvt_s_kEND,
	/* 195: kEOL     */	ABSENT_STRING,
	/* 196: kEXT     */	ABSENT_STRING,
	/* 197: kFND     */	ABSENT_STRING,
	/* 198: kHLP     */	ABSENT_STRING,
	/* 199: kHOM     */	rxvt_s_kHOM,
	/* 200: kIC      */	rxvt_s_kIC,
	/* 201: kLFT     */	rxvt_s_kLFT,
	/* 202: kMSG     */	ABSENT_STRING,
	/* 203: kMOV     */	ABSENT_STRING,
	/* 204: kNXT     */	rxvt_s_kNXT,
	/* 205: kOPT     */	ABSENT_STRING,
	/* 206: kPRV     */	rxvt_s_kPRV,
	/* 207: kPRT     */	ABSENT_STRING,
	/* 208: kRDO     */	ABSENT_STRING,
	/* 209: kRPL     */	ABSENT_STRING,
	/* 210: kRIT     */	rxvt_s_kRIT,
	/* 211: kRES     */	ABSENT_STRING,
	/* 212: kSAV     */	ABSENT_STRING,
	/* 213: kSPD     */	ABSENT_STRING,
	/* 214: kUND     */	ABSENT_STRING,
	/* 215: rfi      */	ABSENT_STRING,
	/* 216: kf11     */	rxvt_s_kf11,
	/* 217: kf12     */	rxvt_s_kf12,
	/* 218: kf13     */	rxvt_s_kf13,
	/* 219: kf14     */	rxvt_s_kf14,
	/* 220: kf15     */	rxvt_s_kf15,
	/* 221: kf16     */	rxvt_s_kf16,
	/* 222: kf17     */	rxvt_s_kf17,
	/* 223: kf18     */	rxvt_s_kf18,
	/* 224: kf19     */	rxvt_s_kf19,
	/* 225: kf20     */	rxvt_s_kf20,
	/* 226: kf21     */	rxvt_s_kf21,
	/* 227: kf22     */	rxvt_s_kf22,
	/* 228: kf23     */	rxvt_s_kf23,
	/* 229: kf24     */	rxvt_s_kf24,
	/* 230: kf25     */	rxvt_s_kf25,
	/* 231: kf26     */	rxvt_s_kf26,
	/* 232: kf27     */	rxvt_s_kf27,
	/* 233: kf28     */	rxvt_s_kf28,
	/* 234: kf29     */	rxvt_s_kf29,
	/* 235: kf30     */	rxvt_s_kf30,
	/* 236: kf31     */	rxvt_s_kf31,
	/* 237: kf32     */	rxvt_s_kf32,
	/* 238: kf33     */	rxvt_s_kf33,
	/* 239: kf34     */	rxvt_s_kf34,
	/* 240: kf35     */	rxvt_s_kf35,
	/* 241: kf36     */	rxvt_s_kf36,
	/* 242: kf37     */	rxvt_s_kf37,
	/* 243: kf38     */	rxvt_s_kf38,
	/* 244: kf39     */	rxvt_s_kf39,
	/* 245: kf40     */	rxvt_s_kf40,
	/* 246: kf41     */	rxvt_s_kf41,
	/* 247: kf42     */	rxvt_s_kf42,
	/* 248: kf43     */	rxvt_s_kf43,
	/* 249: kf44     */	rxvt_s_kf44,
	/* 250: kf45     */	ABSENT_STRING,
	/* 251: kf46     */	ABSENT_STRING,
	/* 252: kf47     */	ABSENT_STRING,
	/* 253: kf48     */	ABSENT_STRING,
	/* 254: kf49     */	ABSENT_STRING,
	/* 255: kf50     */	ABSENT_STRING,
	/* 256: kf51     */	ABSENT_STRING,
	/* 257: kf52     */	ABSENT_STRING,
	/* 258: kf53     */	ABSENT_STRING,
	/* 259: kf54     */	ABSENT_STRING,
	/* 260: kf55     */	ABSENT_STRING,
	/* 261: kf56     */	ABSENT_STRING,
	/* 262: kf57     */	ABSENT_STRING,
	/* 263: kf58     */	ABSENT_STRING,
	/* 264: kf59     */	ABSENT_STRING,
	/* 265: kf60     */	ABSENT_STRING,
	/* 266: kf61     */	ABSENT_STRING,
	/* 267: kf62     */	ABSENT_STRING,
	/* 268: kf63     */	ABSENT_STRING,
	/* 269: el1      */	rxvt_s_el1,
	/* 270: mgc      */	ABSENT_STRING,
	/* 271: smgl     */	ABSENT_STRING,
	/* 272: smgr     */	ABSENT_STRING,
	/* 273: fln      */	ABSENT_STRING,
	/* 274: sclk     */	ABSENT_STRING,
	/* 275: dclk     */	ABSENT_STRING,
	/* 276: rmclk    */	ABSENT_STRING,
	/* 277: cwin     */	ABSENT_STRING,
	/* 278: wingo    */	ABSENT_STRING,
	/* 279: hup      */	ABSENT_STRING,
	/* 280: dial     */	ABSENT_STRING,
	/* 281: qdial    */	ABSENT_STRING,
	/* 282: tone     */	ABSENT_STRING,
	/* 283: pulse    */	ABSENT_STRING,
	/* 284: hook     */	ABSENT_STRING,
	/* 285: pause    */	ABSENT_STRING,
	/* 286: wait     */	ABSENT_STRING,
	/* 287: u0       */	ABSENT_STRING,
	/* 288: u1       */	ABSENT_STRING,
	/* 289: u2       */	ABSENT_STRING,
	/* 290: u3       */	ABSENT_STRING,
	/* 291: u4       */	ABSENT_STRING,
	/* 292: u5       */	ABSENT_STRING,
	/* 293: u6       */	rxvt_s_u6,
	/* 294: u7       */	rxvt_s_u7,
	/* 295: u8       */	rxvt_s_u8,
	/* 296: u9       */	rxvt_s_u9,
	/* 297: op       */	rxvt_s_op,
	/* 298: oc       */	ABSENT_STRING,
	/* 299: initc    */	ABSENT_STRING,
	/* 300: initp    */	ABSENT_STRING,
	/* 301: scp      */	ABSENT_STRING,
	/* 302: setf     */	ABSENT_STRING,
	/* 303: setb     */	ABSENT_STRING,
	/* 304: cpi      */	ABSENT_STRING,
	/* 305: lpi      */	ABSENT_STRING,
	/* 306: chr      */	ABSENT_STRING,
	/* 307: cvr      */	ABSENT_STRING,
	/* 308: defc     */	ABSENT_STRING,
	/* 309: swidm    */	ABSENT_STRING,
	/* 310: sdrfq    */	ABSENT_STRING,
	/* 311: sitm     */	ABSENT_STRING,
	/* 312: slm      */	ABSENT_STRING,
	/* 313: smicm    */	ABSENT_STRING,
	/* 314: snlq     */	ABSENT_STRING,
	/* 315: snrmq    */	ABSENT_STRING,
	/* 316: sshm     */	ABSENT_STRING,
	/* 317: ssubm    */	ABSENT_STRING,
	/* 318: ssupm    */	ABSENT_STRING,
	/* 319: sum      */	ABSENT_STRING,
	/* 320: rwidm    */	ABSENT_STRING,
	/* 321: ritm     */	ABSENT_STRING,
	/* 322: rlm      */	ABSENT_STRING,
	/* 323: rmicm    */	ABSENT_STRING,
	/* 324: rshm     */	ABSENT_STRING,
	/* 325: rsubm    */	ABSENT_STRING,
	/* 326: rsupm    */	ABSENT_STRING,
	/* 327: rum      */	ABSENT_STRING,
	/* 328: mhpa     */	ABSENT_STRING,
	/* 329: mcud1    */	ABSENT_STRING,
	/* 330: mcub1    */	ABSENT_STRING,
	/* 331: mcuf1    */	ABSENT_STRING,
	/* 332: mvpa     */	ABSENT_STRING,
	/* 333: mcuu1    */	ABSENT_STRING,
	/* 334: porder   */	ABSENT_STRING,
	/* 335: mcud     */	ABSENT_STRING,
	/* 336: mcub     */	ABSENT_STRING,
	/* 337: mcuf     */	ABSENT_STRING,
	/* 338: mcuu     */	ABSENT_STRING,
	/* 339: scs      */	ABSENT_STRING,
	/* 340: smgb     */	ABSENT_STRING,
	/* 341: smgbp    */	ABSENT_STRING,
	/* 342: smglp    */	ABSENT_STRING,
	/* 343: smgrp    */	ABSENT_STRING,
	/* 344: smgt     */	ABSENT_STRING,
	/* 345: smgtp    */	ABSENT_STRING,
	/* 346: sbim     */	ABSENT_STRING,
	/* 347: scsd     */	ABSENT_STRING,
	/* 348: rbim     */	ABSENT_STRING,
	/* 349: rcsd     */	ABSENT_STRING,
	/* 350: subcs    */	ABSENT_STRING,
	/* 351: supcs    */	ABSENT_STRING,
	/* 352: docr     */	ABSENT_STRING,
	/* 353: zerom    */	ABSENT_STRING,
	/* 354: csnm     */	ABSENT_STRING,
	/* 355: kmous    */	rxvt_s_kmous,
	/* 356: minfo    */	ABSENT_STRING,
	/* 357: reqmp    */	ABSENT_STRING,
	/* 358: getm     */	ABSENT_STRING,
	/* 359: setaf    */	rxvt_s_setaf,
	/* 360: setab    */	rxvt_s_setab,
	/* 361: pfxl     */	ABSENT_STRING,
	/* 362: devt     */	ABSENT_STRING,
	/* 363: csin     */	ABSENT_STRING,
	/* 364: s0ds     */	rxvt_s_s0ds,
	/* 365: s1ds     */	rxvt_s_s1ds,
	/* 366: s2ds     */	ABSENT_STRING,
	/* 367: s3ds     */	ABSENT_STRING,
	/* 368: smglr    */	ABSENT_STRING,
	/* 369: smgtb    */	ABSENT_STRING,
	/* 370: birep    */	ABSENT_STRING,
	/* 371: binel    */	ABSENT_STRING,
	/* 372: bicr     */	ABSENT_STRING,
	/* 373: colornm  */	ABSENT_STRING,
	/* 374: defbi    */	ABSENT_STRING,
	/* 375: endbi    */	ABSENT_STRING,
	/* 376: setcolor */	ABSENT_STRING,
	/* 377: slines   */	ABSENT_STRING,
	/* 378: dispc    */	ABSENT_STRING,
	/* 379: smpch    */	ABSENT_STRING,
	/* 380: rmpch    */	ABSENT_STRING,
	/* 381: smsc     */	ABSENT_STRING,
	/* 382: rmsc     */	ABSENT_STRING,
	/* 383: pctrm    */	ABSENT_STRING,
	/* 384: scesc    */	ABSENT_STRING,
	/* 385: scesa    */	ABSENT_STRING,
	/* 386: ehhlm    */	ABSENT_STRING,
	/* 387: elhlm    */	ABSENT_STRING,
	/* 388: elohlm   */	ABSENT_STRING,
	/* 389: erhlm    */	ABSENT_STRING,
	/* 390: ethlm    */	ABSENT_STRING,
	/* 391: evhlm    */	ABSENT_STRING,
	/* 392: sgr1     */	ABSENT_STRING,
	/* 393: slength  */	ABSENT_STRING,
	/* 394: OTi2     */	ABSENT_STRING,
	/* 395: OTrs     */	ABSENT_STRING,
	/* 396: OTnl     */	ABSENT_STRING,
	/* 397: OTbc     */	ABSENT_STRING,
	/* 398: OTko     */	ABSENT_STRING,
	/* 399: OTma     */	ABSENT_STRING,
	/* 400: OTG2     */	ABSENT_STRING,
	/* 401: OTG3     */	ABSENT_STRING,
	/* 402: OTG1     */	ABSENT_STRING,
	/* 403: OTG4     */	ABSENT_STRING,
	/* 404: OTGR     */	ABSENT_STRING,
	/* 405: OTGL     */	ABSENT_STRING,
	/* 406: OTGU     */	ABSENT_STRING,
	/* 407: OTGD     */	ABSENT_STRING,
	/* 408: OTGH     */	ABSENT_STRING,
	/* 409: OTGV     */	ABSENT_STRING,
	/* 410: OTGC     */	ABSENT_STRING,
	/* 411: meml     */	ABSENT_STRING,
	/* 412: memu     */	ABSENT_STRING,
	/* 413: box1     */	ABSENT_STRING,
};
/* vt100 */

static char vt100_alias_data[] = "vt100|vt100-am|dec vt100 (w/advanced video)";

static char vt100_s_bel         [] = "\007";
static char vt100_s_cr          [] = "\015";
static char vt100_s_csr         [] = "\033[%i%p1%d;%p2%dr";
static char vt100_s_tbc         [] = "\033[3g";
static char vt100_s_clear       [] = "\033[H\033[J$<50>";
static char vt100_s_el          [] = "\033[K$<3>";
static char vt100_s_ed          [] = "\033[J$<50>";
static char vt100_s_cup         [] = "\033[%i%p1%d;%p2%dH$<5>";
static char vt100_s_cud1        [] = "\012";
static char vt100_s_home        [] = "\033[H";
static char vt100_s_cub1        [] = "\010";
static char vt100_s_cuf1        [] = "\033[C$<2>";
static char vt100_s_cuu1        [] = "\033[A$<2>";
static char vt100_s_smacs       [] = "\016";
static char vt100_s_blink       [] = "\033[5m$<2>";
static char vt100_s_bold        [] = "\033[1m$<2>";
static char vt100_s_rev         [] = "\033[7m$<2>";
static char vt100_s_smso        [] = "\033[7m$<2>";
static char vt100_s_smul        [] = "\033[4m$<2>";
static char vt100_s_rmacs       [] = "\017";
static char vt100_s_sgr0        [] = "\033[m\017$<2>";
static char vt100_s_rmso        [] = "\033[m$<2>";
static char vt100_s_rmul        [] = "\033[m$<2>";
static char vt100_s_kbs         [] = "\010";
static char vt100_s_kcud1       [] = "\033OB";
static char vt100_s_kf0         [] = "\033Oy";
static char vt100_s_kf1         [] = "\033OP";
static char vt100_s_kf10        [] = "\033Ox";
static char vt100_s_kf2         [] = "\033OQ";
static char vt100_s_kf3         [] = "\033OR";
static char vt100_s_kf4         [] = "\033OS";
static char vt100_s_kf5         [] = "\033Ot";
static char vt100_s_kf6         [] = "\033Ou";
static char vt100_s_kf7         [] = "\033Ov";
static char vt100_s_kf8         [] = "\033Ol";
static char vt100_s_kf9         [] = "\033Ow";
static char vt100_s_kcub1       [] = "\033OD";
static char vt100_s_kcuf1       [] = "\033OC";
static char vt100_s_kcuu1       [] = "\033OA";
static char vt100_s_rmkx        [] = "\033[?1l\033>";
static char vt100_s_smkx        [] = "\033[?1h\033=";
static char vt100_s_lf1         [] = "pf1";
static char vt100_s_lf2         [] = "pf2";
static char vt100_s_lf3         [] = "pf3";
static char vt100_s_lf4         [] = "pf4";
static char vt100_s_cud         [] = "\033[%p1%dB";
static char vt100_s_cub         [] = "\033[%p1%dD";
static char vt100_s_cuf         [] = "\033[%p1%dC";
static char vt100_s_cuu         [] = "\033[%p1%dA";
static char vt100_s_mc0         [] = "\033[0i";
static char vt100_s_mc4         [] = "\033[4i";
static char vt100_s_mc5         [] = "\033[5i";
static char vt100_s_rs2         [] = "\033<\033>\033[?3;4;5l\033[?7;8h\033[r";
static char vt100_s_rc          [] = "\0338";
static char vt100_s_sc          [] = "\0337";
static char vt100_s_ind         [] = "\012";
static char vt100_s_ri          [] = "\033M$<5>";
static char vt100_s_sgr         [] = "\033[0%?%p1%p6%|%t;1%;%?%p2%t;4%;%?%p1%p3%|%t;7%;%?%p4%t;5%;m%?%p9%t\016%e\017%;$<2>";
static char vt100_s_hts         [] = "\033H";
static char vt100_s_ht          [] = "\011";
static char vt100_s_ka1         [] = "\033Oq";
static char vt100_s_ka3         [] = "\033Os";
static char vt100_s_kb2         [] = "\033Or";
static char vt100_s_kc1         [] = "\033Op";
static char vt100_s_kc3         [] = "\033On";
static char vt100_s_acsc        [] = "``aaffggjjkkllmmnnooppqqrrssttuuvvwwxxyyzz{{||}}~~";
static char vt100_s_smam        [] = "\033[?7h";
static char vt100_s_rmam        [] = "\033[?7l";
static char vt100_s_enacs       [] = "\033(B\033)0";
static char vt100_s_kent        [] = "\033OM";
static char vt100_s_el1         [] = "\033[1K$<3>";

static char vt100_bool_data[] = {
	/*   0: bw       */	FALSE,
	/*   1: am       */	TRUE,
	/*   2: xsb      */	FALSE,
	/*   3: xhp      */	FALSE,
	/*   4: xenl     */	TRUE,
	/*   5: eo       */	FALSE,
	/*   6: gn       */	FALSE,
	/*   7: hc       */	FALSE,
	/*   8: km       */	FALSE,
	/*   9: hs       */	FALSE,
	/*  10: in       */	FALSE,
	/*  11: da       */	FALSE,
	/*  12: db       */	FALSE,
	/*  13: mir      */	FALSE,
	/*  14: msgr     */	TRUE,
	/*  15: os       */	FALSE,
	/*  16: eslok    */	FALSE,
	/*  17: xt       */	FALSE,
	/*  18: hz       */	FALSE,
	/*  19: ul       */	FALSE,
	/*  20: xon      */	TRUE,
	/*  21: nxon     */	FALSE,
	/*  22: mc5i     */	TRUE,
	/*  23: chts     */	FALSE,
	/*  24: nrrmc    */	FALSE,
	/*  25: npc      */	FALSE,
	/*  26: ndscr    */	FALSE,
	/*  27: ccc      */	FALSE,
	/*  28: bce      */	FALSE,
	/*  29: hls      */	FALSE,
	/*  30: xhpa     */	FALSE,
	/*  31: crxm     */	FALSE,
	/*  32: daisy    */	FALSE,
	/*  33: xvpa     */	FALSE,
	/*  34: sam      */	FALSE,
	/*  35: cpix     */	FALSE,
	/*  36: lpix     */	FALSE,
	/*  37: OTbs     */	TRUE,
	/*  38: OTns     */	FALSE,
	/*  39: OTnc     */	FALSE,
	/*  40: OTMT     */	FALSE,
	/*  41: OTNL     */	FALSE,
	/*  42: OTpt     */	FALSE,
	/*  43: OTxr     */	FALSE,
};
static NCURSES_INT2 vt100_number_data[] = {
	/*   0: cols     */	80,
	/*   1: it       */	8,
	/*   2: lines    */	24,
	/*   3: lm       */	ABSENT_NUMERIC,
	/*   4: xmc      */	ABSENT_NUMERIC,
	/*   5: pb       */	ABSENT_NUMERIC,
	/*   6: vt       */	3,
	/*   7: wsl      */	ABSENT_NUMERIC,
	/*   8: nlab     */	ABSENT_NUMERIC,
	/*   9: lh       */	ABSENT_NUMERIC,
	/*  10: lw       */	ABSENT_NUMERIC,
	/*  11: ma       */	ABSENT_NUMERIC,
	/*  12: wnum     */	ABSENT_NUMERIC,
	/*  13: colors   */	ABSENT_NUMERIC,
	/*  14: pairs    */	ABSENT_NUMERIC,
	/*  15: ncv      */	ABSENT_NUMERIC,
	/*  16: bufsz    */	ABSENT_NUMERIC,
	/*  17: spinv    */	ABSENT_NUMERIC,
	/*  18: spinh    */	ABSENT_NUMERIC,
	/*  19: maddr    */	ABSENT_NUMERIC,
	/*  20: mjump    */	ABSENT_NUMERIC,
	/*  21: mcs      */	ABSENT_NUMERIC,
	/*  22: mls      */	ABSENT_NUMERIC,
	/*  23: npins    */	ABSENT_NUMERIC,
	/*  24: orc      */	ABSENT_NUMERIC,
	/*  25: orl      */	ABSENT_NUMERIC,
	/*  26: orhi     */	ABSENT_NUMERIC,
	/*  27: orvi     */	ABSENT_NUMERIC,
	/*  28: cps      */	ABSENT_NUMERIC,
	/*  29: widcs    */	ABSENT_NUMERIC,
	/*  30: btns     */	ABSENT_NUMERIC,
	/*  31: bitwin   */	ABSENT_NUMERIC,
	/*  32: bitype   */	ABSENT_NUMERIC,
	/*  33: OTug     */	ABSENT_NUMERIC,
	/*  34: OTdC     */	ABSENT_NUMERIC,
	/*  35: OTdN     */	ABSENT_NUMERIC,
	/*  36: OTdB     */	ABSENT_NUMERIC,
	/*  37: OTdT     */	ABSENT_NUMERIC,
	/*  38: OTkn     */	ABSENT_NUMERIC,
};
static char * vt100_string_data[] = {
	/*   0: cbt      */	ABSENT_STRING,
	/*   1: bel      */	vt100_s_bel,
	/*   2: cr       */	vt100_s_cr,
	/*   3: csr      */	vt100_s_csr,
	/*   4: tbc      */	vt100_s_tbc,
	/*   5: clear    */	vt100_s_clear,
	/*   6: el       */	vt100_s_el,
	/*   7: ed       */	vt100_s_ed,
	/*   8: hpa      */	ABSENT_STRING,
	/*   9: cmdch    */	ABSENT_STRING,
	/*  10: cup      */	vt100_s_cup,
	/*  11: cud1     */	vt100_s_cud1,
	/*  12: home     */	vt100_s_home,
	/*  13: civis    */	ABSENT_STRING,
	/*  14: cub1     */	vt100_s_cub1,
	/*  15: mrcup    */	ABSENT_STRING,
	/*  16: cnorm    */	ABSENT_STRING,
	/*  17: cuf1     */	vt100_s_cuf1,
	/*  18: ll       */	ABSENT_STRING,
	/*  19: cuu1     */	vt100_s_cuu1,
	/*  20: cvvis    */	ABSENT_STRING,
	/*  21: dch1     */	ABSENT_STRING,
	/*  22: dl1      */	ABSENT_STRING,
	/*  23: dsl      */	ABSENT_STRING,
	/*  24: hd       */	ABSENT_STRING,
	/*  25: smacs    */	vt100_s_smacs,
	/*  26: blink    */	vt100_s_blink,
	/*  27: bold     */	vt100_s_bold,
	/*  28: smcup    */	ABSENT_STRING,
	/*  29: smdc     */	ABSENT_STRING,
	/*  30: dim      */	ABSENT_STRING,
	/*  31: smir     */	ABSENT_STRING,
	/*  32: invis    */	ABSENT_STRING,
	/*  33: prot     */	ABSENT_STRING,
	/*  34: rev      */	vt100_s_rev,
	/*  35: smso     */	vt100_s_smso,
	/*  36: smul     */	vt100_s_smul,
	/*  37: ech      */	ABSENT_STRING,
	/*  38: rmacs    */	vt100_s_rmacs,
	/*  39: sgr0     */	vt100_s_sgr0,
	/*  40: rmcup    */	ABSENT_STRING,
	/*  41: rmdc     */	ABSENT_STRING,
	/*  42: rmir     */	ABSENT_STRING,
	/*  43: rmso     */	vt100_s_rmso,
	/*  44: rmul     */	vt100_s_rmul,
	/*  45: flash    */	ABSENT_STRING,
	/*  46: ff       */	ABSENT_STRING,
	/*  47: fsl      */	ABSENT_STRING,
	/*  48: is1      */	ABSENT_STRING,
	/*  49: is2      */	ABSENT_STRING,
	/*  50: is3      */	ABSENT_STRING,
	/*  51: if       */	ABSENT_STRING,
	/*  52: ich1     */	ABSENT_STRING,
	/*  53: il1      */	ABSENT_STRING,
	/*  54: ip       */	ABSENT_STRING,
	/*  55: kbs      */	vt100_s_kbs,
	/*  56: ktbc     */	ABSENT_STRING,
	/*  57: kclr     */	ABSENT_STRING,
	/*  58: kctab    */	ABSENT_STRING,
	/*  59: kdch1    */	ABSENT_STRING,
	/*  60: kdl1     */	ABSENT_STRING,
	/*  61: kcud1    */	vt100_s_kcud1,
	/*  62: krmir    */	ABSENT_STRING,
	/*  63: kel      */	ABSENT_STRING,
	/*  64: ked      */	ABSENT_STRING,
	/*  65: kf0      */	vt100_s_kf0,
	/*  66: kf1      */	vt100_s_kf1,
	/*  67: kf10     */	vt100_s_kf10,
	/*  68: kf2      */	vt100_s_kf2,
	/*  69: kf3      */	vt100_s_kf3,
	/*  70: kf4      */	vt100_s_kf4,
	/*  71: kf5      */	vt100_s_kf5,
	/*  72: kf6      */	vt100_s_kf6,
	/*  73: kf7      */	vt100_s_kf7,
	/*  74: kf8      */	vt100_s_kf8,
	/*  75: kf9      */	vt100_s_kf9,
	/*  76: khome    */	ABSENT_STRING,
	/*  77: kich1    */	ABSENT_STRING,
	/*  78: kil1     */	ABSENT_STRING,
	/*  79: kcub1    */	vt100_s_kcub1,
	/*  80: kll      */	ABSENT_STRING,
	/*  81: knp      */	ABSENT_STRING,
	/*  82: kpp      */	ABSENT_STRING,
	/*  83: kcuf1    */	vt100_s_kcuf1,
	/*  84: kind     */	ABSENT_STRING,
	/*  85: kri      */	ABSENT_STRING,
	/*  86: khts     */	ABSENT_STRING,
	/*  87: kcuu1    */	vt100_s_kcuu1,
	/*  88: rmkx     */	vt100_s_rmkx,
	/*  89: smkx     */	vt100_s_smkx,
	/*  90: lf0      */	ABSENT_STRING,
	/*  91: lf1      */	vt100_s_lf1,
	/*  92: lf10     */	ABSENT_STRING,
	/*  93: lf2      */	vt100_s_lf2,
	/*  94: lf3      */	vt100_s_lf3,
	/*  95: lf4      */	vt100_s_lf4,
	/*  96: lf5      */	ABSENT_STRING,
	/*  97: lf6      */	ABSENT_STRING,
	/*  98: lf7      */	ABSENT_STRING,
	/*  99: lf8      */	ABSENT_STRING,
	/* 100: lf9      */	ABSENT_STRING,
	/* 101: rmm      */	ABSENT_STRING,
	/* 102: smm      */	ABSENT_STRING,
	/* 103: nel      */	ABSENT_STRING,
	/* 104: pad      */	ABSENT_STRING,
	/* 105: dch      */	ABSENT_STRING,
	/* 106: dl       */	ABSENT_STRING,
	/* 107: cud      */	vt100_s_cud,
	/* 108: ich      */	ABSENT_STRING,
	/* 109: indn     */	ABSENT_STRING,
	/* 110: il       */	ABSENT_STRING,
	/* 111: cub      */	vt100_s_cub,
	/* 112: cuf      */	vt100_s_cuf,
	/* 113: rin      */	ABSENT_STRING,
	/* 114: cuu      */	vt100_s_cuu,
	/* 115: pfkey    */	ABSENT_STRING,
	/* 116: pfloc    */	ABSENT_STRING,
	/* 117: pfx      */	ABSENT_STRING,
	/* 118: mc0      */	vt100_s_mc0,
	/* 119: mc4      */	vt100_s_mc4,
	/* 120: mc5      */	vt100_s_mc5,
	/* 121: rep      */	ABSENT_STRING,
	/* 122: rs1      */	ABSENT_STRING,
	/* 123: rs2      */	vt100_s_rs2,
	/* 124: rs3      */	ABSENT_STRING,
	/* 125: rf       */	ABSENT_STRING,
	/* 126: rc       */	vt100_s_rc,
	/* 127: vpa      */	ABSENT_STRING,
	/* 128: sc       */	vt100_s_sc,
	/* 129: ind      */	vt100_s_ind,
	/* 130: ri       */	vt100_s_ri,
	/* 131: sgr      */	vt100_s_sgr,
	/* 132: hts      */	vt100_s_hts,
	/* 133: wind     */	ABSENT_STRING,
	/* 134: ht       */	vt100_s_ht,
	/* 135: tsl      */	ABSENT_STRING,
	/* 136: uc       */	ABSENT_STRING,
	/* 137: hu       */	ABSENT_STRING,
	/* 138: iprog    */	ABSENT_STRING,
	/* 139: ka1      */	vt100_s_ka1,
	/* 140: ka3      */	vt100_s_ka3,
	/* 141: kb2      */	vt100_s_kb2,
	/* 142: kc1      */	vt100_s_kc1,
	/* 143: kc3      */	vt100_s_kc3,
	/* 144: mc5p     */	ABSENT_STRING,
	/* 145: rmp      */	ABSENT_STRING,
	/* 146: acsc     */	vt100_s_acsc,
	/* 147: pln      */	ABSENT_STRING,
	/* 148: kcbt     */	ABSENT_STRING,
	/* 149: smxon    */	ABSENT_STRING,
	/* 150: rmxon    */	ABSENT_STRING,
	/* 151: smam     */	vt100_s_smam,
	/* 152: rmam     */	vt100_s_rmam,
	/* 153: xonc     */	ABSENT_STRING,
	/* 154: xoffc    */	ABSENT_STRING,
	/* 155: enacs    */	vt100_s_enacs,
	/* 156: smln     */	ABSENT_STRING,
	/* 157: rmln     */	ABSENT_STRING,
	/* 158: kbeg     */	ABSENT_STRING,
	/* 159: kcan     */	ABSENT_STRING,
	/* 160: kclo     */	ABSENT_STRING,
	/* 161: kcmd     */	ABSENT_STRING,
	/* 162: kcpy     */	ABSENT_STRING,
	/* 163: kcrt     */	ABSENT_STRING,
	/* 164: kend     */	ABSENT_STRING,
	/* 165: kent     */	vt100_s_kent,
	/* 166: kext     */	ABSENT_STRING,
	/* 167: kfnd     */	ABSENT_STRING,
	/* 168: khlp     */	ABSENT_STRING,
	/* 169: kmrk     */	ABSENT_STRING,
	/* 170: kmsg     */	ABSENT_STRING,
	/* 171: kmov     */	ABSENT_STRING,
	/* 172: knxt     */	ABSENT_STRING,
	/* 173: kopn     */	ABSENT_STRING,
	/* 174: kopt     */	ABSENT_STRING,
	/* 175: kprv     */	ABSENT_STRING,
	/* 176: kprt     */	ABSENT_STRING,
	/* 177: krdo     */	ABSENT_STRING,
	/* 178: kref     */	ABSENT_STRING,
	/* 179: krfr     */	ABSENT_STRING,
	/* 180: krpl     */	ABSENT_STRING,
	/* 181: krst     */	ABSENT_STRING,
	/* 182: kres     */	ABSENT_STRING,
	/* 183: ksav     */	ABSENT_STRING,
	/* 184: kspd     */	ABSENT_STRING,
	/* 185: kund     */	ABSENT_STRING,
	/* 186: kBEG     */	ABSENT_STRING,
	/* 187: kCAN     */	ABSENT_STRING,
	/* 188: kCMD     */	ABSENT_STRING,
	/* 189: kCPY     */	ABSENT_STRING,
	/* 190: kCRT     */	ABSENT_STRING,
	/* 191: kDC      */	ABSENT_STRING,
	/* 192: kDL      */	ABSENT_STRING,
	/* 193: kslt     */	ABSENT_STRING,
	/* 194: kEND     */	ABSENT_STRING,
	/* 195: kEOL     */	ABSENT_STRING,
	/* 196: kEXT     */	ABSENT_STRING,
	/* 197: kFND     */	ABSENT_STRING,
	/* 198: kHLP     */	ABSENT_STRING,
	/* 199: kHOM     */	ABSENT_STRING,
	/* 200: kIC      */	ABSENT_STRING,
	/* 201: kLFT     */	ABSENT_STRING,
	/* 202: kMSG     */	ABSENT_STRING,
	/* 203: kMOV     */	ABSENT_STRING,
	/* 204: kNXT     */	ABSENT_STRING,
	/* 205: kOPT     */	ABSENT_STRING,
	/* 206: kPRV     */	ABSENT_STRING,
	/* 207: kPRT     */	ABSENT_STRING,
	/* 208: kRDO     */	ABSENT_STRING,
	/* 209: kRPL     */	ABSENT_STRING,
	/* 210: kRIT     */	ABSENT_STRING,
	/* 211: kRES     */	ABSENT_STRING,
	/* 212: kSAV     */	ABSENT_STRING,
	/* 213: kSPD     */	ABSENT_STRING,
	/* 214: kUND     */	ABSENT_STRING,
	/* 215: rfi      */	ABSENT_STRING,
	/* 216: kf11     */	ABSENT_STRING,
	/* 217: kf12     */	ABSENT_STRING,
	/* 218: kf13     */	ABSENT_STRING,
	/* 219: kf14     */	ABSENT_STRING,
	/* 220: kf15     */	ABSENT_STRING,
	/* 221: kf16     */	ABSENT_STRING,
	/* 222: kf17     */	ABSENT_STRING,
	/* 223: kf18     */	ABSENT_STRING,
	/* 224: kf19     */	ABSENT_STRING,
	/* 225: kf20     */	ABSENT_STRING,
	/* 226: kf21     */	ABSENT_STRING,
	/* 227: kf22     */	ABSENT_STRING,
	/* 228: kf23     */	ABSENT_STRING,
	/* 229: kf24     */	ABSENT_STRING,
	/* 230: kf25     */	ABSENT_STRING,
	/* 231: kf26     */	ABSENT_STRING,
	/* 232: kf27     */	ABSENT_STRING,
	/* 233: kf28     */	ABSENT_STRING,
	/* 234: kf29     */	ABSENT_STRING,
	/* 235: kf30     */	ABSENT_STRING,
	/* 236: kf31     */	ABSENT_STRING,
	/* 237: kf32     */	ABSENT_STRING,
	/* 238: kf33     */	ABSENT_STRING,
	/* 239: kf34     */	ABSENT_STRING,
	/* 240: kf35     */	ABSENT_STRING,
	/* 241: kf36     */	ABSENT_STRING,
	/* 242: kf37     */	ABSENT_STRING,
	/* 243: kf38     */	ABSENT_STRING,
	/* 244: kf39     */	ABSENT_STRING,
	/* 245: kf40     */	ABSENT_STRING,
	/* 246: kf41     */	ABSENT_STRING,
	/* 247: kf42     */	ABSENT_STRING,
	/* 248: kf43     */	ABSENT_STRING,
	/* 249: kf44     */	ABSENT_STRING,
	/* 250: kf45     */	ABSENT_STRING,
	/* 251: kf46     */	ABSENT_STRING,
	/* 252: kf47     */	ABSENT_STRING,
	/* 253: kf48     */	ABSENT_STRING,
	/* 254: kf49     */	ABSENT_STRING,
	/* 255: kf50     */	ABSENT_STRING,
	/* 256: kf51     */	ABSENT_STRING,
	/* 257: kf52     */	ABSENT_STRING,
	/* 258: kf53     */	ABSENT_STRING,
	/* 259: kf54     */	ABSENT_STRING,
	/* 260: kf55     */	ABSENT_STRING,
	/* 261: kf56     */	ABSENT_STRING,
	/* 262: kf57     */	ABSENT_STRING,
	/* 263: kf58     */	ABSENT_STRING,
	/* 264: kf59     */	ABSENT_STRING,
	/* 265: kf60     */	ABSENT_STRING,
	/* 266: kf61     */	ABSENT_STRING,
	/* 267: kf62     */	ABSENT_STRING,
	/* 268: kf63     */	ABSENT_STRING,
	/* 269: el1      */	vt100_s_el1,
	/* 270: mgc      */	ABSENT_STRING,
	/* 271: smgl     */	ABSENT_STRING,
	/* 272: smgr     */	ABSENT_STRING,
	/* 273: fln      */	ABSENT_STRING,
	/* 274: sclk     */	ABSENT_STRING,
	/* 275: dclk     */	ABSENT_STRING,
	/* 276: rmclk    */	ABSENT_STRING,
	/* 277: cwin     */	ABSENT_STRING,
	/* 278: wingo    */	ABSENT_STRING,
	/* 279: hup      */	ABSENT_STRING,
	/* 280: dial     */	ABSENT_STRING,
	/* 281: qdial    */	ABSENT_STRING,
	/* 282: tone     */	ABSENT_STRING,
	/* 283: pulse    */	ABSENT_STRING,
	/* 284: hook     */	ABSENT_STRING,
	/* 285: pause    */	ABSENT_STRING,
	/* 286: wait     */	ABSENT_STRING,
	/* 287: u0       */	ABSENT_STRING,
	/* 288: u1       */	ABSENT_STRING,
	/* 289: u2       */	ABSENT_STRING,
	/* 290: u3       */	ABSENT_STRING,
	/* 291: u4       */	ABSENT_STRING,
	/* 292: u5       */	ABSENT_STRING,
	/* 293: u6       */	ABSENT_STRING,
	/* 294: u7       */	ABSENT_STRING,
	/* 295: u8       */	ABSENT_STRING,
	/* 296: u9       */	ABSENT_STRING,
	/* 297: op       */	ABSENT_STRING,
	/* 298: oc       */	ABSENT_STRING,
	/* 299: initc    */	ABSENT_STRING,
	/* 300: initp    */	ABSENT_STRING,
	/* 301: scp      */	ABSENT_STRING,
	/* 302: setf     */	ABSENT_STRING,
	/* 303: setb     */	ABSENT_STRING,
	/* 304: cpi      */	ABSENT_STRING,
	/* 305: lpi      */	ABSENT_STRING,
	/* 306: chr      */	ABSENT_STRING,
	/* 307: cvr      */	ABSENT_STRING,
	/* 308: defc     */	ABSENT_STRING,
	/* 309: swidm    */	ABSENT_STRING,
	/* 310: sdrfq    */	ABSENT_STRING,
	/* 311: sitm     */	ABSENT_STRING,
	/* 312: slm      */	ABSENT_STRING,
	/* 313: smicm    */	ABSENT_STRING,
	/* 314: snlq     */	ABSENT_STRING,
	/* 315: snrmq    */	ABSENT_STRING,
	/* 316: sshm     */	ABSENT_STRING,
	/* 317: ssubm    */	ABSENT_STRING,
	/* 318: ssupm    */	ABSENT_STRING,
	/* 319: sum      */	ABSENT_STRING,
	/* 320: rwidm    */	ABSENT_STRING,
	/* 321: ritm     */	ABSENT_STRING,
	/* 322: rlm      */	ABSENT_STRING,
	/* 323: rmicm    */	ABSENT_STRING,
	/* 324: rshm     */	ABSENT_STRING,
	/* 325: rsubm    */	ABSENT_STRING,
	/* 326: rsupm    */	ABSENT_STRING,
	/* 327: rum      */	ABSENT_STRING,
	/* 328: mhpa     */	ABSENT_STRING,
	/* 329: mcud1    */	ABSENT_STRING,
	/* 330: mcub1    */	ABSENT_STRING,
	/* 331: mcuf1    */	ABSENT_STRING,
	/* 332: mvpa     */	ABSENT_STRING,
	/* 333: mcuu1    */	ABSENT_STRING,
	/* 334: porder   */	ABSENT_STRING,
	/* 335: mcud     */	ABSENT_STRING,
	/* 336: mcub     */	ABSENT_STRING,
	/* 337: mcuf     */	ABSENT_STRING,
	/* 338: mcuu     */	ABSENT_STRING,
	/* 339: scs      */	ABSENT_STRING,
	/* 340: smgb     */	ABSENT_STRING,
	/* 341: smgbp    */	ABSENT_STRING,
	/* 342: smglp    */	ABSENT_STRING,
	/* 343: smgrp    */	ABSENT_STRING,
	/* 344: smgt     */	ABSENT_STRING,
	/* 345: smgtp    */	ABSENT_STRING,
	/* 346: sbim     */	ABSENT_STRING,
	/* 347: scsd     */	ABSENT_STRING,
	/* 348: rbim     */	ABSENT_STRING,
	/* 349: rcsd     */	ABSENT_STRING,
	/* 350: subcs    */	ABSENT_STRING,
	/* 351: supcs    */	ABSENT_STRING,
	/* 352: docr     */	ABSENT_STRING,
	/* 353: zerom    */	ABSENT_STRING,
	/* 354: csnm     */	ABSENT_STRING,
	/* 355: kmous    */	ABSENT_STRING,
	/* 356: minfo    */	ABSENT_STRING,
	/* 357: reqmp    */	ABSENT_STRING,
	/* 358: getm     */	ABSENT_STRING,
	/* 359: setaf    */	ABSENT_STRING,
	/* 360: setab    */	ABSENT_STRING,
	/* 361: pfxl     */	ABSENT_STRING,
	/* 362: devt     */	ABSENT_STRING,
	/* 363: csin     */	ABSENT_STRING,
	/* 364: s0ds     */	ABSENT_STRING,
	/* 365: s1ds     */	ABSENT_STRING,
	/* 366: s2ds     */	ABSENT_STRING,
	/* 367: s3ds     */	ABSENT_STRING,
	/* 368: smglr    */	ABSENT_STRING,
	/* 369: smgtb    */	ABSENT_STRING,
	/* 370: birep    */	ABSENT_STRING,
	/* 371: binel    */	ABSENT_STRING,
	/* 372: bicr     */	ABSENT_STRING,
	/* 373: colornm  */	ABSENT_STRING,
	/* 374: defbi    */	ABSENT_STRING,
	/* 375: endbi    */	ABSENT_STRING,
	/* 376: setcolor */	ABSENT_STRING,
	/* 377: slines   */	ABSENT_STRING,
	/* 378: dispc    */	ABSENT_STRING,
	/* 379: smpch    */	ABSENT_STRING,
	/* 380: rmpch    */	ABSENT_STRING,
	/* 381: smsc     */	ABSENT_STRING,
	/* 382: rmsc     */	ABSENT_STRING,
	/* 383: pctrm    */	ABSENT_STRING,
	/* 384: scesc    */	ABSENT_STRING,
	/* 385: scesa    */	ABSENT_STRING,
	/* 386: ehhlm    */	ABSENT_STRING,
	/* 387: elhlm    */	ABSENT_STRING,
	/* 388: elohlm   */	ABSENT_STRING,
	/* 389: erhlm    */	ABSENT_STRING,
	/* 390: ethlm    */	ABSENT_STRING,
	/* 391: evhlm    */	ABSENT_STRING,
	/* 392: sgr1     */	ABSENT_STRING,
	/* 393: slength  */	ABSENT_STRING,
	/* 394: OTi2     */	ABSENT_STRING,
	/* 395: OTrs     */	ABSENT_STRING,
	/* 396: OTnl     */	ABSENT_STRING,
	/* 397: OTbc     */	ABSENT_STRING,
	/* 398: OTko     */	ABSENT_STRING,
	/* 399: OTma     */	ABSENT_STRING,
	/* 400: OTG2     */	ABSENT_STRING,
	/* 401: OTG3     */	ABSENT_STRING,
	/* 402: OTG1     */	ABSENT_STRING,
	/* 403: OTG4     */	ABSENT_STRING,
	/* 404: OTGR     */	ABSENT_STRING,
	/* 405: OTGL     */	ABSENT_STRING,
	/* 406: OTGU     */	ABSENT_STRING,
	/* 407: OTGD     */	ABSENT_STRING,
	/* 408: OTGH     */	ABSENT_STRING,
	/* 409: OTGV     */	ABSENT_STRING,
	/* 410: OTGC     */	ABSENT_STRING,
	/* 411: meml     */	ABSENT_STRING,
	/* 412: memu     */	ABSENT_STRING,
	/* 413: box1     */	ABSENT_STRING,
};
/* xterm */

static char xterm_alias_data[] = "xterm|xterm terminal emulator (X Window System)";

static char xterm_s_cbt         [] = "\033[Z";
static char xterm_s_bel         [] = "\007";
static char xterm_s_cr          [] = "\015";
static char xterm_s_csr         [] = "\033[%i%p1%d;%p2%dr";
static char xterm_s_tbc         [] = "\033[3g";
static char xterm_s_clear       [] = "\033[H\033[2J";
static char xterm_s_el          [] = "\033[K";
static char xterm_s_ed          [] = "\033[J";
static char xterm_s_hpa         [] = "\033[%i%p1%dG";
static char xterm_s_cup         [] = "\033[%i%p1%d;%p2%dH";
static char xterm_s_cud1        [] = "\012";
static char xterm_s_home        [] = "\033[H";
static char xterm_s_civis       [] = "\033[?25l";
static char xterm_s_cub1        [] = "\010";
static char xterm_s_cnorm       [] = "\033[?12l\033[?25h";
static char xterm_s_cuf1        [] = "\033[C";
static char xterm_s_cuu1        [] = "\033[A";
static char xterm_s_cvvis       [] = "\033[?12;25h";
static char xterm_s_dch1        [] = "\033[P";
static char xterm_s_dl1         [] = "\033[M";
static char xterm_s_smacs       [] = "\033(0";
static char xterm_s_blink       [] = "\033[5m";
static char xterm_s_bold        [] = "\033[1m";
static char xterm_s_smcup       [] = "\033[?1049h\033[22;0;0t";
static char xterm_s_dim         [] = "\033[2m";
static char xterm_s_smir        [] = "\033[4h";
static char xterm_s_invis       [] = "\033[8m";
static char xterm_s_rev         [] = "\033[7m";
static char xterm_s_smso        [] = "\033[7m";
static char xterm_s_smul        [] = "\033[4m";
static char xterm_s_ech         [] = "\033[%p1%dX";
static char xterm_s_rmacs       [] = "\033(B";
static char xterm_s_sgr0        [] = "\033(B\033[m";
static char xterm_s_rmcup       [] = "\033[?1049l\033[23;0;0t";
static char xterm_s_rmir        [] = "\033[4l";
static char xterm_s_rmso        [] = "\033[27m";
static char xterm_s_rmul        [] = "\033[24m";
static char xterm_s_flash       [] = "\033[?5h$<100/>\033[?5l";
static char xterm_s_is2         [] = "\033[!p\033[?3;4l\033[4l\033>";
static char xterm_s_il1         [] = "\033[L";
static char xterm_s_kbs         [] = "\010";
static char xterm_s_kdch1       [] = "\033[3~";
static char xterm_s_kcud1       [] = "\033OB";
static char xterm_s_kf1         [] = "\033OP";
static char xterm_s_kf10        [] = "\033[21~";
static char xterm_s_kf2         [] = "\033OQ";
static char xterm_s_kf3         [] = "\033OR";
static char xterm_s_kf4         [] = "\033OS";
static char xterm_s_kf5         [] = "\033[15~";
static char xterm_s_kf6         [] = "\033[17~";
static char xterm_s_kf7         [] = "\033[18~";
static char xterm_s_kf8         [] = "\033[19~";
static char xterm_s_kf9         [] = "\033[20~";
static char xterm_s_khome       [] = "\033OH";
static char xterm_s_kich1       [] = "\033[2~";
static char xterm_s_kcub1       [] = "\033OD";
static char xterm_s_knp         [] = "\033[6~";
static char xterm_s_kpp         [] = "\033[5~";
static char xterm_s_kcuf1       [] = "\033OC";
static char xterm_s_kind        [] = "\033[1;2B";
static char xterm_s_kri         [] = "\033[1;2A";
static char xterm_s_kcuu1       [] = "\033OA";
static char xterm_s_rmkx        [] = "\033[?1l\033>";
static char xterm_s_smkx        [] = "\033[?1h\033=";
static char xterm_s_rmm         [] = "\033[?1034l";
static char xterm_s_smm         [] = "\033[?1034h";
static char xterm_s_dch         [] = "\033[%p1%dP";
static char xterm_s_dl          [] = "\033[%p1%dM";
static char xterm_s_cud         [] = "\033[%p1%dB";
static char xterm_s_ich         [] = "\033[%p1%d@";
static char xterm_s_indn        [] = "\033[%p1%dS";
static char xterm_s_il          [] = "\033[%p1%dL";
static char xterm_s_cub         [] = "\033[%p1%dD";
static char xterm_s_cuf         [] = "\033[%p1%dC";
static char xterm_s_rin         [] = "\033[%p1%dT";
static char xterm_s_cuu         [] = "\033[%p1%dA";
static char xterm_s_mc0         [] = "\033[i";
static char xterm_s_mc4         [] = "\033[4i";
static char xterm_s_mc5         [] = "\033[5i";
static char xterm_s_rep         [] = "%p1%c\033[%p2%{1}%-%db";
static char xterm_s_rs1         [] = "\033c";
static char xterm_s_rs2         [] = "\033[!p\033[?3;4l\033[4l\033>";
static char xterm_s_rc          [] = "\0338";
static char xterm_s_vpa         [] = "\033[%i%p1%dd";
static char xterm_s_sc          [] = "\0337";
static char xterm_s_ind         [] = "\012";
static char xterm_s_ri          [] = "\033M";
static char xterm_s_sgr         [] = "%?%p9%t\033(0%e\033(B%;\033[0%?%p6%t;1%;%?%p5%t;2%;%?%p2%t;4%;%?%p1%p3%|%t;7%;%?%p4%t;5%;%?%p7%t;8%;m";
static char xterm_s_hts         [] = "\033H";
static char xterm_s_ht          [] = "\011";
static char xterm_s_kb2         [] = "\033OE";
static char xterm_s_acsc        [] = "``aaffggiijjkkllmmnnooppqqrrssttuuvvwwxxyyzz{{||}}~~";
static char xterm_s_kcbt        [] = "\033[Z";
static char xterm_s_smam        [] = "\033[?7h";
static char xterm_s_rmam        [] = "\033[?7l";
static char xterm_s_kend        [] = "\033OF";
static char xterm_s_kent        [] = "\033OM";
static char xterm_s_kDC         [] = "\033[3;2~";
static char xterm_s_kEND        [] = "\033[1;2F";
static char xterm_s_kHOM        [] = "\033[1;2H";
static char xterm_s_kIC         [] = "\033[2;2~";
static char xterm_s_kLFT        [] = "\033[1;2D";
static char xterm_s_kNXT        [] = "\033[6;2~";
static char xterm_s_kPRV        [] = "\033[5;2~";
static char xterm_s_kRIT        [] = "\033[1;2C";
static char xterm_s_kf11        [] = "\033[23~";
static char xterm_s_kf12        [] = "\033[24~";
static char xterm_s_kf13        [] = "\033[1;2P";
static char xterm_s_kf14        [] = "\033[1;2Q";
static char xterm_s_kf15        [] = "\033[1;2R";
static char xterm_s_kf16        [] = "\033[1;2S";
static char xterm_s_kf17        [] = "\033[15;2~";
static char xterm_s_kf18        [] = "\033[17;2~";
static char xterm_s_kf19        [] = "\033[18;2~";
static char xterm_s_kf20        [] = "\033[19;2~";
static char xterm_s_kf21        [] = "\033[20;2~";
static char xterm_s_kf22        [] = "\033[21;2~";
static char xterm_s_kf23        [] = "\033[23;2~";
static char xterm_s_kf24        [] = "\033[24;2~";
static char xterm_s_kf25        [] = "\033[1;5P";
static char xterm_s_kf26        [] = "\033[1;5Q";
static char xterm_s_kf27        [] = "\033[1;5R";
static char xterm_s_kf28        [] = "\033[1;5S";
static char xterm_s_kf29        [] = "\033[15;5~";
static char xterm_s_kf30        [] = "\033[17;5~";
static char xterm_s_kf31        [] = "\033[18;5~";
static char xterm_s_kf32        [] = "\033[19;5~";
static char xterm_s_kf33        [] = "\033[20;5~";
static char xterm_s_kf34        [] = "\033[21;5~";
static char xterm_s_kf35        [] = "\033[23;5~";
static char xterm_s_kf36        [] = "\033[24;5~";
static char xterm_s_kf37        [] = "\033[1;6P";
static char xterm_s_kf38        [] = "\033[1;6Q";
static char xterm_s_kf39        [] = "\033[1;6R";
static char xterm_s_kf40        [] = "\033[1;6S";
static char xterm_s_kf41        [] = "\033[15;6~";
static char xterm_s_kf42        [] = "\033[17;6~";
static char xterm_s_kf43        [] = "\033[18;6~";
static char xterm_s_kf44        [] = "\033[19;6~";
static char xterm_s_kf45        [] = "\033[20;6~";
static char xterm_s_kf46        [] = "\033[21;6~";
static char xterm_s_kf47        [] = "\033[23;6~";
static char xterm_s_kf48        [] = "\033[24;6~";
static char xterm_s_kf49        [] = "\033[1;3P";
static char xterm_s_kf50        [] = "\033[1;3Q";
static char xterm_s_kf51        [] = "\033[1;3R";
static char xterm_s_kf52        [] = "\033[1;3S";
static char xterm_s_kf53        [] = "\033[15;3~";
static char xterm_s_kf54        [] = "\033[17;3~";
static char xterm_s_kf55        [] = "\033[18;3~";
static char xterm_s_kf56        [] = "\033[19;3~";
static char xterm_s_kf57        [] = "\033[20;3~";
static char xterm_s_kf58        [] = "\033[21;3~";
static char xterm_s_kf59        [] = "\033[23;3~";
static char xterm_s_kf60        [] = "\033[24;3~";
static char xterm_s_kf61        [] = "\033[1;4P";
static char xterm_s_kf62        [] = "\033[1;4Q";
static char xterm_s_kf63        [] = "\033[1;4R";
static char xterm_s_el1         [] = "\033[1K";
static char xterm_s_u6          [] = "\033[%i%d;%dR";
static char xterm_s_u7          [] = "\033[6n";
static char xterm_s_u8          [] = "\033[?%[;0123456789]c";
static char xterm_s_u9          [] = "\033[c";
static char xterm_s_op          [] = "\033[39;49m";
static char xterm_s_setf        [] = "\033[3%?%p1%{1}%=%t4%e%p1%{3}%=%t6%e%p1%{4}%=%t1%e%p1%{6}%=%t3%e%p1%d%;m";
static char xterm_s_setb        [] = "\033[4%?%p1%{1}%=%t4%e%p1%{3}%=%t6%e%p1%{4}%=%t1%e%p1%{6}%=%t3%e%p1%d%;m";
static char xterm_s_sitm        [] = "\033[3m";
static char xterm_s_ritm        [] = "\033[23m";
static char xterm_s_kmous       [] = "\033[<";
static char xterm_s_setaf       [] = "\033[3%p1%dm";
static char xterm_s_setab       [] = "\033[4%p1%dm";
static char xterm_s_meml        [] = "\033l";
static char xterm_s_memu        [] = "\033m";

static char xterm_bool_data[] = {
	/*   0: bw       */	FALSE,
	/*   1: am       */	TRUE,
	/*   2: xsb      */	FALSE,
	/*   3: xhp      */	FALSE,
	/*   4: xenl     */	TRUE,
	/*   5: eo       */	FALSE,
	/*   6: gn       */	FALSE,
	/*   7: hc       */	FALSE,
	/*   8: km       */	TRUE,
	/*   9: hs       */	FALSE,
	/*  10: in       */	FALSE,
	/*  11: da       */	FALSE,
	/*  12: db       */	FALSE,
	/*  13: mir      */	TRUE,
	/*  14: msgr     */	TRUE,
	/*  15: os       */	FALSE,
	/*  16: eslok    */	FALSE,
	/*  17: xt       */	FALSE,
	/*  18: hz       */	FALSE,
	/*  19: ul       */	FALSE,
	/*  20: xon      */	FALSE,
	/*  21: nxon     */	FALSE,
	/*  22: mc5i     */	TRUE,
	/*  23: chts     */	FALSE,
	/*  24: nrrmc    */	FALSE,
	/*  25: npc      */	TRUE,
	/*  26: ndscr    */	FALSE,
	/*  27: ccc      */	FALSE,
	/*  28: bce      */	TRUE,
	/*  29: hls      */	FALSE,
	/*  30: xhpa     */	FALSE,
	/*  31: crxm     */	FALSE,
	/*  32: daisy    */	FALSE,
	/*  33: xvpa     */	FALSE,
	/*  34: sam      */	FALSE,
	/*  35: cpix     */	FALSE,
	/*  36: lpix     */	FALSE,
	/*  37: OTbs     */	TRUE,
	/*  38: OTns     */	FALSE,
	/*  39: OTnc     */	FALSE,
	/*  40: OTMT     */	FALSE,
	/*  41: OTNL     */	FALSE,
	/*  42: OTpt     */	FALSE,
	/*  43: OTxr     */	FALSE,
};
static NCURSES_INT2 xterm_number_data[] = {
	/*   0: cols     */	80,
	/*   1: it       */	8,
	/*   2: lines    */	24,
	/*   3: lm       */	ABSENT_NUMERIC,
	/*   4: xmc      */	ABSENT_NUMERIC,
	/*   5: pb       */	ABSENT_NUMERIC,
	/*   6: vt       */	ABSENT_NUMERIC,
	/*   7: wsl      */	ABSENT_NUMERIC,
	/*   8: nlab     */	ABSENT_NUMERIC,
	/*   9: lh       */	ABSENT_NUMERIC,
	/*  10: lw       */	ABSENT_NUMERIC,
	/*  11: ma       */	ABSENT_NUMERIC,
	/*  12: wnum     */	ABSENT_NUMERIC,
	/*  13: colors   */	8,
	/*  14: pairs    */	64,
	/*  15: ncv      */	ABSENT_NUMERIC,
	/*  16: bufsz    */	ABSENT_NUMERIC,
	/*  17: spinv    */	ABSENT_NUMERIC,
	/*  18: spinh    */	ABSENT_NUMERIC,
	/*  19: maddr    */	ABSENT_NUMERIC,
	/*  20: mjump    */	ABSENT_NUMERIC,
	/*  21: mcs      */	ABSENT_NUMERIC,
	/*  22: mls      */	ABSENT_NUMERIC,
	/*  23: npins    */	ABSENT_NUMERIC,
	/*  24: orc      */	ABSENT_NUMERIC,
	/*  25: orl      */	ABSENT_NUMERIC,
	/*  26: orhi     */	ABSENT_NUMERIC,
	/*  27: orvi     */	ABSENT_NUMERIC,
	/*  28: cps      */	ABSENT_NUMERIC,
	/*  29: widcs    */	ABSENT_NUMERIC,
	/*  30: btns     */	ABSENT_NUMERIC,
	/*  31: bitwin   */	ABSENT_NUMERIC,
	/*  32: bitype   */	ABSENT_NUMERIC,
	/*  33: OTug     */	ABSENT_NUMERIC,
	/*  34: OTdC     */	ABSENT_NUMERIC,
	/*  35: OTdN     */	ABSENT_NUMERIC,
	/*  36: OTdB     */	ABSENT_NUMERIC,
	/*  37: OTdT     */	ABSENT_NUMERIC,
	/*  38: OTkn     */	ABSENT_NUMERIC,
};
static char * xterm_string_data[] = {
	/*   0: cbt      */	xterm_s_cbt,
	/*   1: bel      */	xterm_s_bel,
	/*   2: cr       */	xterm_s_cr,
	/*   3: csr      */	xterm_s_csr,
	/*   4: tbc      */	xterm_s_tbc,
	/*   5: clear    */	xterm_s_clear,
	/*   6: el       */	xterm_s_el,
	/*   7: ed       */	xterm_s_ed,
	/*   8: hpa      */	xterm_s_hpa,
	/*   9: cmdch    */	ABSENT_STRING,
	/*  10: cup      */	xterm_s_cup,
	/*  11: cud1     */	xterm_s_cud1,
	/*  12: home     */	xterm_s_home,
	/*  13: civis    */	xterm_s_civis,
	/*  14: cub1     */	xterm_s_cub1,
	/*  15: mrcup    */	ABSENT_STRING,
	/*  16: cnorm    */	xterm_s_cnorm,
	/*  17: cuf1     */	xterm_s_cuf1,
	/*  18: ll       */	ABSENT_STRING,
	/*  19: cuu1     */	xterm_s_cuu1,
	/*  20: cvvis    */	xterm_s_cvvis,
	/*  21: dch1     */	xterm_s_dch1,
	/*  22: dl1      */	xterm_s_dl1,
	/*  23: dsl      */	ABSENT_STRING,
	/*  24: hd       */	ABSENT_STRING,
	/*  25: smacs    */	xterm_s_smacs,
	/*  26: blink    */	xterm_s_blink,
	/*  27: bold     */	xterm_s_bold,
	/*  28: smcup    */	xterm_s_smcup,
	/*  29: smdc     */	ABSENT_STRING,
	/*  30: dim      */	xterm_s_dim,
	/*  31: smir     */	xterm_s_smir,
	/*  32: invis    */	xterm_s_invis,
	/*  33: prot     */	ABSENT_STRING,
	/*  34: rev      */	xterm_s_rev,
	/*  35: smso     */	xterm_s_smso,
	/*  36: smul     */	xterm_s_smul,
	/*  37: ech      */	xterm_s_ech,
	/*  38: rmacs    */	xterm_s_rmacs,
	/*  39: sgr0     */	xterm_s_sgr0,
	/*  40: rmcup    */	xterm_s_rmcup,
	/*  41: rmdc     */	ABSENT_STRING,
	/*  42: rmir     */	xterm_s_rmir,
	/*  43: rmso     */	xterm_s_rmso,
	/*  44: rmul     */	xterm_s_rmul,
	/*  45: flash    */	xterm_s_flash,
	/*  46: ff       */	ABSENT_STRING,
	/*  47: fsl      */	ABSENT_STRING,
	/*  48: is1      */	ABSENT_STRING,
	/*  49: is2      */	xterm_s_is2,
	/*  50: is3      */	ABSENT_STRING,
	/*  51: if       */	ABSENT_STRING,
	/*  52: ich1     */	ABSENT_STRING,
	/*  53: il1      */	xterm_s_il1,
	/*  54: ip       */	ABSENT_STRING,
	/*  55: kbs      */	xterm_s_kbs,
	/*  56: ktbc     */	ABSENT_STRING,
	/*  57: kclr     */	ABSENT_STRING,
	/*  58: kctab    */	ABSENT_STRING,
	/*  59: kdch1    */	xterm_s_kdch1,
	/*  60: kdl1     */	ABSENT_STRING,
	/*  61: kcud1    */	xterm_s_kcud1,
	/*  62: krmir    */	ABSENT_STRING,
	/*  63: kel      */	ABSENT_STRING,
	/*  64: ked      */	ABSENT_STRING,
	/*  65: kf0      */	ABSENT_STRING,
	/*  66: kf1      */	xterm_s_kf1,
	/*  67: kf10     */	xterm_s_kf10,
	/*  68: kf2      */	xterm_s_kf2,
	/*  69: kf3      */	xterm_s_kf3,
	/*  70: kf4      */	xterm_s_kf4,
	/*  71: kf5      */	xterm_s_kf5,
	/*  72: kf6      */	xterm_s_kf6,
	/*  73: kf7      */	xterm_s_kf7,
	/*  74: kf8      */	xterm_s_kf8,
	/*  75: kf9      */	xterm_s_kf9,
	/*  76: khome    */	xterm_s_khome,
	/*  77: kich1    */	xterm_s_kich1,
	/*  78: kil1     */	ABSENT_STRING,
	/*  79: kcub1    */	xterm_s_kcub1,
	/*  80: kll      */	ABSENT_STRING,
	/*  81: knp      */	xterm_s_knp,
	/*  82: kpp      */	xterm_s_kpp,
	/*  83: kcuf1    */	xterm_s_kcuf1,
	/*  84: kind     */	xterm_s_kind,
	/*  85: kri      */	xterm_s_kri,
	/*  86: khts     */	ABSENT_STRING,
	/*  87: kcuu1    */	xterm_s_kcuu1,
	/*  88: rmkx     */	xterm_s_rmkx,
	/*  89: smkx     */	xterm_s_smkx,
	/*  90: lf0      */	ABSENT_STRING,
	/*  91: lf1      */	ABSENT_STRING,
	/*  92: lf10     */	ABSENT_STRING,
	/*  93: lf2      */	ABSENT_STRING,
	/*  94: lf3      */	ABSENT_STRING,
	/*  95: lf4      */	ABSENT_STRING,
	/*  96: lf5      */	ABSENT_STRING,
	/*  97: lf6      */	ABSENT_STRING,
	/*  98: lf7      */	ABSENT_STRING,
	/*  99: lf8      */	ABSENT_STRING,
	/* 100: lf9      */	ABSENT_STRING,
	/* 101: rmm      */	xterm_s_rmm,
	/* 102: smm      */	xterm_s_smm,
	/* 103: nel      */	ABSENT_STRING,
	/* 104: pad      */	ABSENT_STRING,
	/* 105: dch      */	xterm_s_dch,
	/* 106: dl       */	xterm_s_dl,
	/* 107: cud      */	xterm_s_cud,
	/* 108: ich      */	xterm_s_ich,
	/* 109: indn     */	xterm_s_indn,
	/* 110: il       */	xterm_s_il,
	/* 111: cub      */	xterm_s_cub,
	/* 112: cuf      */	xterm_s_cuf,
	/* 113: rin      */	xterm_s_rin,
	/* 114: cuu      */	xterm_s_cuu,
	/* 115: pfkey    */	ABSENT_STRING,
	/* 116: pfloc    */	ABSENT_STRING,
	/* 117: pfx      */	ABSENT_STRING,
	/* 118: mc0      */	xterm_s_mc0,
	/* 119: mc4      */	xterm_s_mc4,
	/* 120: mc5      */	xterm_s_mc5,
	/* 121: rep      */	xterm_s_rep,
	/* 122: rs1      */	xterm_s_rs1,
	/* 123: rs2      */	xterm_s_rs2,
	/* 124: rs3      */	ABSENT_STRING,
	/* 125: rf       */	ABSENT_STRING,
	/* 126: rc       */	xterm_s_rc,
	/* 127: vpa      */	xterm_s_vpa,
	/* 128: sc       */	xterm_s_sc,
	/* 129: ind      */	xterm_s_ind,
	/* 130: ri       */	xterm_s_ri,
	/* 131: sgr      */	xterm_s_sgr,
	/* 132: hts      */	xterm_s_hts,
	/* 133: wind     */	ABSENT_STRING,
	/* 134: ht       */	xterm_s_ht,
	/* 135: tsl      */	ABSENT_STRING,
	/* 136: uc       */	ABSENT_STRING,
	/* 137: hu       */	ABSENT_STRING,
	/* 138: iprog    */	ABSENT_STRING,
	/* 139: ka1      */	ABSENT_STRING,
	/* 140: ka3      */	ABSENT_STRING,
	/* 141: kb2      */	xterm_s_kb2,
	/* 142: kc1      */	ABSENT_STRING,
	/* 143: kc3      */	ABSENT_STRING,
	/* 144: mc5p     */	ABSENT_STRING,
	/* 145: rmp      */	ABSENT_STRING,
	/* 146: acsc     */	xterm_s_acsc,
	/* 147: pln      */	ABSENT_STRING,
	/* 148: kcbt     */	xterm_s_kcbt,
	/* 149: smxon    */	ABSENT_STRING,
	/* 150: rmxon    */	ABSENT_STRING,
	/* 151: smam     */	xterm_s_smam,
	/* 152: rmam     */	xterm_s_rmam,
	/* 153: xonc     */	ABSENT_STRING,
	/* 154: xoffc    */	ABSENT_STRING,
	/* 155: enacs    */	ABSENT_STRING,
	/* 156: smln     */	ABSENT_STRING,
	/* 157: rmln     */	ABSENT_STRING,
	/* 158: kbeg     */	ABSENT_STRING,
	/* 159: kcan     */	ABSENT_STRING,
	/* 160: kclo     */	ABSENT_STRING,
	/* 161: kcmd     */	ABSENT_STRING,
	/* 162: kcpy     */	ABSENT_STRING,
	/* 163: kcrt     */	ABSENT_STRING,
	/* 164: kend     */	xterm_s_kend,
	/* 165: kent     */	xterm_s_kent,
	/* 166: kext     */	ABSENT_STRING,
	/* 167: kfnd     */	ABSENT_STRING,
	/* 168: khlp     */	ABSENT_STRING,
	/* 169: kmrk     */	ABSENT_STRING,
	/* 170: kmsg     */	ABSENT_STRING,
	/* 171: kmov     */	ABSENT_STRING,
	/* 172: knxt     */	ABSENT_STRING,
	/* 173: kopn     */	ABSENT_STRING,
	/* 174: kopt     */	ABSENT_STRING,
	/* 175: kprv     */	ABSENT_STRING,
	/* 176: kprt     */	ABSENT_STRING,
	/* 177: krdo     */	ABSENT_STRING,
	/* 178: kref     */	ABSENT_STRING,
	/* 179: krfr     */	ABSENT_STRING,
	/* 180: krpl     */	ABSENT_STRING,
	/* 181: krst     */	ABSENT_STRING,
	/* 182: kres     */	ABSENT_STRING,
	/* 183: ksav     */	ABSENT_STRING,
	/* 184: kspd     */	ABSENT_STRING,
	/* 185: kund     */	ABSENT_STRING,
	/* 186: kBEG     */	ABSENT_STRING,
	/* 187: kCAN     */	ABSENT_STRING,
	/* 188: kCMD     */	ABSENT_STRING,
	/* 189: kCPY     */	ABSENT_STRING,
	/* 190: kCRT     */	ABSENT_STRING,
	/* 191: kDC      */	xterm_s_kDC,
	/* 192: kDL      */	ABSENT_STRING,
	/* 193: kslt     */	ABSENT_STRING,
	/* 194: kEND     */	xterm_s_kEND,
	/* 195: kEOL     */	ABSENT_STRING,
	/* 196: kEXT     */	ABSENT_STRING,
	/* 197: kFND     */	ABSENT_STRING,
	/* 198: kHLP     */	ABSENT_STRING,
	/* 199: kHOM     */	xterm_s_kHOM,
	/* 200: kIC      */	xterm_s_kIC,
	/* 201: kLFT     */	xterm_s_kLFT,
	/* 202: kMSG     */	ABSENT_STRING,
	/* 203: kMOV     */	ABSENT_STRING,
	/* 204: kNXT     */	xterm_s_kNXT,
	/* 205: kOPT     */	ABSENT_STRING,
	/* 206: kPRV     */	xterm_s_kPRV,
	/* 207: kPRT     */	ABSENT_STRING,
	/* 208: kRDO     */	ABSENT_STRING,
	/* 209: kRPL     */	ABSENT_STRING,
	/* 210: kRIT     */	xterm_s_kRIT,
	/* 211: kRES     */	ABSENT_STRING,
	/* 212: kSAV     */	ABSENT_STRING,
	/* 213: kSPD     */	ABSENT_STRING,
	/* 214: kUND     */	ABSENT_STRING,
	/* 215: rfi      */	ABSENT_STRING,
	/* 216: kf11     */	xterm_s_kf11,
	/* 217: kf12     */	xterm_s_kf12,
	/* 218: kf13     */	xterm_s_kf13,
	/* 219: kf14     */	xterm_s_kf14,
	/* 220: kf15     */	xterm_s_kf15,
	/* 221: kf16     */	xterm_s_kf16,
	/* 222: kf17     */	xterm_s_kf17,
	/* 223: kf18     */	xterm_s_kf18,
	/* 224: kf19     */	xterm_s_kf19,
	/* 225: kf20     */	xterm_s_kf20,
	/* 226: kf21     */	xterm_s_kf21,
	/* 227: kf22     */	xterm_s_kf22,
	/* 228: kf23     */	xterm_s_kf23,
	/* 229: kf24     */	xterm_s_kf24,
	/* 230: kf25     */	xterm_s_kf25,
	/* 231: kf26     */	xterm_s_kf26,
	/* 232: kf27     */	xterm_s_kf27,
	/* 233: kf28     */	xterm_s_kf28,
	/* 234: kf29     */	xterm_s_kf29,
	/* 235: kf30     */	xterm_s_kf30,
	/* 236: kf31     */	xterm_s_kf31,
	/* 237: kf32     */	xterm_s_kf32,
	/* 238: kf33     */	xterm_s_kf33,
	/* 239: kf34     */	xterm_s_kf34,
	/* 240: kf35     */	xterm_s_kf35,
	/* 241: kf36     */	xterm_s_kf36,
	/* 242: kf37     */	xterm_s_kf37,
	/* 243: kf38     */	xterm_s_kf38,
	/* 244: kf39     */	xterm_s_kf39,
	/* 245: kf40     */	xterm_s_kf40,
	/* 246: kf41     */	xterm_s_kf41,
	/* 247: kf42     */	xterm_s_kf42,
	/* 248: kf43     */	xterm_s_kf43,
	/* 249: kf44     */	xterm_s_kf44,
	/* 250: kf45     */	xterm_s_kf45,
	/* 251: kf46     */	xterm_s_kf46,
	/* 252: kf47     */	xterm_s_kf47,
	/* 253: kf48     */	xterm_s_kf48,
	/* 254: kf49     */	xterm_s_kf49,
	/* 255: kf50     */	xterm_s_kf50,
	/* 256: kf51     */	xterm_s_kf51,
	/* 257: kf52     */	xterm_s_kf52,
	/* 258: kf53     */	xterm_s_kf53,
	/* 259: kf54     */	xterm_s_kf54,
	/* 260: kf55     */	xterm_s_kf55,
	/* 261: kf56     */	xterm_s_kf56,
	/* 262: kf57     */	xterm_s_kf57,
	/* 263: kf58     */	xterm_s_kf58,
	/* 264: kf59     */	xterm_s_kf59,
	/* 265: kf60     */	xterm_s_kf60,
	/* 266: kf61     */	xterm_s_kf61,
	/* 267: kf62     */	xterm_s_kf62,
	/* 268: kf63     */	xterm_s_kf63,
	/* 269: el1      */	xterm_s_el1,
	/* 270: mgc      */	ABSENT_STRING,
	/* 271: smgl     */	ABSENT_STRING,
	/* 272: smgr     */	ABSENT_STRING,
	/* 273: fln      */	ABSENT_STRING,
	/* 274: sclk     */	ABSENT_STRING,
	/* 275: dclk     */	ABSENT_STRING,
	/* 276: rmclk    */	ABSENT_STRING,
	/* 277: cwin     */	ABSENT_STRING,
	/* 278: wingo    */	ABSENT_STRING,
	/* 279: hup      */	ABSENT_STRING,
	/* 280: dial     */	ABSENT_STRING,
	/* 281: qdial    */	ABSENT_STRING,
	/* 282: tone     */	ABSENT_STRING,
	/* 283: pulse    */	ABSENT_STRING,
	/* 284: hook     */	ABSENT_STRING,
	/* 285: pause    */	ABSENT_STRING,
	/* 286: wait     */	ABSENT_STRING,
	/* 287: u0       */	ABSENT_STRING,
	/* 288: u1       */	ABSENT_STRING,
	/* 289: u2       */	ABSENT_STRING,
	/* 290: u3       */	ABSENT_STRING,
	/* 291: u4       */	ABSENT_STRING,
	/* 292: u5       */	ABSENT_STRING,
	/* 293: u6       */	xterm_s_u6,
	/* 294: u7       */	xterm_s_u7,
	/* 295: u8       */	xterm_s_u8,
	/* 296: u9       */	xterm_s_u9,
	/* 297: op       */	xterm_s_op,
	/* 298: oc       */	ABSENT_STRING,
	/* 299: initc    */	ABSENT_STRING,
	/* 300: initp    */	ABSENT_STRING,
	/* 301: scp      */	ABSENT_STRING,
	/* 302: setf     */	xterm_s_setf,
	/* 303: setb     */	xterm_s_setb,
	/* 304: cpi      */	ABSENT_STRING,
	/* 305: lpi      */	ABSENT_STRING,
	/* 306: chr      */	ABSENT_STRING,
	/* 307: cvr      */	ABSENT_STRING,
	/* 308: defc     */	ABSENT_STRING,
	/* 309: swidm    */	ABSENT_STRING,
	/* 310: sdrfq    */	ABSENT_STRING,
	/* 311: sitm     */	xterm_s_sitm,
	/* 312: slm      */	ABSENT_STRING,
	/* 313: smicm    */	ABSENT_STRING,
	/* 314: snlq     */	ABSENT_STRING,
	/* 315: snrmq    */	ABSENT_STRING,
	/* 316: sshm     */	ABSENT_STRING,
	/* 317: ssubm    */	ABSENT_STRING,
	/* 318: ssupm    */	ABSENT_STRING,
	/* 319: sum      */	ABSENT_STRING,
	/* 320: rwidm    */	ABSENT_STRING,
	/* 321: ritm     */	xterm_s_ritm,
	/* 322: rlm      */	ABSENT_STRING,
	/* 323: rmicm    */	ABSENT_STRING,
	/* 324: rshm     */	ABSENT_STRING,
	/* 325: rsubm    */	ABSENT_STRING,
	/* 326: rsupm    */	ABSENT_STRING,
	/* 327: rum      */	ABSENT_STRING,
	/* 328: mhpa     */	ABSENT_STRING,
	/* 329: mcud1    */	ABSENT_STRING,
	/* 330: mcub1    */	ABSENT_STRING,
	/* 331: mcuf1    */	ABSENT_STRING,
	/* 332: mvpa     */	ABSENT_STRING,
	/* 333: mcuu1    */	ABSENT_STRING,
	/* 334: porder   */	ABSENT_STRING,
	/* 335: mcud     */	ABSENT_STRING,
	/* 336: mcub     */	ABSENT_STRING,
	/* 337: mcuf     */	ABSENT_STRING,
	/* 338: mcuu     */	ABSENT_STRING,
	/* 339: scs      */	ABSENT_STRING,
	/* 340: smgb     */	ABSENT_STRING,
	/* 341: smgbp    */	ABSENT_STRING,
	/* 342: smglp    */	ABSENT_STRING,
	/* 343: smgrp    */	ABSENT_STRING,
	/* 344: smgt     */	ABSENT_STRING,
	/* 345: smgtp    */	ABSENT_STRING,
	/* 346: sbim     */	ABSENT_STRING,
	/* 347: scsd     */	ABSENT_STRING,
	/* 348: rbim     */	ABSENT_STRING,
	/* 349: rcsd     */	ABSENT_STRING,
	/* 350: subcs    */	ABSENT_STRING,
	/* 351: supcs    */	ABSENT_STRING,
	/* 352: docr     */	ABSENT_STRING,
	/* 353: zerom    */	ABSENT_STRING,
	/* 354: csnm     */	ABSENT_STRING,
	/* 355: kmous    */	xterm_s_kmous,
	/* 356: minfo    */	ABSENT_STRING,
	/* 357: reqmp    */	ABSENT_STRING,
	/* 358: getm     */	ABSENT_STRING,
	/* 359: setaf    */	xterm_s_setaf,
	/* 360: setab    */	xterm_s_setab,
	/* 361: pfxl     */	ABSENT_STRING,
	/* 362: devt     */	ABSENT_STRING,
	/* 363: csin     */	ABSENT_STRING,
	/* 364: s0ds     */	ABSENT_STRING,
	/* 365: s1ds     */	ABSENT_STRING,
	/* 366: s2ds     */	ABSENT_STRING,
	/* 367: s3ds     */	ABSENT_STRING,
	/* 368: smglr    */	ABSENT_STRING,
	/* 369: smgtb    */	ABSENT_STRING,
	/* 370: birep    */	ABSENT_STRING,
	/* 371: binel    */	ABSENT_STRING,
	/* 372: bicr     */	ABSENT_STRING,
	/* 373: colornm  */	ABSENT_STRING,
	/* 374: defbi    */	ABSENT_STRING,
	/* 375: endbi    */	ABSENT_STRING,
	/* 376: setcolor */	ABSENT_STRING,
	/* 377: slines   */	ABSENT_STRING,
	/* 378: dispc    */	ABSENT_STRING,
	/* 379: smpch    */	ABSENT_STRING,
	/* 380: rmpch    */	ABSENT_STRING,
	/* 381: smsc     */	ABSENT_STRING,
	/* 382: rmsc     */	ABSENT_STRING,
	/* 383: pctrm    */	ABSENT_STRING,
	/* 384: scesc    */	ABSENT_STRING,
	/* 385: scesa    */	ABSENT_STRING,
	/* 386: ehhlm    */	ABSENT_STRING,
	/* 387: elhlm    */	ABSENT_STRING,
	/* 388: elohlm   */	ABSENT_STRING,
	/* 389: erhlm    */	ABSENT_STRING,
	/* 390: ethlm    */	ABSENT_STRING,
	/* 391: evhlm    */	ABSENT_STRING,
	/* 392: sgr1     */	ABSENT_STRING,
	/* 393: slength  */	ABSENT_STRING,
	/* 394: OTi2     */	ABSENT_STRING,
	/* 395: OTrs     */	ABSENT_STRING,
	/* 396: OTnl     */	ABSENT_STRING,
	/* 397: OTbc     */	ABSENT_STRING,
	/* 398: OTko     */	ABSENT_STRING,
	/* 399: OTma     */	ABSENT_STRING,
	/* 400: OTG2     */	ABSENT_STRING,
	/* 401: OTG3     */	ABSENT_STRING,
	/* 402: OTG1     */	ABSENT_STRING,
	/* 403: OTG4     */	ABSENT_STRING,
	/* 404: OTGR     */	ABSENT_STRING,
	/* 405: OTGL     */	ABSENT_STRING,
	/* 406: OTGU     */	ABSENT_STRING,
	/* 407: OTGD     */	ABSENT_STRING,
	/* 408: OTGH     */	ABSENT_STRING,
	/* 409: OTGV     */	ABSENT_STRING,
	/* 410: OTGC     */	ABSENT_STRING,
	/* 411: meml     */	xterm_s_meml,
	/* 412: memu     */	xterm_s_memu,
	/* 413: box1     */	ABSENT_STRING,
};
/* xterm-256color */

static char xterm_256color_alias_data[] = "xterm-256color|xterm with 256 colors";

static char xterm_256color_s_cbt[] = "\033[Z";
static char xterm_256color_s_bel[] = "\007";
static char xterm_256color_s_cr [] = "\015";
static char xterm_256color_s_csr[] = "\033[%i%p1%d;%p2%dr";
static char xterm_256color_s_tbc[] = "\033[3g";
static char xterm_256color_s_clear[] = "\033[H\033[2J";
static char xterm_256color_s_el [] = "\033[K";
static char xterm_256color_s_ed [] = "\033[J";
static char xterm_256color_s_hpa[] = "\033[%i%p1%dG";
static char xterm_256color_s_cup[] = "\033[%i%p1%d;%p2%dH";
static char xterm_256color_s_cud1[] = "\012";
static char xterm_256color_s_home[] = "\033[H";
static char xterm_256color_s_civis[] = "\033[?25l";
static char xterm_256color_s_cub1[] = "\010";
static char xterm_256color_s_cnorm[] = "\033[?12l\033[?25h";
static char xterm_256color_s_cuf1[] = "\033[C";
static char xterm_256color_s_cuu1[] = "\033[A";
static char xterm_256color_s_cvvis[] = "\033[?12;25h";
static char xterm_256color_s_dch1[] = "\033[P";
static char xterm_256color_s_dl1[] = "\033[M";
static char xterm_256color_s_smacs[] = "\033(0";
static char xterm_256color_s_blink[] = "\033[5m";
static char xterm_256color_s_bold[] = "\033[1m";
static char xterm_256color_s_smcup[] = "\033[?1049h\033[22;0;0t";
static char xterm_256color_s_dim[] = "\033[2m";
static char xterm_256color_s_smir[] = "\033[4h";
static char xterm_256color_s_invis[] = "\033[8m";
static char xterm_256color_s_rev[] = "\033[7m";
static char xterm_256color_s_smso[] = "\033[7m";
static char xterm_256color_s_smul[] = "\033[4m";
static char xterm_256color_s_ech[] = "\033[%p1%dX";
static char xterm_256color_s_rmacs[] = "\033(B";
static char xterm_256color_s_sgr0[] = "\033(B\033[m";
static char xterm_256color_s_rmcup[] = "\033[?1049l\033[23;0;0t";
static char xterm_256color_s_rmir[] = "\033[4l";
static char xterm_256color_s_rmso[] = "\033[27m";
static char xterm_256color_s_rmul[] = "\033[24m";
static char xterm_256color_s_flash[] = "\033[?5h$<100/>\033[?5l";
static char xterm_256color_s_is2[] = "\033[!p\033[?3;4l\033[4l\033>";
static char xterm_256color_s_il1[] = "\033[L";
static char xterm_256color_s_kbs[] = "\010";
static char xterm_256color_s_kdch1[] = "\033[3~";
static char xterm_256color_s_kcud1[] = "\033OB";
static char xterm_256color_s_kf1[] = "\033OP";
static char xterm_256color_s_kf10[] = "\033[21~";
static char xterm_256color_s_kf2[] = "\033OQ";
static char xterm_256color_s_kf3[] = "\033OR";
static char xterm_256color_s_kf4[] = "\033OS";
static char xterm_256color_s_kf5[] = "\033[15~";
static char xterm_256color_s_kf6[] = "\033[17~";
static char xterm_256color_s_kf7[] = "\033[18~";
static char xterm_256color_s_kf8[] = "\033[19~";
static char xterm_256color_s_kf9[] = "\033[20~";
static char xterm_256color_s_khome[] = "\033OH";
static char xterm_256color_s_kich1[] = "\033[2~";
static char xterm_256color_s_kcub1[] = "\033OD";
static char xterm_256color_s_knp[] = "\033[6~";
static char xterm_256color_s_kpp[] = "\033[5~";
static char xterm_256color_s_kcuf1[] = "\033OC";
static char xterm_256color_s_kind[] = "\033[1;2B";
static char xterm_256color_s_kri[] = "\033[1;2A";
static char xterm_256color_s_kcuu1[] = "\033OA";
static char xterm_256color_s_rmkx[] = "\033[?1l\033>";
static char xterm_256color_s_smkx[] = "\033[?1h\033=";
static char xterm_256color_s_rmm[] = "\033[?1034l";
static char xterm_256color_s_smm[] = "\033[?1034h";
static char xterm_256color_s_dch[] = "\033[%p1%dP";
static char xterm_256color_s_dl [] = "\033[%p1%dM";
static char xterm_256color_s_cud[] = "\033[%p1%dB";
static char xterm_256color_s_ich[] = "\033[%p1%d@";
static char xterm_256color_s_indn[] = "\033[%p1%dS";
static char xterm_256color_s_il [] = "\033[%p1%dL";
static char xterm_256color_s_cub[] = "\033[%p1%dD";
static char xterm_256color_s_cuf[] = "\033[%p1%dC";
static char xterm_256color_s_rin[] = "\033[%p1%dT";
static char xterm_256color_s_cuu[] = "\033[%p1%dA";
static char xterm_256color_s_mc0[] = "\033[i";
static char xterm_256color_s_mc4[] = "\033[4i";
static char xterm_256color_s_mc5[] = "\033[5i";
static char xterm_256color_s_rep[] = "%p1%c\033[%p2%{1}%-%db";
static char xterm_256color_s_rs1[] = "\033c\033]104\007";
static char xterm_256color_s_rs2[] = "\033[!p\033[?3;4l\033[4l\033>";
static char xterm_256color_s_rc [] = "\0338";
static char xterm_256color_s_vpa[] = "\033[%i%p1%dd";
static char xterm_256color_s_sc [] = "\0337";
static char xterm_256color_s_ind[] = "\012";
static char xterm_256color_s_ri [] = "\033M";
static char xterm_256color_s_sgr[] = "%?%p9%t\033(0%e\033(B%;\033[0%?%p6%t;1%;%?%p5%t;2%;%?%p2%t;4%;%?%p1%p3%|%t;7%;%?%p4%t;5%;%?%p7%t;8%;m";
static char xterm_256color_s_hts[] = "\033H";
static char xterm_256color_s_ht [] = "\011";
static char xterm_256color_s_kb2[] = "\033OE";
static char xterm_256color_s_acsc[] = "``aaffggiijjkkllmmnnooppqqrrssttuuvvwwxxyyzz{{||}}~~";
static char xterm_256color_s_kcbt[] = "\033[Z";
static char xterm_256color_s_smam[] = "\033[?7h";
static char xterm_256color_s_rmam[] = "\033[?7l";
static char xterm_256color_s_kend[] = "\033OF";
static char xterm_256color_s_kent[] = "\033OM";
static char xterm_256color_s_kDC[] = "\033[3;2~";
static char xterm_256color_s_kEND[] = "\033[1;2F";
static char xterm_256color_s_kHOM[] = "\033[1;2H";
static char xterm_256color_s_kIC[] = "\033[2;2~";
static char xterm_256color_s_kLFT[] = "\033[1;2D";
static char xterm_256color_s_kNXT[] = "\033[6;2~";
static char xterm_256color_s_kPRV[] = "\033[5;2~";
static char xterm_256color_s_kRIT[] = "\033[1;2C";
static char xterm_256color_s_kf11[] = "\033[23~";
static char xterm_256color_s_kf12[] = "\033[24~";
static char xterm_256color_s_kf13[] = "\033[1;2P";
static char xterm_256color_s_kf14[] = "\033[1;2Q";
static char xterm_256color_s_kf15[] = "\033[1;2R";
static char xterm_256color_s_kf16[] = "\033[1;2S";
static char xterm_256color_s_kf17[] = "\033[15;2~";
static char xterm_256color_s_kf18[] = "\033[17;2~";
static char xterm_256color_s_kf19[] = "\033[18;2~";
static char xterm_256color_s_kf20[] = "\033[19;2~";
static char xterm_256color_s_kf21[] = "\033[20;2~";
static char xterm_256color_s_kf22[] = "\033[21;2~";
static char xterm_256color_s_kf23[] = "\033[23;2~";
static char xterm_256color_s_kf24[] = "\033[24;2~";
static char xterm_256color_s_kf25[] = "\033[1;5P";
static char xterm_256color_s_kf26[] = "\033[1;5Q";
static char xterm_256color_s_kf27[] = "\033[1;5R";
static char xterm_256color_s_kf28[] = "\033[1;5S";
static char xterm_256color_s_kf29[] = "\033[15;5~";
static char xterm_256color_s_kf30[] = "\033[17;5~";
static char xterm_256color_s_kf31[] = "\033[18;5~";
static char xterm_256color_s_kf32[] = "\033[19;5~";
static char xterm_256color_s_kf33[] = "\033[20;5~";
static char xterm_256color_s_kf34[] = "\033[21;5~";
static char xterm_256color_s_kf35[] = "\033[23;5~";
static char xterm_256color_s_kf36[] = "\033[24;5~";
static char xterm_256color_s_kf37[] = "\033[1;6P";
static char xterm_256color_s_kf38[] = "\033[1;6Q";
static char xterm_256color_s_kf39[] = "\033[1;6R";
static char xterm_256color_s_kf40[] = "\033[1;6S";
static char xterm_256color_s_kf41[] = "\033[15;6~";
static char xterm_256color_s_kf42[] = "\033[17;6~";
static char xterm_256color_s_kf43[] = "\033[18;6~";
static char xterm_256color_s_kf44[] = "\033[19;6~";
static char xterm_256color_s_kf45[] = "\033[20;6~";
static char xterm_256color_s_kf46[] = "\033[21;6~";
static char xterm_256color_s_kf47[] = "\033[23;6~";
static char xterm_256color_s_kf48[] = "\033[24;6~";
static char xterm_256color_s_kf49[] = "\033[1;3P";
static char xterm_256color_s_kf50[] = "\033[1;3Q";
static char xterm_256color_s_kf51[] = "\033[1;3R";
static char xterm_256color_s_kf52[] = "\033[1;3S";
static char xterm_256color_s_kf53[] = "\033[15;3~";
static char xterm_256color_s_kf54[] = "\033[17;3~";
static char xterm_256color_s_kf55[] = "\033[18;3~";
static char xterm_256color_s_kf56[] = "\033[19;3~";
static char xterm_256color_s_kf57[] = "\033[20;3~";
static char xterm_256color_s_kf58[] = "\033[21;3~";
static char xterm_256color_s_kf59[] = "\033[23;3~";
static char xterm_256color_s_kf60[] = "\033[24;3~";
static char xterm_256color_s_kf61[] = "\033[1;4P";
static char xterm_256color_s_kf62[] = "\033[1;4Q";
static char xterm_256color_s_kf63[] = "\033[1;4R";
static char xterm_256color_s_el1[] = "\033[1K";
static char xterm_256color_s_u6 [] = "\033[%i%d;%dR";
static char xterm_256color_s_u7 [] = "\033[6n";
static char xterm_256color_s_u8 [] = "\033[?%[;0123456789]c";
static char xterm_256color_s_u9 [] = "\033[c";
static char xterm_256color_s_op [] = "\033[39;49m";
static char xterm_256color_s_oc [] = "\033]104\007";
static char xterm_256color_s_initc[] = "\033]4;%p1%d;rgb:%p2%{255}%*%{1000}%/%2.2X/%p3%{255}%*%{1000}%/%2.2X/%p4%{255}%*%{1000}%/%2.2X\033\134";
static char xterm_256color_s_sitm[] = "\033[3m";
static char xterm_256color_s_ritm[] = "\033[23m";
static char xterm_256color_s_kmous[] = "\033[<";
static char xterm_256color_s_setaf[] = "\033[%?%p1%{8}%<%t3%p1%d%e%p1%{16}%<%t9%p1%{8}%-%d%e38;5;%p1%d%;m";
static char xterm_256color_s_setab[] = "\033[%?%p1%{8}%<%t4%p1%d%e%p1%{16}%<%t10%p1%{8}%-%d%e48;5;%p1%d%;m";
static char xterm_256color_s_meml[] = "\033l";
static char xterm_256color_s_memu[] = "\033m";

static char xterm_256color_bool_data[] = {
	/*   0: bw       */	FALSE,
	/*   1: am       */	TRUE,
	/*   2: xsb      */	FALSE,
	/*   3: xhp      */	FALSE,
	/*   4: xenl     */	TRUE,
	/*   5: eo       */	FALSE,
	/*   6: gn       */	FALSE,
	/*   7: hc       */	FALSE,
	/*   8: km       */	TRUE,
	/*   9: hs       */	FALSE,
	/*  10: in       */	FALSE,
	/*  11: da       */	FALSE,
	/*  12: db       */	FALSE,
	/*  13: mir      */	TRUE,
	/*  14: msgr     */	TRUE,
	/*  15: os       */	FALSE,
	/*  16: eslok    */	FALSE,
	/*  17: xt       */	FALSE,
	/*  18: hz       */	FALSE,
	/*  19: ul       */	FALSE,
	/*  20: xon      */	FALSE,
	/*  21: nxon     */	FALSE,
	/*  22: mc5i     */	TRUE,
	/*  23: chts     */	FALSE,
	/*  24: nrrmc    */	FALSE,
	/*  25: npc      */	TRUE,
	/*  26: ndscr    */	FALSE,
	/*  27: ccc      */	TRUE,
	/*  28: bce      */	TRUE,
	/*  29: hls      */	FALSE,
	/*  30: xhpa     */	FALSE,
	/*  31: crxm     */	FALSE,
	/*  32: daisy    */	FALSE,
	/*  33: xvpa     */	FALSE,
	/*  34: sam      */	FALSE,
	/*  35: cpix     */	FALSE,
	/*  36: lpix     */	FALSE,
	/*  37: OTbs     */	TRUE,
	/*  38: OTns     */	FALSE,
	/*  39: OTnc     */	FALSE,
	/*  40: OTMT     */	FALSE,
	/*  41: OTNL     */	FALSE,
	/*  42: OTpt     */	FALSE,
	/*  43: OTxr     */	FALSE,
};
static NCURSES_INT2 xterm_256color_number_data[] = {
	/*   0: cols     */	80,
	/*   1: it       */	8,
	/*   2: lines    */	24,
	/*   3: lm       */	ABSENT_NUMERIC,
	/*   4: xmc      */	ABSENT_NUMERIC,
	/*   5: pb       */	ABSENT_NUMERIC,
	/*   6: vt       */	ABSENT_NUMERIC,
	/*   7: wsl      */	ABSENT_NUMERIC,
	/*   8: nlab     */	ABSENT_NUMERIC,
	/*   9: lh       */	ABSENT_NUMERIC,
	/*  10: lw       */	ABSENT_NUMERIC,
	/*  11: ma       */	ABSENT_NUMERIC,
	/*  12: wnum     */	ABSENT_NUMERIC,
	/*  13: colors   */	256,
	/*  14: pairs    */	32767,
	/*  15: ncv      */	ABSENT_NUMERIC,
	/*  16: bufsz    */	ABSENT_NUMERIC,
	/*  17: spinv    */	ABSENT_NUMERIC,
	/*  18: spinh    */	ABSENT_NUMERIC,
	/*  19: maddr    */	ABSENT_NUMERIC,
	/*  20: mjump    */	ABSENT_NUMERIC,
	/*  21: mcs      */	ABSENT_NUMERIC,
	/*  22: mls      */	ABSENT_NUMERIC,
	/*  23: npins    */	ABSENT_NUMERIC,
	/*  24: orc      */	ABSENT_NUMERIC,
	/*  25: orl      */	ABSENT_NUMERIC,
	/*  26: orhi     */	ABSENT_NUMERIC,
	/*  27: orvi     */	ABSENT_NUMERIC,
	/*  28: cps      */	ABSENT_NUMERIC,
	/*  29: widcs    */	ABSENT_NUMERIC,
	/*  30: btns     */	ABSENT_NUMERIC,
	/*  31: bitwin   */	ABSENT_NUMERIC,
	/*  32: bitype   */	ABSENT_NUMERIC,
	/*  33: OTug     */	ABSENT_NUMERIC,
	/*  34: OTdC     */	ABSENT_NUMERIC,
	/*  35: OTdN     */	ABSENT_NUMERIC,
	/*  36: OTdB     */	ABSENT_NUMERIC,
	/*  37: OTdT     */	ABSENT_NUMERIC,
	/*  38: OTkn     */	ABSENT_NUMERIC,
};
static char * xterm_256color_string_data[] = {
	/*   0: cbt      */	xterm_256color_s_cbt,
	/*   1: bel      */	xterm_256color_s_bel,
	/*   2: cr       */	xterm_256color_s_cr,
	/*   3: csr      */	xterm_256color_s_csr,
	/*   4: tbc      */	xterm_256color_s_tbc,
	/*   5: clear    */	xterm_256color_s_clear,
	/*   6: el       */	xterm_256color_s_el,
	/*   7: ed       */	xterm_256color_s_ed,
	/*   8: hpa      */	xterm_256color_s_hpa,
	/*   9: cmdch    */	ABSENT_STRING,
	/*  10: cup      */	xterm_256color_s_cup,
	/*  11: cud1     */	xterm_256color_s_cud1,
	/*  12: home     */	xterm_256color_s_home,
	/*  13: civis    */	xterm_256color_s_civis,
	/*  14: cub1     */	xterm_256color_s_cub1,
	/*  15: mrcup    */	ABSENT_STRING,
	/*  16: cnorm    */	xterm_256color_s_cnorm,
	/*  17: cuf1     */	xterm_256color_s_cuf1,
	/*  18: ll       */	ABSENT_STRING,
	/*  19: cuu1     */	xterm_256color_s_cuu1,
	/*  20: cvvis    */	xterm_256color_s_cvvis,
	/*  21: dch1     */	xterm_256color_s_dch1,
	/*  22: dl1      */	xterm_256color_s_dl1,
	/*  23: dsl      */	ABSENT_STRING,
	/*  24: hd       */	ABSENT_STRING,
	/*  25: smacs    */	xterm_256color_s_smacs,
	/*  26: blink    */	xterm_256color_s_blink,
	/*  27: bold     */	xterm_256color_s_bold,
	/*  28: smcup    */	xterm_256color_s_smcup,
	/*  29: smdc     */	ABSENT_STRING,
	/*  30: dim      */	xterm_256color_s_dim,
	/*  31: smir     */	xterm_256color_s_smir,
	/*  32: invis    */	xterm_256color_s_invis,
	/*  33: prot     */	ABSENT_STRING,
	/*  34: rev      */	xterm_256color_s_rev,
	/*  35: smso     */	xterm_256color_s_smso,
	/*  36: smul     */	xterm_256color_s_smul,
	/*  37: ech      */	xterm_256color_s_ech,
	/*  38: rmacs    */	xterm_256color_s_rmacs,
	/*  39: sgr0     */	xterm_256color_s_sgr0,
	/*  40: rmcup    */	xterm_256color_s_rmcup,
	/*  41: rmdc     */	ABSENT_STRING,
	/*  42: rmir     */	xterm_256color_s_rmir,
	/*  43: rmso     */	xterm_256color_s_rmso,
	/*  44: rmul     */	xterm_256color_s_rmul,
	/*  45: flash    */	xterm_256color_s_flash,
	/*  46: ff       */	ABSENT_STRING,
	/*  47: fsl      */	ABSENT_STRING,
	/*  48: is1      */	ABSENT_STRING,
	/*  49: is2      */	xterm_256color_s_is2,
	/*  50: is3      */	ABSENT_STRING,
	/*  51: if       */	ABSENT_STRING,
	/*  52: ich1     */	ABSENT_STRING,
	/*  53: il1      */	xterm_256color_s_il1,
	/*  54: ip       */	ABSENT_STRING,
	/*  55: kbs      */	xterm_256color_s_kbs,
	/*  56: ktbc     */	ABSENT_STRING,
	/*  57: kclr     */	ABSENT_STRING,
	/*  58: kctab    */	ABSENT_STRING,
	/*  59: kdch1    */	xterm_256color_s_kdch1,
	/*  60: kdl1     */	ABSENT_STRING,
	/*  61: kcud1    */	xterm_256color_s_kcud1,
	/*  62: krmir    */	ABSENT_STRING,
	/*  63: kel      */	ABSENT_STRING,
	/*  64: ked      */	ABSENT_STRING,
	/*  65: kf0      */	ABSENT_STRING,
	/*  66: kf1      */	xterm_256color_s_kf1,
	/*  67: kf10     */	xterm_256color_s_kf10,
	/*  68: kf2      */	xterm_256color_s_kf2,
	/*  69: kf3      */	xterm_256color_s_kf3,
	/*  70: kf4      */	xterm_256color_s_kf4,
	/*  71: kf5      */	xterm_256color_s_kf5,
	/*  72: kf6      */	xterm_256color_s_kf6,
	/*  73: kf7      */	xterm_256color_s_kf7,
	/*  74: kf8      */	xterm_256color_s_kf8,
	/*  75: kf9      */	xterm_256color_s_kf9,
	/*  76: khome    */	xterm_256color_s_khome,
	/*  77: kich1    */	xterm_256color_s_kich1,
	/*  78: kil1     */	ABSENT_STRING,
	/*  79: kcub1    */	xterm_256color_s_kcub1,
	/*  80: kll      */	ABSENT_STRING,
	/*  81: knp      */	xterm_256color_s_knp,
	/*  82: kpp      */	xterm_256color_s_kpp,
	/*  83: kcuf1    */	xterm_256color_s_kcuf1,
	/*  84: kind     */	xterm_256color_s_kind,
	/*  85: kri      */	xterm_256color_s_kri,
	/*  86: khts     */	ABSENT_STRING,
	/*  87: kcuu1    */	xterm_256color_s_kcuu1,
	/*  88: rmkx     */	xterm_256color_s_rmkx,
	/*  89: smkx     */	xterm_256color_s_smkx,
	/*  90: lf0      */	ABSENT_STRING,
	/*  91: lf1      */	ABSENT_STRING,
	/*  92: lf10     */	ABSENT_STRING,
	/*  93: lf2      */	ABSENT_STRING,
	/*  94: lf3      */	ABSENT_STRING,
	/*  95: lf4      */	ABSENT_STRING,
	/*  96: lf5      */	ABSENT_STRING,
	/*  97: lf6      */	ABSENT_STRING,
	/*  98: lf7      */	ABSENT_STRING,
	/*  99: lf8      */	ABSENT_STRING,
	/* 100: lf9      */	ABSENT_STRING,
	/* 101: rmm      */	xterm_256color_s_rmm,
	/* 102: smm      */	xterm_256color_s_smm,
	/* 103: nel      */	ABSENT_STRING,
	/* 104: pad      */	ABSENT_STRING,
	/* 105: dch      */	xterm_256color_s_dch,
	/* 106: dl       */	xterm_256color_s_dl,
	/* 107: cud      */	xterm_256color_s_cud,
	/* 108: ich      */	xterm_256color_s_ich,
	/* 109: indn     */	xterm_256color_s_indn,
	/* 110: il       */	xterm_256color_s_il,
	/* 111: cub      */	xterm_256color_s_cub,
	/* 112: cuf      */	xterm_256color_s_cuf,
	/* 113: rin      */	xterm_256color_s_rin,
	/* 114: cuu      */	xterm_256color_s_cuu,
	/* 115: pfkey    */	ABSENT_STRING,
	/* 116: pfloc    */	ABSENT_STRING,
	/* 117: pfx      */	ABSENT_STRING,
	/* 118: mc0      */	xterm_256color_s_mc0,
	/* 119: mc4      */	xterm_256color_s_mc4,
	/* 120: mc5      */	xterm_256color_s_mc5,
	/* 121: rep      */	xterm_256color_s_rep,
	/* 122: rs1      */	xterm_256color_s_rs1,
	/* 123: rs2      */	xterm_256color_s_rs2,
	/* 124: rs3      */	ABSENT_STRING,
	/* 125: rf       */	ABSENT_STRING,
	/* 126: rc       */	xterm_256color_s_rc,
	/* 127: vpa      */	xterm_256color_s_vpa,
	/* 128: sc       */	xterm_256color_s_sc,
	/* 129: ind      */	xterm_256color_s_ind,
	/* 130: ri       */	xterm_256color_s_ri,
	/* 131: sgr      */	xterm_256color_s_sgr,
	/* 132: hts      */	xterm_256color_s_hts,
	/* 133: wind     */	ABSENT_STRING,
	/* 134: ht       */	xterm_256color_s_ht,
	/* 135: tsl      */	ABSENT_STRING,
	/* 136: uc       */	ABSENT_STRING,
	/* 137: hu       */	ABSENT_STRING,
	/* 138: iprog    */	ABSENT_STRING,
	/* 139: ka1      */	ABSENT_STRING,
	/* 140: ka3      */	ABSENT_STRING,
	/* 141: kb2      */	xterm_256color_s_kb2,
	/* 142: kc1      */	ABSENT_STRING,
	/* 143: kc3      */	ABSENT_STRING,
	/* 144: mc5p     */	ABSENT_STRING,
	/* 145: rmp      */	ABSENT_STRING,
	/* 146: acsc     */	xterm_256color_s_acsc,
	/* 147: pln      */	ABSENT_STRING,
	/* 148: kcbt     */	xterm_256color_s_kcbt,
	/* 149: smxon    */	ABSENT_STRING,
	/* 150: rmxon    */	ABSENT_STRING,
	/* 151: smam     */	xterm_256color_s_smam,
	/* 152: rmam     */	xterm_256color_s_rmam,
	/* 153: xonc     */	ABSENT_STRING,
	/* 154: xoffc    */	ABSENT_STRING,
	/* 155: enacs    */	ABSENT_STRING,
	/* 156: smln     */	ABSENT_STRING,
	/* 157: rmln     */	ABSENT_STRING,
	/* 158: kbeg     */	ABSENT_STRING,
	/* 159: kcan     */	ABSENT_STRING,
	/* 160: kclo     */	ABSENT_STRING,
	/* 161: kcmd     */	ABSENT_STRING,
	/* 162: kcpy     */	ABSENT_STRING,
	/* 163: kcrt     */	ABSENT_STRING,
	/* 164: kend     */	xterm_256color_s_kend,
	/* 165: kent     */	xterm_256color_s_kent,
	/* 166: kext     */	ABSENT_STRING,
	/* 167: kfnd     */	ABSENT_STRING,
	/* 168: khlp     */	ABSENT_STRING,
	/* 169: kmrk     */	ABSENT_STRING,
	/* 170: kmsg     */	ABSENT_STRING,
	/* 171: kmov     */	ABSENT_STRING,
	/* 172: knxt     */	ABSENT_STRING,
	/* 173: kopn     */	ABSENT_STRING,
	/* 174: kopt     */	ABSENT_STRING,
	/* 175: kprv     */	ABSENT_STRING,
	/* 176: kprt     */	ABSENT_STRING,
	/* 177: krdo     */	ABSENT_STRING,
	/* 178: kref     */	ABSENT_STRING,
	/* 179: krfr     */	ABSENT_STRING,
	/* 180: krpl     */	ABSENT_STRING,
	/* 181: krst     */	ABSENT_STRING,
	/* 182: kres     */	ABSENT_STRING,
	/* 183: ksav     */	ABSENT_STRING,
	/* 184: kspd     */	ABSENT_STRING,
	/* 185: kund     */	ABSENT_STRING,
	/* 186: kBEG     */	ABSENT_STRING,
	/* 187: kCAN     */	ABSENT_STRING,
	/* 188: kCMD     */	ABSENT_STRING,
	/* 189: kCPY     */	ABSENT_STRING,
	/* 190: kCRT     */	ABSENT_STRING,
	/* 191: kDC      */	xterm_256color_s_kDC,
	/* 192: kDL      */	ABSENT_STRING,
	/* 193: kslt     */	ABSENT_STRING,
	/* 194: kEND     */	xterm_256color_s_kEND,
	/* 195: kEOL     */	ABSENT_STRING,
	/* 196: kEXT     */	ABSENT_STRING,
	/* 197: kFND     */	ABSENT_STRING,
	/* 198: kHLP     */	ABSENT_STRING,
	/* 199: kHOM     */	xterm_256color_s_kHOM,
	/* 200: kIC      */	xterm_256color_s_kIC,
	/* 201: kLFT     */	xterm_256color_s_kLFT,
	/* 202: kMSG     */	ABSENT_STRING,
	/* 203: kMOV     */	ABSENT_STRING,
	/* 204: kNXT     */	xterm_256color_s_kNXT,
	/* 205: kOPT     */	ABSENT_STRING,
	/* 206: kPRV     */	xterm_256color_s_kPRV,
	/* 207: kPRT     */	ABSENT_STRING,
	/* 208: kRDO     */	ABSENT_STRING,
	/* 209: kRPL     */	ABSENT_STRING,
	/* 210: kRIT     */	xterm_256color_s_kRIT,
	/* 211: kRES     */	ABSENT_STRING,
	/* 212: kSAV     */	ABSENT_STRING,
	/* 213: kSPD     */	ABSENT_STRING,
	/* 214: kUND     */	ABSENT_STRING,
	/* 215: rfi      */	ABSENT_STRING,
	/* 216: kf11     */	xterm_256color_s_kf11,
	/* 217: kf12     */	xterm_256color_s_kf12,
	/* 218: kf13     */	xterm_256color_s_kf13,
	/* 219: kf14     */	xterm_256color_s_kf14,
	/* 220: kf15     */	xterm_256color_s_kf15,
	/* 221: kf16     */	xterm_256color_s_kf16,
	/* 222: kf17     */	xterm_256color_s_kf17,
	/* 223: kf18     */	xterm_256color_s_kf18,
	/* 224: kf19     */	xterm_256color_s_kf19,
	/* 225: kf20     */	xterm_256color_s_kf20,
	/* 226: kf21     */	xterm_256color_s_kf21,
	/* 227: kf22     */	xterm_256color_s_kf22,
	/* 228: kf23     */	xterm_256color_s_kf23,
	/* 229: kf24     */	xterm_256color_s_kf24,
	/* 230: kf25     */	xterm_256color_s_kf25,
	/* 231: kf26     */	xterm_256color_s_kf26,
	/* 232: kf27     */	xterm_256color_s_kf27,
	/* 233: kf28     */	xterm_256color_s_kf28,
	/* 234: kf29     */	xterm_256color_s_kf29,
	/* 235: kf30     */	xterm_256color_s_kf30,
	/* 236: kf31     */	xterm_256color_s_kf31,
	/* 237: kf32     */	xterm_256color_s_kf32,
	/* 238: kf33     */	xterm_256color_s_kf33,
	/* 239: kf34     */	xterm_256color_s_kf34,
	/* 240: kf35     */	xterm_256color_s_kf35,
	/* 241: kf36     */	xterm_256color_s_kf36,
	/* 242: kf37     */	xterm_256color_s_kf37,
	/* 243: kf38     */	xterm_256color_s_kf38,
	/* 244: kf39     */	xterm_256color_s_kf39,
	/* 245: kf40     */	xterm_256color_s_kf40,
	/* 246: kf41     */	xterm_256color_s_kf41,
	/* 247: kf42     */	xterm_256color_s_kf42,
	/* 248: kf43     */	xterm_256color_s_kf43,
	/* 249: kf44     */	xterm_256color_s_kf44,
	/* 250: kf45     */	xterm_256color_s_kf45,
	/* 251: kf46     */	xterm_256color_s_kf46,
	/* 252: kf47     */	xterm_256color_s_kf47,
	/* 253: kf48     */	xterm_256color_s_kf48,
	/* 254: kf49     */	xterm_256color_s_kf49,
	/* 255: kf50     */	xterm_256color_s_kf50,
	/* 256: kf51     */	xterm_256color_s_kf51,
	/* 257: kf52     */	xterm_256color_s_kf52,
	/* 258: kf53     */	xterm_256color_s_kf53,
	/* 259: kf54     */	xterm_256color_s_kf54,
	/* 260: kf55     */	xterm_256color_s_kf55,
	/* 261: kf56     */	xterm_256color_s_kf56,
	/* 262: kf57     */	xterm_256color_s_kf57,
	/* 263: kf58     */	xterm_256color_s_kf58,
	/* 264: kf59     */	xterm_256color_s_kf59,
	/* 265: kf60     */	xterm_256color_s_kf60,
	/* 266: kf61     */	xterm_256color_s_kf61,
	/* 267: kf62     */	xterm_256color_s_kf62,
	/* 268: kf63     */	xterm_256color_s_kf63,
	/* 269: el1      */	xterm_256color_s_el1,
	/* 270: mgc      */	ABSENT_STRING,
	/* 271: smgl     */	ABSENT_STRING,
	/* 272: smgr     */	ABSENT_STRING,
	/* 273: fln      */	ABSENT_STRING,
	/* 274: sclk     */	ABSENT_STRING,
	/* 275: dclk     */	ABSENT_STRING,
	/* 276: rmclk    */	ABSENT_STRING,
	/* 277: cwin     */	ABSENT_STRING,
	/* 278: wingo    */	ABSENT_STRING,
	/* 279: hup      */	ABSENT_STRING,
	/* 280: dial     */	ABSENT_STRING,
	/* 281: qdial    */	ABSENT_STRING,
	/* 282: tone     */	ABSENT_STRING,
	/* 283: pulse    */	ABSENT_STRING,
	/* 284: hook     */	ABSENT_STRING,
	/* 285: pause    */	ABSENT_STRING,
	/* 286: wait     */	ABSENT_STRING,
	/* 287: u0       */	ABSENT_STRING,
	/* 288: u1       */	ABSENT_STRING,
	/* 289: u2       */	ABSENT_STRING,
	/* 290: u3       */	ABSENT_STRING,
	/* 291: u4       */	ABSENT_STRING,
	/* 292: u5       */	ABSENT_STRING,
	/* 293: u6       */	xterm_256color_s_u6,
	/* 294: u7       */	xterm_256color_s_u7,
	/* 295: u8       */	xterm_256color_s_u8,
	/* 296: u9       */	xterm_256color_s_u9,
	/* 297: op       */	xterm_256color_s_op,
	/* 298: oc       */	xterm_256color_s_oc,
	/* 299: initc    */	xterm_256color_s_initc,
	/* 300: initp    */	ABSENT_STRING,
	/* 301: scp      */	ABSENT_STRING,
	/* 302: setf     */	ABSENT_STRING,
	/* 303: setb     */	ABSENT_STRING,
	/* 304: cpi      */	ABSENT_STRING,
	/* 305: lpi      */	ABSENT_STRING,
	/* 306: chr      */	ABSENT_STRING,
	/* 307: cvr      */	ABSENT_STRING,
	/* 308: defc     */	ABSENT_STRING,
	/* 309: swidm    */	ABSENT_STRING,
	/* 310: sdrfq    */	ABSENT_STRING,
	/* 311: sitm     */	xterm_256color_s_sitm,
	/* 312: slm      */	ABSENT_STRING,
	/* 313: smicm    */	ABSENT_STRING,
	/* 314: snlq     */	ABSENT_STRING,
	/* 315: snrmq    */	ABSENT_STRING,
	/* 316: sshm     */	ABSENT_STRING,
	/* 317: ssubm    */	ABSENT_STRING,
	/* 318: ssupm    */	ABSENT_STRING,
	/* 319: sum      */	ABSENT_STRING,
	/* 320: rwidm    */	ABSENT_STRING,
	/* 321: ritm     */	xterm_256color_s_ritm,
	/* 322: rlm      */	ABSENT_STRING,
	/* 323: rmicm    */	ABSENT_STRING,
	/* 324: rshm     */	ABSENT_STRING,
	/* 325: rsubm    */	ABSENT_STRING,
	/* 326: rsupm    */	ABSENT_STRING,
	/* 327: rum      */	ABSENT_STRING,
	/* 328: mhpa     */	ABSENT_STRING,
	/* 329: mcud1    */	ABSENT_STRING,
	/* 330: mcub1    */	ABSENT_STRING,
	/* 331: mcuf1    */	ABSENT_STRING,
	/* 332: mvpa     */	ABSENT_STRING,
	/* 333: mcuu1    */	ABSENT_STRING,
	/* 334: porder   */	ABSENT_STRING,
	/* 335: mcud     */	ABSENT_STRING,
	/* 336: mcub     */	ABSENT_STRING,
	/* 337: mcuf     */	ABSENT_STRING,
	/* 338: mcuu     */	ABSENT_STRING,
	/* 339: scs      */	ABSENT_STRING,
	/* 340: smgb     */	ABSENT_STRING,
	/* 341: smgbp    */	ABSENT_STRING,
	/* 342: smglp    */	ABSENT_STRING,
	/* 343: smgrp    */	ABSENT_STRING,
	/* 344: smgt     */	ABSENT_STRING,
	/* 345: smgtp    */	ABSENT_STRING,
	/* 346: sbim     */	ABSENT_STRING,
	/* 347: scsd     */	ABSENT_STRING,
	/* 348: rbim     */	ABSENT_STRING,
	/* 349: rcsd     */	ABSENT_STRING,
	/* 350: subcs    */	ABSENT_STRING,
	/* 351: supcs    */	ABSENT_STRING,
	/* 352: docr     */	ABSENT_STRING,
	/* 353: zerom    */	ABSENT_STRING,
	/* 354: csnm     */	ABSENT_STRING,
	/* 355: kmous    */	xterm_256color_s_kmous,
	/* 356: minfo    */	ABSENT_STRING,
	/* 357: reqmp    */	ABSENT_STRING,
	/* 358: getm     */	ABSENT_STRING,
	/* 359: setaf    */	xterm_256color_s_setaf,
	/* 360: setab    */	xterm_256color_s_setab,
	/* 361: pfxl     */	ABSENT_STRING,
	/* 362: devt     */	ABSENT_STRING,
	/* 363: csin     */	ABSENT_STRING,
	/* 364: s0ds     */	ABSENT_STRING,
	/* 365: s1ds     */	ABSENT_STRING,
	/* 366: s2ds     */	ABSENT_STRING,
	/* 367: s3ds     */	ABSENT_STRING,
	/* 368: smglr    */	ABSENT_STRING,
	/* 369: smgtb    */	ABSENT_STRING,
	/* 370: birep    */	ABSENT_STRING,
	/* 371: binel    */	ABSENT_STRING,
	/* 372: bicr     */	ABSENT_STRING,
	/* 373: colornm  */	ABSENT_STRING,
	/* 374: defbi    */	ABSENT_STRING,
	/* 375: endbi    */	ABSENT_STRING,
	/* 376: setcolor */	ABSENT_STRING,
	/* 377: slines   */	ABSENT_STRING,
	/* 378: dispc    */	ABSENT_STRING,
	/* 379: smpch    */	ABSENT_STRING,
	/* 380: rmpch    */	ABSENT_STRING,
	/* 381: smsc     */	ABSENT_STRING,
	/* 382: rmsc     */	ABSENT_STRING,
	/* 383: pctrm    */	ABSENT_STRING,
	/* 384: scesc    */	ABSENT_STRING,
	/* 385: scesa    */	ABSENT_STRING,
	/* 386: ehhlm    */	ABSENT_STRING,
	/* 387: elhlm    */	ABSENT_STRING,
	/* 388: elohlm   */	ABSENT_STRING,
	/* 389: erhlm    */	ABSENT_STRING,
	/* 390: ethlm    */	ABSENT_STRING,
	/* 391: evhlm    */	ABSENT_STRING,
	/* 392: sgr1     */	ABSENT_STRING,
	/* 393: slength  */	ABSENT_STRING,
	/* 394: OTi2     */	ABSENT_STRING,
	/* 395: OTrs     */	ABSENT_STRING,
	/* 396: OTnl     */	ABSENT_STRING,
	/* 397: OTbc     */	ABSENT_STRING,
	/* 398: OTko     */	ABSENT_STRING,
	/* 399: OTma     */	ABSENT_STRING,
	/* 400: OTG2     */	ABSENT_STRING,
	/* 401: OTG3     */	ABSENT_STRING,
	/* 402: OTG1     */	ABSENT_STRING,
	/* 403: OTG4     */	ABSENT_STRING,
	/* 404: OTGR     */	ABSENT_STRING,
	/* 405: OTGL     */	ABSENT_STRING,
	/* 406: OTGU     */	ABSENT_STRING,
	/* 407: OTGD     */	ABSENT_STRING,
	/* 408: OTGH     */	ABSENT_STRING,
	/* 409: OTGV     */	ABSENT_STRING,
	/* 410: OTGC     */	ABSENT_STRING,
	/* 411: meml     */	xterm_256color_s_meml,
	/* 412: memu     */	xterm_256color_s_memu,
	/* 413: box1     */	ABSENT_STRING,
};
/* screen */

static char screen_alias_data[] = "screen|VT 100/ANSI X3.64 virtual terminal";

static char screen_s_cbt        [] = "\033[Z";
static char screen_s_bel        [] = "\007";
static char screen_s_cr         [] = "\015";
static char screen_s_csr        [] = "\033[%i%p1%d;%p2%dr";
static char screen_s_tbc        [] = "\033[3g";
static char screen_s_clear      [] = "\033[H\033[J";
static char screen_s_el         [] = "\033[K";
static char screen_s_ed         [] = "\033[J";
static char screen_s_hpa        [] = "\033[%i%p1%dG";
static char screen_s_cup        [] = "\033[%i%p1%d;%p2%dH";
static char screen_s_cud1       [] = "\012";
static char screen_s_home       [] = "\033[H";
static char screen_s_civis      [] = "\033[?25l";
static char screen_s_cub1       [] = "\010";
static char screen_s_cnorm      [] = "\033[34h\033[?25h";
static char screen_s_cuf1       [] = "\033[C";
static char screen_s_cuu1       [] = "\033M";
static char screen_s_cvvis      [] = "\033[34l";
static char screen_s_dch1       [] = "\033[P";
static char screen_s_dl1        [] = "\033[M";
static char screen_s_smacs      [] = "\016";
static char screen_s_blink      [] = "\033[5m";
static char screen_s_bold       [] = "\033[1m";
static char screen_s_smcup      [] = "\033[?1049h";
static char screen_s_dim        [] = "\033[2m";
static char screen_s_smir       [] = "\033[4h";
static char screen_s_rev        [] = "\033[7m";
static char screen_s_smso       [] = "\033[3m";
static char screen_s_smul       [] = "\033[4m";
static char screen_s_rmacs      [] = "\017";
static char screen_s_sgr0       [] = "\033[m\017";
static char screen_s_rmcup      [] = "\033[?1049l";
static char screen_s_rmir       [] = "\033[4l";
static char screen_s_rmso       [] = "\033[23m";
static char screen_s_rmul       [] = "\033[24m";
static char screen_s_flash      [] = "\033g";
static char screen_s_is2        [] = "\033)0";
static char screen_s_il1        [] = "\033[L";
static char screen_s_kbs        [] = "\010";
static char screen_s_kdch1      [] = "\033[3~";
static char screen_s_kcud1      [] = "\033OB";
static char screen_s_kf1        [] = "\033OP";
static char screen_s_kf10       [] = "\033[21~";
static char screen_s_kf2        [] = "\033OQ";
static char screen_s_kf3        [] = "\033OR";
static char screen_s_kf4        [] = "\033OS";
static char screen_s_kf5        [] = "\033[15~";
static char screen_s_kf6        [] = "\033[17~";
static char screen_s_kf7        [] = "\033[18~";
static char screen_s_kf8        [] = "\033[19~";
static char screen_s_kf9        [] = "\033[20~";
static char screen_s_khome      [] = "\033[1~";
static char screen_s_kich1      [] = "\033[2~";
static char screen_s_kcub1      [] = "\033OD";
static char screen_s_knp        [] = "\033[6~";
static char screen_s_kpp        [] = "\033[5~";
static char screen_s_kcuf1      [] = "\033OC";
static char screen_s_kcuu1      [] = "\033OA";
static char screen_s_rmkx       [] = "\033[?1l\033>";
static char screen_s_smkx       [] = "\033[?1h\033=";
static char screen_s_nel        [] = "\033E";
static char screen_s_dch        [] = "\033[%p1%dP";
static char screen_s_dl         [] = "\033[%p1%dM";
static char screen_s_cud        [] = "\033[%p1%dB";
static char screen_s_ich        [] = "\033[%p1%d@";
static char screen_s_indn       [] = "\033[%p1%dS";
static char screen_s_il         [] = "\033[%p1%dL";
static char screen_s_cub        [] = "\033[%p1%dD";
static char screen_s_cuf        [] = "\033[%p1%dC";
static char screen_s_cuu        [] = "\033[%p1%dA";
static char screen_s_rs2        [] = "\033c\033[?1000l\033[?25h";
static char screen_s_rc         [] = "\0338";
static char screen_s_vpa        [] = "\033[%i%p1%dd";
static char screen_s_sc         [] = "\0337";
static char screen_s_ind        [] = "\012";
static char screen_s_ri         [] = "\033M";
static char screen_s_sgr        [] = "\033[0%?%p6%t;1%;%?%p1%t;3%;%?%p2%t;4%;%?%p3%t;7%;%?%p4%t;5%;%?%p5%t;2%;m%?%p9%t\016%e\017%;";
static char screen_s_hts        [] = "\033H";
static char screen_s_ht         [] = "\011";
static char screen_s_acsc       [] = "++,,--..00``aaffgghhiijjkkllmmnnooppqqrrssttuuvvwwxxyyzz{{||}}~~";
static char screen_s_kcbt       [] = "\033[Z";
static char screen_s_enacs      [] = "\033(B\033)0";
static char screen_s_kend       [] = "\033[4~";
static char screen_s_kf11       [] = "\033[23~";
static char screen_s_kf12       [] = "\033[24~";
static char screen_s_el1        [] = "\033[1K";
static char screen_s_op         [] = "\033[39;49m";
static char screen_s_kmous      [] = "\033[M";
static char screen_s_setaf      [] = "\033[3%p1%dm";
static char screen_s_setab      [] = "\033[4%p1%dm";

static char screen_bool_data[] = {
	/*   0: bw       */	FALSE,
	/*   1: am       */	TRUE,
	/*   2: xsb      */	FALSE,
	/*   3: xhp      */	FALSE,
	/*   4: xenl     */	TRUE,
	/*   5: eo       */	FALSE,
	/*   6: gn       */	FALSE,
	/*   7: hc       */	FALSE,
	/*   8: km       */	TRUE,
	/*   9: hs       */	FALSE,
	/*  10: in       */	FALSE,
	/*  11: da       */	FALSE,
	/*  12: db       */	FALSE,
	/*  13: mir      */	TRUE,
	/*  14: msgr     */	TRUE,
	/*  15: os       */	FALSE,
	/*  16: eslok    */	FALSE,
	/*  17: xt       */	FALSE,
	/*  18: hz       */	FALSE,
	/*  19: ul       */	FALSE,
	/*  20: xon      */	FALSE,
	/*  21: nxon     */	FALSE,
	/*  22: mc5i     */	FALSE,
	/*  23: chts     */	FALSE,
	/*  24: nrrmc    */	FALSE,
	/*  25: npc      */	FALSE,
	/*  26: ndscr    */	FALSE,
	/*  27: ccc      */	FALSE,
	/*  28: bce      */	FALSE,
	/*  29: hls      */	FALSE,
	/*  30: xhpa     */	FALSE,
	/*  31: crxm     */	FALSE,
	/*  32: daisy    */	FALSE,
	/*  33: xvpa     */	FALSE,
	/*  34: sam      */	FALSE,
	/*  35: cpix     */	FALSE,
	/*  36: lpix     */	FALSE,
	/*  37: OTbs     */	TRUE,
	/*  38: OTns     */	FALSE,
	/*  39: OTnc     */	FALSE,
	/*  40: OTMT     */	FALSE,
	/*  41: OTNL     */	FALSE,
	/*  42: OTpt     */	TRUE,
	/*  43: OTxr     */	FALSE,
};
static NCURSES_INT2 screen_number_data[] = {
	/*   0: cols     */	80,
	/*   1: it       */	8,
	/*   2: lines    */	24,
	/*   3: lm       */	ABSENT_NUMERIC,
	/*   4: xmc      */	ABSENT_NUMERIC,
	/*   5: pb       */	ABSENT_NUMERIC,
	/*   6: vt       */	ABSENT_NUMERIC,
	/*   7: wsl      */	ABSENT_NUMERIC,
	/*   8: nlab     */	ABSENT_NUMERIC,
	/*   9: lh       */	ABSENT_NUMERIC,
	/*  10: lw       */	ABSENT_NUMERIC,
	/*  11: ma       */	ABSENT_NUMERIC,
	/*  12: wnum     */	ABSENT_NUMERIC,
	/*  13: colors   */	8,
	/*  14: pairs    */	64,
	/*  15: ncv      */	CANCELLED_NUMERIC,
	/*  16: bufsz    */	ABSENT_NUMERIC,
	/*  17: spinv    */	ABSENT_NUMERIC,
	/*  18: spinh    */	ABSENT_NUMERIC,
	/*  19: maddr    */	ABSENT_NUMERIC,
	/*  20: mjump    */	ABSENT_NUMERIC,
	/*  21: mcs      */	ABSENT_NUMERIC,
	/*  22: mls      */	ABSENT_NUMERIC,
	/*  23: npins    */	ABSENT_NUMERIC,
	/*  24: orc      */	ABSENT_NUMERIC,
	/*  25: orl      */	ABSENT_NUMERIC,
	/*  26: orhi     */	ABSENT_NUMERIC,
	/*  27: orvi     */	ABSENT_NUMERIC,
	/*  28: cps      */	ABSENT_NUMERIC,
	/*  29: widcs    */	ABSENT_NUMERIC,
	/*  30: btns     */	ABSENT_NUMERIC,
	/*  31: bitwin   */	ABSENT_NUMERIC,
	/*  32: bitype   */	ABSENT_NUMERIC,
	/*  33: OTug     */	ABSENT_NUMERIC,
	/*  34: OTdC     */	ABSENT_NUMERIC,
	/*  35: OTdN     */	ABSENT_NUMERIC,
	/*  36: OTdB     */	ABSENT_NUMERIC,
	/*  37: OTdT     */	ABSENT_NUMERIC,
	/*  38: OTkn     */	ABSENT_NUMERIC,
};
static char * screen_string_data[] = {
	/*   0: cbt      */	screen_s_cbt,
	/*   1: bel      */	screen_s_bel,
	/*   2: cr       */	screen_s_cr,
	/*   3: csr      */	screen_s_csr,
	/*   4: tbc      */	screen_s_tbc,
	/*   5: clear    */	screen_s_clear,
	/*   6: el       */	screen_s_el,
	/*   7: ed       */	screen_s_ed,
	/*   8: hpa      */	screen_s_hpa,
	/*   9: cmdch    */	ABSENT_STRING,
	/*  10: cup      */	screen_s_cup,
	/*  11: cud1     */	screen_s_cud1,
	/*  12: home     */	screen_s_home,
	/*  13: civis    */	screen_s_civis,
	/*  14: cub1     */	screen_s_cub1,
	/*  15: mrcup    */	ABSENT_STRING,
	/*  16: cnorm    */	screen_s_cnorm,
	/*  17: cuf1     */	screen_s_cuf1,
	/*  18: ll       */	ABSENT_STRING,
	/*  19: cuu1     */	screen_s_cuu1,
	/*  20: cvvis    */	screen_s_cvvis,
	/*  21: dch1     */	screen_s_dch1,
	/*  22: dl1      */	screen_s_dl1,
	/*  23: dsl      */	ABSENT_STRING,
	/*  24: hd       */	ABSENT_STRING,
	/*  25: smacs    */	screen_s_smacs,
	/*  26: blink    */	screen_s_blink,
	/*  27: bold     */	screen_s_bold,
	/*  28: smcup    */	screen_s_smcup,
	/*  29: smdc     */	ABSENT_STRING,
	/*  30: dim      */	screen_s_dim,
	/*  31: smir     */	screen_s_smir,
	/*  32: invis    */	ABSENT_STRING,
	/*  33: prot     */	ABSENT_STRING,
	/*  34: rev      */	screen_s_rev,
	/*  35: smso     */	screen_s_smso,
	/*  36: smul     */	screen_s_smul,
	/*  37: ech      */	ABSENT_STRING,
	/*  38: rmacs    */	screen_s_rmacs,
	/*  39: sgr0     */	screen_s_sgr0,
	/*  40: rmcup    */	screen_s_rmcup,
	/*  41: rmdc     */	ABSENT_STRING,
	/*  42: rmir     */	screen_s_rmir,
	/*  43: rmso     */	screen_s_rmso,
	/*  44: rmul     */	screen_s_rmul,
	/*  45: flash    */	screen_s_flash,
	/*  46: ff       */	ABSENT_STRING,
	/*  47: fsl      */	ABSENT_STRING,
	/*  48: is1      */	ABSENT_STRING,
	/*  49: is2      */	screen_s_is2,
	/*  50: is3      */	ABSENT_STRING,
	/*  51: if       */	ABSENT_STRING,
	/*  52: ich1     */	ABSENT_STRING,
	/*  53: il1      */	screen_s_il1,
	/*  54: ip       */	ABSENT_STRING,
	/*  55: kbs      */	screen_s_kbs,
	/*  56: ktbc     */	ABSENT_STRING,
	/*  57: kclr     */	ABSENT_STRING,
	/*  58: kctab    */	ABSENT_STRING,
	/*  59: kdch1    */	screen_s_kdch1,
	/*  60: kdl1     */	ABSENT_STRING,
	/*  61: kcud1    */	screen_s_kcud1,
	/*  62: krmir    */	ABSENT_STRING,
	/*  63: kel      */	ABSENT_STRING,
	/*  64: ked      */	ABSENT_STRING,
	/*  65: kf0      */	ABSENT_STRING,
	/*  66: kf1      */	screen_s_kf1,
	/*  67: kf10     */	screen_s_kf10,
	/*  68: kf2      */	screen_s_kf2,
	/*  69: kf3      */	screen_s_kf3,
	/*  70: kf4      */	screen_s_kf4,
	/*  71: kf5      */	screen_s_kf5,
	/*  72: kf6      */	screen_s_kf6,
	/*  73: kf7      */	screen_s_kf7,
	/*  74: kf8      */	screen_s_kf8,
	/*  75: kf9      */	screen_s_kf9,
	/*  76: khome    */	screen_s_khome,
	/*  77: kich1    */	screen_s_kich1,
	/*  78: kil1     */	ABSENT_STRING,
	/*  79: kcub1    */	screen_s_kcub1,
	/*  80: kll      */	ABSENT_STRING,
	/*  81: knp      */	screen_s_knp,
	/*  82: kpp      */	screen_s_kpp,
	/*  83: kcuf1    */	screen_s_kcuf1,
	/*  84: kind     */	ABSENT_STRING,
	/*  85: kri      */	ABSENT_STRING,
	/*  86: khts     */	ABSENT_STRING,
	/*  87: kcuu1    */	screen_s_kcuu1,
	/*  88: rmkx     */	screen_s_rmkx,
	/*  89: smkx     */	screen_s_smkx,
	/*  90: lf0      */	ABSENT_STRING,
	/*  91: lf1      */	ABSENT_STRING,
	/*  92: lf10     */	ABSENT_STRING,
	/*  93: lf2      */	ABSENT_STRING,
	/*  94: lf3      */	ABSENT_STRING,
	/*  95: lf4      */	ABSENT_STRING,
	/*  96: lf5      */	ABSENT_STRING,
	/*  97: lf6      */	ABSENT_STRING,
	/*  98: lf7      */	ABSENT_STRING,
	/*  99: lf8      */	ABSENT_STRING,
	/* 100: lf9      */	ABSENT_STRING,
	/* 101: rmm      */	ABSENT_STRING,
	/* 102: smm      */	ABSENT_STRING,
	/* 103: nel      */	screen_s_nel,
	/* 104: pad      */	ABSENT_STRING,
	/* 105: dch      */	screen_s_dch,
	/* 106: dl       */	screen_s_dl,
	/* 107: cud      */	screen_s_cud,
	/* 108: ich      */	screen_s_ich,
	/* 109: indn     */	screen_s_indn,
	/* 110: il       */	screen_s_il,
	/* 111: cub      */	screen_s_cub,
	/* 112: cuf      */	screen_s_cuf,
	/* 113: rin      */	ABSENT_STRING,
	/* 114: cuu      */	screen_s_cuu,
	/* 115: pfkey    */	ABSENT_STRING,
	/* 116: pfloc    */	ABSENT_STRING,
	/* 117: pfx      */	ABSENT_STRING,
	/* 118: mc0      */	ABSENT_STRING,
	/* 119: mc4      */	ABSENT_STRING,
	/* 120: mc5      */	ABSENT_STRING,
	/* 121: rep      */	ABSENT_STRING,
	/* 122: rs1      */	ABSENT_STRING,
	/* 123: rs2      */	screen_s_rs2,
	/* 124: rs3      */	ABSENT_STRING,
	/* 125: rf       */	ABSENT_STRING,
	/* 126: rc       */	screen_s_rc,
	/* 127: vpa      */	screen_s_vpa,
	/* 128: sc       */	screen_s_sc,
	/* 129: ind      */	screen_s_ind,
	/* 130: ri       */	screen_s_ri,
	/* 131: sgr      */	screen_s_sgr,
	/* 132: hts      */	screen_s_hts,
	/* 133: wind     */	ABSENT_STRING,
	/* 134: ht       */	screen_s_ht,
	/* 135: tsl      */	ABSENT_STRING,
	/* 136: uc       */	ABSENT_STRING,
	/* 137: hu       */	ABSENT_STRING,
	/* 138: iprog    */	ABSENT_STRING,
	/* 139: ka1      */	ABSENT_STRING,
	/* 140: ka3      */	ABSENT_STRING,
	/* 141: kb2      */	ABSENT_STRING,
	/* 142: kc1      */	ABSENT_STRING,
	/* 143: kc3      */	ABSENT_STRING,
	/* 144: mc5p     */	ABSENT_STRING,
	/* 145: rmp      */	ABSENT_STRING,
	/* 146: acsc     */	screen_s_acsc,
	/* 147: pln      */	ABSENT_STRING,
	/* 148: kcbt     */	screen_s_kcbt,
	/* 149: smxon    */	ABSENT_STRING,
	/* 150: rmxon    */	ABSENT_STRING,
	/* 151: smam     */	ABSENT_STRING,
	/* 152: rmam     */	ABSENT_STRING,
	/* 153: xonc     */	ABSENT_STRING,
	/* 154: xoffc    */	ABSENT_STRING,
	/* 155: enacs    */	screen_s_enacs,
	/* 156: smln     */	ABSENT_STRING,
	/* 157: rmln     */	ABSENT_STRING,
	/* 158: kbeg     */	ABSENT_STRING,
	/* 159: kcan     */	ABSENT_STRING,
	/* 160: kclo     */	ABSENT_STRING,
	/* 161: kcmd     */	ABSENT_STRING,
	/* 162: kcpy     */	ABSENT_STRING,
	/* 163: kcrt     */	ABSENT_STRING,
	/* 164: kend     */	screen_s_kend,
	/* 165: kent     */	ABSENT_STRING,
	/* 166: kext     */	ABSENT_STRING,
	/* 167: kfnd     */	ABSENT_STRING,
	/* 168: khlp     */	ABSENT_STRING,
	/* 169: kmrk     */	ABSENT_STRING,
	/* 170: kmsg     */	ABSENT_STRING,
	/* 171: kmov     */	ABSENT_STRING,
	/* 172: knxt     */	ABSENT_STRING,
	/* 173: kopn     */	ABSENT_STRING,
	/* 174: kopt     */	ABSENT_STRING,
	/* 175: kprv     */	ABSENT_STRING,
	/* 176: kprt     */	ABSENT_STRING,
	/* 177: krdo     */	ABSENT_STRING,
	/* 178: kref     */	ABSENT_STRING,
	/* 179: krfr     */	ABSENT_STRING,
	/* 180: krpl     */	ABSENT_STRING,
	/* 181: krst     */	ABSENT_STRING,
	/* 182: kres     */	ABSENT_STRING,
	/* 183: ksav     */	ABSENT_STRING,
	/* 184: kspd     */	ABSENT_STRING,
	/* 185: kund     */	ABSENT_STRING,
	/* 186: kBEG     */	ABSENT_STRING,
	/* 187: kCAN     */	ABSENT_STRING,
	/* 188: kCMD     */	ABSENT_STRING,
	/* 189: kCPY     */	ABSENT_STRING,
	/* 190: kCRT     */	ABSENT_STRING,
	/* 191: kDC      */	ABSENT_STRING,
	/* 192: kDL      */	ABSENT_STRING,
	/* 193: kslt     */	ABSENT_STRING,
	/* 194: kEND     */	ABSENT_STRING,
	/* 195: kEOL     */	ABSENT_STRING,
	/* 196: kEXT     */	ABSENT_STRING,
	/* 197: kFND     */	ABSENT_STRING,
	/* 198: kHLP     */	ABSENT_STRING,
	/* 199: kHOM     */	ABSENT_STRING,
	/* 200: kIC      */	ABSENT_STRING,
	/* 201: kLFT     */	ABSENT_STRING,
	/* 202: kMSG     */	ABSENT_STRING,
	/* 203: kMOV     */	ABSENT_STRING,
	/* 204: kNXT     */	ABSENT_STRING,
	/* 205: kOPT     */	ABSENT_STRING,
	/* 206: kPRV     */	ABSENT_STRING,
	/* 207: kPRT     */	ABSENT_STRING,
	/* 208: kRDO     */	ABSENT_STRING,
	/* 209: kRPL     */	ABSENT_STRING,
	/* 210: kRIT     */	ABSENT_STRING,
	/* 211: kRES     */	ABSENT_STRING,
	/* 212: kSAV     */	ABSENT_STRING,
	/* 213: kSPD     */	ABSENT_STRING,
	/* 214: kUND     */	ABSENT_STRING,
	/* 215: rfi      */	ABSENT_STRING,
	/* 216: kf11     */	screen_s_kf11,
	/* 217: kf12     */	screen_s_kf12,
	/* 218: kf13     */	ABSENT_STRING,
	/* 219: kf14     */	ABSENT_STRING,
	/* 220: kf15     */	ABSENT_STRING,
	/* 221: kf16     */	ABSENT_STRING,
	/* 222: kf17     */	ABSENT_STRING,
	/* 223: kf18     */	ABSENT_STRING,
	/* 224: kf19     */	ABSENT_STRING,
	/* 225: kf20     */	ABSENT_STRING,
	/* 226: kf21     */	ABSENT_STRING,
	/* 227: kf22     */	ABSENT_STRING,
	/* 228: kf23     */	ABSENT_STRING,
	/* 229: kf24     */	ABSENT_STRING,
	/* 230: kf25     */	ABSENT_STRING,
	/* 231: kf26     */	ABSENT_STRING,
	/* 232: kf27     */	ABSENT_STRING,
	/* 233: kf28     */	ABSENT_STRING,
	/* 234: kf29     */	ABSENT_STRING,
	/* 235: kf30     */	ABSENT_STRING,
	/* 236: kf31     */	ABSENT_STRING,
	/* 237: kf32     */	ABSENT_STRING,
	/* 238: kf33     */	ABSENT_STRING,
	/* 239: kf34     */	ABSENT_STRING,
	/* 240: kf35     */	ABSENT_STRING,
	/* 241: kf36     */	ABSENT_STRING,
	/* 242: kf37     */	ABSENT_STRING,
	/* 243: kf38     */	ABSENT_STRING,
	/* 244: kf39     */	ABSENT_STRING,
	/* 245: kf40     */	ABSENT_STRING,
	/* 246: kf41     */	ABSENT_STRING,
	/* 247: kf42     */	ABSENT_STRING,
	/* 248: kf43     */	ABSENT_STRING,
	/* 249: kf44     */	ABSENT_STRING,
	/* 250: kf45     */	ABSENT_STRING,
	/* 251: kf46     */	ABSENT_STRING,
	/* 252: kf47     */	ABSENT_STRING,
	/* 253: kf48     */	ABSENT_STRING,
	/* 254: kf49     */	ABSENT_STRING,
	/* 255: kf50     */	ABSENT_STRING,
	/* 256: kf51     */	ABSENT_STRING,
	/* 257: kf52     */	ABSENT_STRING,
	/* 258: kf53     */	ABSENT_STRING,
	/* 259: kf54     */	ABSENT_STRING,
	/* 260: kf55     */	ABSENT_STRING,
	/* 261: kf56     */	ABSENT_STRING,
	/* 262: kf57     */	ABSENT_STRING,
	/* 263: kf58     */	ABSENT_STRING,
	/* 264: kf59     */	ABSENT_STRING,
	/* 265: kf60     */	ABSENT_STRING,
	/* 266: kf61     */	ABSENT_STRING,
	/* 267: kf62     */	ABSENT_STRING,
	/* 268: kf63     */	ABSENT_STRING,
	/* 269: el1      */	screen_s_el1,
	/* 270: mgc      */	ABSENT_STRING,
	/* 271: smgl     */	ABSENT_STRING,
	/* 272: smgr     */	ABSENT_STRING,
	/* 273: fln      */	ABSENT_STRING,
	/* 274: sclk     */	ABSENT_STRING,
	/* 275: dclk     */	ABSENT_STRING,
	/* 276: rmclk    */	ABSENT_STRING,
	/* 277: cwin     */	ABSENT_STRING,
	/* 278: wingo    */	ABSENT_STRING,
	/* 279: hup      */	ABSENT_STRING,
	/* 280: dial     */	ABSENT_STRING,
	/* 281: qdial    */	ABSENT_STRING,
	/* 282: tone     */	ABSENT_STRING,
	/* 283: pulse    */	ABSENT_STRING,
	/* 284: hook     */	ABSENT_STRING,
	/* 285: pause    */	ABSENT_STRING,
	/* 286: wait     */	ABSENT_STRING,
	/* 287: u0       */	ABSENT_STRING,
	/* 288: u1       */	ABSENT_STRING,
	/* 289: u2       */	ABSENT_STRING,
	/* 290: u3       */	ABSENT_STRING,
	/* 291: u4       */	ABSENT_STRING,
	/* 292: u5       */	ABSENT_STRING,
	/* 293: u6       */	ABSENT_STRING,
	/* 294: u7       */	ABSENT_STRING,
	/* 295: u8       */	ABSENT_STRING,
	/* 296: u9       */	ABSENT_STRING,
	/* 297: op       */	screen_s_op,
	/* 298: oc       */	ABSENT_STRING,
	/* 299: initc    */	ABSENT_STRING,
	/* 300: initp    */	ABSENT_STRING,
	/* 301: scp      */	ABSENT_STRING,
	/* 302: setf     */	ABSENT_STRING,
	/* 303: setb     */	ABSENT_STRING,
	/* 304: cpi      */	ABSENT_STRING,
	/* 305: lpi      */	ABSENT_STRING,
	/* 306: chr      */	ABSENT_STRING,
	/* 307: cvr      */	ABSENT_STRING,
	/* 308: defc     */	ABSENT_STRING,
	/* 309: swidm    */	ABSENT_STRING,
	/* 310: sdrfq    */	ABSENT_STRING,
	/* 311: sitm     */	ABSENT_STRING,
	/* 312: slm      */	ABSENT_STRING,
	/* 313: smicm    */	ABSENT_STRING,
	/* 314: snlq     */	ABSENT_STRING,
	/* 315: snrmq    */	ABSENT_STRING,
	/* 316: sshm     */	ABSENT_STRING,
	/* 317: ssubm    */	ABSENT_STRING,
	/* 318: ssupm    */	ABSENT_STRING,
	/* 319: sum      */	ABSENT_STRING,
	/* 320: rwidm    */	ABSENT_STRING,
	/* 321: ritm     */	ABSENT_STRING,
	/* 322: rlm      */	ABSENT_STRING,
	/* 323: rmicm    */	ABSENT_STRING,
	/* 324: rshm     */	ABSENT_STRING,
	/* 325: rsubm    */	ABSENT_STRING,
	/* 326: rsupm    */	ABSENT_STRING,
	/* 327: rum      */	ABSENT_STRING,
	/* 328: mhpa     */	ABSENT_STRING,
	/* 329: mcud1    */	ABSENT_STRING,
	/* 330: mcub1    */	ABSENT_STRING,
	/* 331: mcuf1    */	ABSENT_STRING,
	/* 332: mvpa     */	ABSENT_STRING,
	/* 333: mcuu1    */	ABSENT_STRING,
	/* 334: porder   */	ABSENT_STRING,
	/* 335: mcud     */	ABSENT_STRING,
	/* 336: mcub     */	ABSENT_STRING,
	/* 337: mcuf     */	ABSENT_STRING,
	/* 338: mcuu     */	ABSENT_STRING,
	/* 339: scs      */	ABSENT_STRING,
	/* 340: smgb     */	ABSENT_STRING,
	/* 341: smgbp    */	ABSENT_STRING,
	/* 342: smglp    */	ABSENT_STRING,
	/* 343: smgrp    */	ABSENT_STRING,
	/* 344: smgt     */	ABSENT_STRING,
	/* 345: smgtp    */	ABSENT_STRING,
	/* 346: sbim     */	ABSENT_STRING,
	/* 347: scsd     */	ABSENT_STRING,
	/* 348: rbim     */	ABSENT_STRING,
	/* 349: rcsd     */	ABSENT_STRING,
	/* 350: subcs    */	ABSENT_STRING,
	/* 351: supcs    */	ABSENT_STRING,
	/* 352: docr     */	ABSENT_STRING,
	/* 353: zerom    */	ABSENT_STRING,
	/* 354: csnm     */	ABSENT_STRING,
	/* 355: kmous    */	screen_s_kmous,
	/* 356: minfo    */	ABSENT_STRING,
	/* 357: reqmp    */	ABSENT_STRING,
	/* 358: getm     */	ABSENT_STRING,
	/* 359: setaf    */	screen_s_setaf,
	/* 360: setab    */	screen_s_setab,
	/* 361: pfxl     */	ABSENT_STRING,
	/* 362: devt     */	ABSENT_STRING,
	/* 363: csin     */	ABSENT_STRING,
	/* 364: s0ds     */	ABSENT_STRING,
	/* 365: s1ds     */	ABSENT_STRING,
	/* 366: s2ds     */	ABSENT_STRING,
	/* 367: s3ds     */	ABSENT_STRING,
	/* 368: smglr    */	ABSENT_STRING,
	/* 369: smgtb    */	ABSENT_STRING,
	/* 370: birep    */	ABSENT_STRING,
	/* 371: binel    */	ABSENT_STRING,
	/* 372: bicr     */	ABSENT_STRING,
	/* 373: colornm  */	ABSENT_STRING,
	/* 374: defbi    */	ABSENT_STRING,
	/* 375: endbi    */	ABSENT_STRING,
	/* 376: setcolor */	ABSENT_STRING,
	/* 377: slines   */	ABSENT_STRING,
	/* 378: dispc    */	ABSENT_STRING,
	/* 379: smpch    */	ABSENT_STRING,
	/* 380: rmpch    */	ABSENT_STRING,
	/* 381: smsc     */	ABSENT_STRING,
	/* 382: rmsc     */	ABSENT_STRING,
	/* 383: pctrm    */	ABSENT_STRING,
	/* 384: scesc    */	ABSENT_STRING,
	/* 385: scesa    */	ABSENT_STRING,
	/* 386: ehhlm    */	ABSENT_STRING,
	/* 387: elhlm    */	ABSENT_STRING,
	/* 388: elohlm   */	ABSENT_STRING,
	/* 389: erhlm    */	ABSENT_STRING,
	/* 390: ethlm    */	ABSENT_STRING,
	/* 391: evhlm    */	ABSENT_STRING,
	/* 392: sgr1     */	ABSENT_STRING,
	/* 393: slength  */	ABSENT_STRING,
	/* 394: OTi2     */	ABSENT_STRING,
	/* 395: OTrs     */	ABSENT_STRING,
	/* 396: OTnl     */	ABSENT_STRING,
	/* 397: OTbc     */	ABSENT_STRING,
	/* 398: OTko     */	ABSENT_STRING,
	/* 399: OTma     */	ABSENT_STRING,
	/* 400: OTG2     */	ABSENT_STRING,
	/* 401: OTG3     */	ABSENT_STRING,
	/* 402: OTG1     */	ABSENT_STRING,
	/* 403: OTG4     */	ABSENT_STRING,
	/* 404: OTGR     */	ABSENT_STRING,
	/* 405: OTGL     */	ABSENT_STRING,
	/* 406: OTGU     */	ABSENT_STRING,
	/* 407: OTGD     */	ABSENT_STRING,
	/* 408: OTGH     */	ABSENT_STRING,
	/* 409: OTGV     */	ABSENT_STRING,
	/* 410: OTGC     */	ABSENT_STRING,
	/* 411: meml     */	ABSENT_STRING,
	/* 412: memu     */	ABSENT_STRING,
	/* 413: box1     */	ABSENT_STRING,
};
/* screen.linux */

static char screen_linux_alias_data[] = "screen.linux|screen in linux console";

static char screen_linux_s_cbt  [] = "\033[Z";
static char screen_linux_s_bel  [] = "\007";
static char screen_linux_s_cr   [] = "\015";
static char screen_linux_s_csr  [] = "\033[%i%p1%d;%p2%dr";
static char screen_linux_s_tbc  [] = "\033[3g";
static char screen_linux_s_clear[] = "\033[H\033[J";
static char screen_linux_s_el   [] = "\033[K";
static char screen_linux_s_ed   [] = "\033[J";
static char screen_linux_s_hpa  [] = "\033[%i%p1%dG";
static char screen_linux_s_cup  [] = "\033[%i%p1%d;%p2%dH";
static char screen_linux_s_cud1 [] = "\012";
static char screen_linux_s_home [] = "\033[H";
static char screen_linux_s_civis[] = "\033[?25l";
static char screen_linux_s_cub1 [] = "\010";
static char screen_linux_s_cnorm[] = "\033[34h\033[?25h";
static char screen_linux_s_cuf1 [] = "\033[C";
static char screen_linux_s_cuu1 [] = "\033M";
static char screen_linux_s_cvvis[] = "\033[34l";
static char screen_linux_s_dch1 [] = "\033[P";
static char screen_linux_s_dl1  [] = "\033[M";
static char screen_linux_s_smacs[] = "\016";
static char screen_linux_s_blink[] = "\033[5m";
static char screen_linux_s_bold [] = "\033[1m";
static char screen_linux_s_smcup[] = "\033[?1049h";
static char screen_linux_s_dim  [] = "\033[2m";
static char screen_linux_s_smir [] = "\033[4h";
static char screen_linux_s_rev  [] = "\033[7m";
static char screen_linux_s_smso [] = "\033[3m";
static char screen_linux_s_smul [] = "\033[4m";
static char screen_linux_s_rmacs[] = "\017";
static char screen_linux_s_sgr0 [] = "\033[m\017";
static char screen_linux_s_rmcup[] = "\033[?1049l";
static char screen_linux_s_rmir [] = "\033[4l";
static char screen_linux_s_rmso [] = "\033[23m";
static char screen_linux_s_rmul [] = "\033[24m";
static char screen_linux_s_flash[] = "\033g";
static char screen_linux_s_is2  [] = "\033)0";
static char screen_linux_s_il1  [] = "\033[L";
static char screen_linux_s_kbs  [] = "\177";
static char screen_linux_s_kdch1[] = "\033[3~";
static char screen_linux_s_kcud1[] = "\033OB";
static char screen_linux_s_kf1  [] = "\033OP";
static char screen_linux_s_kf10 [] = "\033[21~";
static char screen_linux_s_kf2  [] = "\033OQ";
static char screen_linux_s_kf3  [] = "\033OR";
static char screen_linux_s_kf4  [] = "\033OS";
static char screen_linux_s_kf5  [] = "\033[15~";
static char screen_linux_s_kf6  [] = "\033[17~";
static char screen_linux_s_kf7  [] = "\033[18~";
static char screen_linux_s_kf8  [] = "\033[19~";
static char screen_linux_s_kf9  [] = "\033[20~";
static char screen_linux_s_khome[] = "\033[1~";
static char screen_linux_s_kich1[] = "\033[2~";
static char screen_linux_s_kcub1[] = "\033OD";
static char screen_linux_s_knp  [] = "\033[6~";
static char screen_linux_s_kpp  [] = "\033[5~";
static char screen_linux_s_kcuf1[] = "\033OC";
static char screen_linux_s_kcuu1[] = "\033OA";
static char screen_linux_s_rmkx [] = "\033[?1l\033>";
static char screen_linux_s_smkx [] = "\033[?1h\033=";
static char screen_linux_s_nel  [] = "\033E";
static char screen_linux_s_dch  [] = "\033[%p1%dP";
static char screen_linux_s_dl   [] = "\033[%p1%dM";
static char screen_linux_s_cud  [] = "\033[%p1%dB";
static char screen_linux_s_ich  [] = "\033[%p1%d@";
static char screen_linux_s_indn [] = "\033[%p1%dS";
static char screen_linux_s_il   [] = "\033[%p1%dL";
static char screen_linux_s_cub  [] = "\033[%p1%dD";
static char screen_linux_s_cuf  [] = "\033[%p1%dC";
static char screen_linux_s_cuu  [] = "\033[%p1%dA";
static char screen_linux_s_rs2  [] = "\033c\033[?1000l\033[?25h";
static char screen_linux_s_rc   [] = "\0338";
static char screen_linux_s_vpa  [] = "\033[%i%p1%dd";
static char screen_linux_s_sc   [] = "\0337";
static char screen_linux_s_ind  [] = "\012";
static char screen_linux_s_ri   [] = "\033M";
static char screen_linux_s_sgr  [] = "\033[0%?%p6%t;1%;%?%p1%t;3%;%?%p2%t;4%;%?%p3%t;7%;%?%p4%t;5%;%?%p5%t;2%;m%?%p9%t\016%e\017%;";
static char screen_linux_s_hts  [] = "\033H";
static char screen_linux_s_ht   [] = "\011";
static char screen_linux_s_acsc [] = "++,,--..00``aaffgghhiijjkkllmmnnooppqqrrssttuuvvwwxxyyzz{{||}}~~";
static char screen_linux_s_enacs[] = "\033(B\033)0";
static char screen_linux_s_kend [] = "\033[4~";
static char screen_linux_s_kf11 [] = "\033[23~";
static char screen_linux_s_kf12 [] = "\033[24~";
static char screen_linux_s_el1  [] = "\033[1K";
static char screen_linux_s_op   [] = "\033[39;49m";
static char screen_linux_s_kmous[] = "\033[M";
static char screen_linux_s_setaf[] = "\033[3%p1%dm";
static char screen_linux_s_setab[] = "\033[4%p1%dm";

static char screen_linux_bool_data[] = {
	/*   0: bw       */	TRUE,
	/*   1: am       */	TRUE,
	/*   2: xsb      */	FALSE,
	/*   3: xhp      */	FALSE,
	/*   4: xenl     */	TRUE,
	/*   5: eo       */	FALSE,
	/*   6: gn       */	FALSE,
	/*   7: hc       */	FALSE,
	/*   8: km       */	TRUE,
	/*   9: hs       */	FALSE,
	/*  10: in       */	FALSE,
	/*  11: da       */	FALSE,
	/*  12: db       */	FALSE,
	/*  13: mir      */	TRUE,
	/*  14: msgr     */	TRUE,
	/*  15: os       */	FALSE,
	/*  16: eslok    */	FALSE,
	/*  17: xt       */	FALSE,
	/*  18: hz       */	FALSE,
	/*  19: ul       */	FALSE,
	/*  20: xon      */	FALSE,
	/*  21: nxon     */	FALSE,
	/*  22: mc5i     */	FALSE,
	/*  23: chts     */	FALSE,
	/*  24: nrrmc    */	FALSE,
	/*  25: npc      */	FALSE,
	/*  26: ndscr    */	FALSE,
	/*  27: ccc      */	FALSE,
	/*  28: bce      */	FALSE,
	/*  29: hls      */	FALSE,
	/*  30: xhpa     */	FALSE,
	/*  31: crxm     */	FALSE,
	/*  32: daisy    */	FALSE,
	/*  33: xvpa     */	FALSE,
	/*  34: sam      */	FALSE,
	/*  35: cpix     */	FALSE,
	/*  36: lpix     */	FALSE,
	/*  37: OTbs     */	TRUE,
	/*  38: OTns     */	FALSE,
	/*  39: OTnc     */	FALSE,
	/*  40: OTMT     */	FALSE,
	/*  41: OTNL     */	FALSE,
	/*  42: OTpt     */	TRUE,
	/*  43: OTxr     */	FALSE,
};
static NCURSES_INT2 screen_linux_number_data[] = {
	/*   0: cols     */	80,
	/*   1: it       */	8,
	/*   2: lines    */	24,
	/*   3: lm       */	ABSENT_NUMERIC,
	/*   4: xmc      */	ABSENT_NUMERIC,
	/*   5: pb       */	ABSENT_NUMERIC,
	/*   6: vt       */	ABSENT_NUMERIC,
	/*   7: wsl      */	ABSENT_NUMERIC,
	/*   8: nlab     */	ABSENT_NUMERIC,
	/*   9: lh       */	ABSENT_NUMERIC,
	/*  10: lw       */	ABSENT_NUMERIC,
	/*  11: ma       */	ABSENT_NUMERIC,
	/*  12: wnum     */	ABSENT_NUMERIC,
	/*  13: colors   */	8,
	/*  14: pairs    */	64,
	/*  15: ncv      */	ABSENT_NUMERIC,
	/*  16: bufsz    */	ABSENT_NUMERIC,
	/*  17: spinv    */	ABSENT_NUMERIC,
	/*  18: spinh    */	ABSENT_NUMERIC,
	/*  19: maddr    */	ABSENT_NUMERIC,
	/*  20: mjump    */	ABSENT_NUMERIC,
	/*  21: mcs      */	ABSENT_NUMERIC,
	/*  22: mls      */	ABSENT_NUMERIC,
	/*  23: npins    */	ABSENT_NUMERIC,
	/*  24: orc      */	ABSENT_NUMERIC,
	/*  25: orl      */	ABSENT_NUMERIC,
	/*  26: orhi     */	ABSENT_NUMERIC,
	/*  27: orvi     */	ABSENT_NUMERIC,
	/*  28: cps      */	ABSENT_NUMERIC,
	/*  29: widcs    */	ABSENT_NUMERIC,
	/*  30: btns     */	ABSENT_NUMERIC,
	/*  31: bitwin   */	ABSENT_NUMERIC,
	/*  32: bitype   */	ABSENT_NUMERIC,
	/*  33: OTug     */	ABSENT_NUMERIC,
	/*  34: OTdC     */	ABSENT_NUMERIC,
	/*  35: OTdN     */	ABSENT_NUMERIC,
	/*  36: OTdB     */	ABSENT_NUMERIC,
	/*  37: OTdT     */	ABSENT_NUMERIC,
	/*  38: OTkn     */	ABSENT_NUMERIC,
};
static char * screen_linux_string_data[] = {
	/*   0: cbt      */	screen_linux_s_cbt,
	/*   1: bel      */	screen_linux_s_bel,
	/*   2: cr       */	screen_linux_s_cr,
	/*   3: csr      */	screen_linux_s_csr,
	/*   4: tbc      */	screen_linux_s_tbc,
	/*   5: clear    */	screen_linux_s_clear,
	/*   6: el       */	screen_linux_s_el,
	/*   7: ed       */	screen_linux_s_ed,
	/*   8: hpa      */	screen_linux_s_hpa,
	/*   9: cmdch    */	ABSENT_STRING,
	/*  10: cup      */	screen_linux_s_cup,
	/*  11: cud1     */	screen_linux_s_cud1,
	/*  12: home     */	screen_linux_s_home,
	/*  13: civis    */	screen_linux_s_civis,
	/*  14: cub1     */	screen_linux_s_cub1,
	/*  15: mrcup    */	ABSENT_STRING,
	/*  16: cnorm    */	screen_linux_s_cnorm,
	/*  17: cuf1     */	screen_linux_s_cuf1,
	/*  18: ll       */	ABSENT_STRING,
	/*  19: cuu1     */	screen_linux_s_cuu1,
	/*  20: cvvis    */	screen_linux_s_cvvis,
	/*  21: dch1     */	screen_linux_s_dch1,
	/*  22: dl1      */	screen_linux_s_dl1,
	/*  23: dsl      */	ABSENT_STRING,
	/*  24: hd       */	ABSENT_STRING,
	/*  25: smacs    */	screen_linux_s_smacs,
	/*  26: blink    */	screen_linux_s_blink,
	/*  27: bold     */	screen_linux_s_bold,
	/*  28: smcup    */	screen_linux_s_smcup,
	/*  29: smdc     */	ABSENT_STRING,
	/*  30: dim      */	screen_linux_s_dim,
	/*  31: smir     */	screen_linux_s_smir,
	/*  32: invis    */	ABSENT_STRING,
	/*  33: prot     */	ABSENT_STRING,
	/*  34: rev      */	screen_linux_s_rev,
	/*  35: smso     */	screen_linux_s_smso,
	/*  36: smul     */	screen_linux_s_smul,
	/*  37: ech      */	ABSENT_STRING,
	/*  38: rmacs    */	screen_linux_s_rmacs,
	/*  39: sgr0     */	screen_linux_s_sgr0,
	/*  40: rmcup    */	screen_linux_s_rmcup,
	/*  41: rmdc     */	ABSENT_STRING,
	/*  42: rmir     */	screen_linux_s_rmir,
	/*  43: rmso     */	screen_linux_s_rmso,
	/*  44: rmul     */	screen_linux_s_rmul,
	/*  45: flash    */	screen_linux_s_flash,
	/*  46: ff       */	ABSENT_STRING,
	/*  47: fsl      */	ABSENT_STRING,
	/*  48: is1      */	ABSENT_STRING,
	/*  49: is2      */	screen_linux_s_is2,
	/*  50: is3      */	ABSENT_STRING,
	/*  51: if       */	ABSENT_STRING,
	/*  52: ich1     */	ABSENT_STRING,
	/*  53: il1      */	screen_linux_s_il1,
	/*  54: ip       */	ABSENT_STRING,
	/*  55: kbs      */	screen_linux_s_kbs,
	/*  56: ktbc     */	ABSENT_STRING,
	/*  57: kclr     */	ABSENT_STRING,
	/*  58: kctab    */	ABSENT_STRING,
	/*  59: kdch1    */	screen_linux_s_kdch1,
	/*  60: kdl1     */	ABSENT_STRING,
	/*  61: kcud1    */	screen_linux_s_kcud1,
	/*  62: krmir    */	ABSENT_STRING,
	/*  63: kel      */	ABSENT_STRING,
	/*  64: ked      */	ABSENT_STRING,
	/*  65: kf0      */	ABSENT_STRING,
	/*  66: kf1      */	screen_linux_s_kf1,
	/*  67: kf10     */	screen_linux_s_kf10,
	/*  68: kf2      */	screen_linux_s_kf2,
	/*  69: kf3      */	screen_linux_s_kf3,
	/*  70: kf4      */	screen_linux_s_kf4,
	/*  71: kf5      */	screen_linux_s_kf5,
	/*  72: kf6      */	screen_linux_s_kf6,
	/*  73: kf7      */	screen_linux_s_kf7,
	/*  74: kf8      */	screen_linux_s_kf8,
	/*  75: kf9      */	screen_linux_s_kf9,
	/*  76: khome    */	screen_linux_s_khome,
	/*  77: kich1    */	screen_linux_s_kich1,
	/*  78: kil1     */	ABSENT_STRING,
	/*  79: kcub1    */	screen_linux_s_kcub1,
	/*  80: kll      */	ABSENT_STRING,
	/*  81: knp      */	screen_linux_s_knp,
	/*  82: kpp      */	screen_linux_s_kpp,
	/*  83: kcuf1    */	screen_linux_s_kcuf1,
	/*  84: kind     */	ABSENT_STRING,
	/*  85: kri      */	ABSENT_STRING,
	/*  86: khts     */	ABSENT_STRING,
	/*  87: kcuu1    */	screen_linux_s_kcuu1,
	/*  88: rmkx     */	screen_linux_s_rmkx,
	/*  89: smkx     */	screen_linux_s_smkx,
	/*  90: lf0      */	ABSENT_STRING,
	/*  91: lf1      */	ABSENT_STRING,
	/*  92: lf10     */	ABSENT_STRING,
	/*  93: lf2      */	ABSENT_STRING,
	/*  94: lf3      */	ABSENT_STRING,
	/*  95: lf4      */	ABSENT_STRING,
	/*  96: lf5      */	ABSENT_STRING,
	/*  97: lf6      */	ABSENT_STRING,
	/*  98: lf7      */	ABSENT_STRING,
	/*  99: lf8      */	ABSENT_STRING,
	/* 100: lf9      */	ABSENT_STRING,
	/* 101: rmm      */	ABSENT_STRING,
	/* 102: smm      */	ABSENT_STRING,
	/* 103: nel      */	screen_linux_s_nel,
	/* 104: pad      */	ABSENT_STRING,
	/* 105: dch      */	screen_linux_s_dch,
	/* 106: dl       */	screen_linux_s_dl,
	/* 107: cud      */	screen_linux_s_cud,
	/* 108: ich      */	screen_linux_s_ich,
	/* 109: indn     */	screen_linux_s_indn,
	/* 110: il       */	screen_linux_s_il,
	/* 111: cub      */	screen_linux_s_cub,
	/* 112: cuf      */	screen_linux_s_cuf,
	/* 113: rin      */	ABSENT_STRING,
	/* 114: cuu      */	screen_linux_s_cuu,
	/* 115: pfkey    */	ABSENT_STRING,
	/* 116: pfloc    */	ABSENT_STRING,
	/* 117: pfx      */	ABSENT_STRING,
	/* 118: mc0      */	ABSENT_STRING,
	/* 119: mc4      */	ABSENT_STRING,
	/* 120: mc5      */	ABSENT_STRING,
	/* 121: rep      */	ABSENT_STRING,
	/* 122: rs1      */	ABSENT_STRING,
	/* 123: rs2      */	screen_linux_s_rs2,
	/* 124: rs3      */	ABSENT_STRING,
	/* 125: rf       */	ABSENT_STRING,
	/* 126: rc       */	screen_linux_s_rc,
	/* 127: vpa      */	screen_linux_s_vpa,
	/* 128: sc       */	screen_linux_s_sc,
	/* 129: ind      */	screen_linux_s_ind,
	/* 130: ri       */	screen_linux_s_ri,
	/* 131: sgr      */	screen_linux_s_sgr,
	/* 132: hts      */	screen_linux_s_hts,
	/* 133: wind     */	ABSENT_STRING,
	/* 134: ht       */	screen_linux_s_ht,
	/* 135: tsl      */	ABSENT_STRING,
	/* 136: uc       */	ABSENT_STRING,
	/* 137: hu       */	ABSENT_STRING,
	/* 138: iprog    */	ABSENT_STRING,
	/* 139: ka1      */	ABSENT_STRING,
	/* 140: ka3      */	ABSENT_STRING,
	/* 141: kb2      */	ABSENT_STRING,
	/* 142: kc1      */	ABSENT_STRING,
	/* 143: kc3      */	ABSENT_STRING,
	/* 144: mc5p     */	ABSENT_STRING,
	/* 145: rmp      */	ABSENT_STRING,
	/* 146: acsc     */	screen_linux_s_acsc,
	/* 147: pln      */	ABSENT_STRING,
	/* 148: kcbt     */	CANCELLED_STRING,
	/* 149: smxon    */	ABSENT_STRING,
	/* 150: rmxon    */	ABSENT_STRING,
	/* 151: smam     */	ABSENT_STRING,
	/* 152: rmam     */	ABSENT_STRING,
	/* 153: xonc     */	ABSENT_STRING,
	/* 154: xoffc    */	ABSENT_STRING,
	/* 155: enacs    */	screen_linux_s_enacs,
	/* 156: smln     */	ABSENT_STRING,
	/* 157: rmln     */	ABSENT_STRING,
	/* 158: kbeg     */	ABSENT_STRING,
	/* 159: kcan     */	ABSENT_STRING,
	/* 160: kclo     */	ABSENT_STRING,
	/* 161: kcmd     */	ABSENT_STRING,
	/* 162: kcpy     */	ABSENT_STRING,
	/* 163: kcrt     */	ABSENT_STRING,
	/* 164: kend     */	screen_linux_s_kend,
	/* 165: kent     */	ABSENT_STRING,
	/* 166: kext     */	ABSENT_STRING,
	/* 167: kfnd     */	ABSENT_STRING,
	/* 168: khlp     */	ABSENT_STRING,
	/* 169: kmrk     */	ABSENT_STRING,
	/* 170: kmsg     */	ABSENT_STRING,
	/* 171: kmov     */	ABSENT_STRING,
	/* 172: knxt     */	ABSENT_STRING,
	/* 173: kopn     */	ABSENT_STRING,
	/* 174: kopt     */	ABSENT_STRING,
	/* 175: kprv     */	ABSENT_STRING,
	/* 176: kprt     */	ABSENT_STRING,
	/* 177: krdo     */	ABSENT_STRING,
	/* 178: kref     */	ABSENT_STRING,
	/* 179: krfr     */	ABSENT_STRING,
	/* 180: krpl     */	ABSENT_STRING,
	/* 181: krst     */	ABSENT_STRING,
	/* 182: kres     */	ABSENT_STRING,
	/* 183: ksav     */	ABSENT_STRING,
	/* 184: kspd     */	ABSENT_STRING,
	/* 185: kund     */	ABSENT_STRING,
	/* 186: kBEG     */	ABSENT_STRING,
	/* 187: kCAN     */	ABSENT_STRING,
	/* 188: kCMD     */	ABSENT_STRING,
	/* 189: kCPY     */	ABSENT_STRING,
	/* 190: kCRT     */	ABSENT_STRING,
	/* 191: kDC      */	ABSENT_STRING,
	/* 192: kDL      */	ABSENT_STRING,
	/* 193: kslt     */	ABSENT_STRING,
	/* 194: kEND     */	ABSENT_STRING,
	/* 195: kEOL     */	ABSENT_STRING,
	/* 196: kEXT     */	ABSENT_STRING,
	/* 197: kFND     */	ABSENT_STRING,
	/* 198: kHLP     */	ABSENT_STRING,
	/* 199: kHOM     */	ABSENT_STRING,
	/* 200: kIC      */	ABSENT_STRING,
	/* 201: kLFT     */	ABSENT_STRING,
	/* 202: kMSG     */	ABSENT_STRING,
	/* 203: kMOV     */	ABSENT_STRING,
	/* 204: kNXT     */	ABSENT_STRING,
	/* 205: kOPT     */	ABSENT_STRING,
	/* 206: kPRV     */	ABSENT_STRING,
	/* 207: kPRT     */	ABSENT_STRING,
	/* 208: kRDO     */	ABSENT_STRING,
	/* 209: kRPL     */	ABSENT_STRING,
	/* 210: kRIT     */	ABSENT_STRING,
	/* 211: kRES     */	ABSENT_STRING,
	/* 212: kSAV     */	ABSENT_STRING,
	/* 213: kSPD     */	ABSENT_STRING,
	/* 214: kUND     */	ABSENT_STRING,
	/* 215: rfi      */	ABSENT_STRING,
	/* 216: kf11     */	screen_linux_s_kf11,
	/* 217: kf12     */	screen_linux_s_kf12,
	/* 218: kf13     */	ABSENT_STRING,
	/* 219: kf14     */	ABSENT_STRING,
	/* 220: kf15     */	ABSENT_STRING,
	/* 221: kf16     */	ABSENT_STRING,
	/* 222: kf17     */	ABSENT_STRING,
	/* 223: kf18     */	ABSENT_STRING,
	/* 224: kf19     */	ABSENT_STRING,
	/* 225: kf20     */	ABSENT_STRING,
	/* 226: kf21     */	ABSENT_STRING,
	/* 227: kf22     */	ABSENT_STRING,
	/* 228: kf23     */	ABSENT_STRING,
	/* 229: kf24     */	ABSENT_STRING,
	/* 230: kf25     */	ABSENT_STRING,
	/* 231: kf26     */	ABSENT_STRING,
	/* 232: kf27     */	ABSENT_STRING,
	/* 233: kf28     */	ABSENT_STRING,
	/* 234: kf29     */	ABSENT_STRING,
	/* 235: kf30     */	ABSENT_STRING,
	/* 236: kf31     */	ABSENT_STRING,
	/* 237: kf32     */	ABSENT_STRING,
	/* 238: kf33     */	ABSENT_STRING,
	/* 239: kf34     */	ABSENT_STRING,
	/* 240: kf35     */	ABSENT_STRING,
	/* 241: kf36     */	ABSENT_STRING,
	/* 242: kf37     */	ABSENT_STRING,
	/* 243: kf38     */	ABSENT_STRING,
	/* 244: kf39     */	ABSENT_STRING,
	/* 245: kf40     */	ABSENT_STRING,
	/* 246: kf41     */	ABSENT_STRING,
	/* 247: kf42     */	ABSENT_STRING,
	/* 248: kf43     */	ABSENT_STRING,
	/* 249: kf44     */	ABSENT_STRING,
	/* 250: kf45     */	ABSENT_STRING,
	/* 251: kf46     */	ABSENT_STRING,
	/* 252: kf47     */	ABSENT_STRING,
	/* 253: kf48     */	ABSENT_STRING,
	/* 254: kf49     */	ABSENT_STRING,
	/* 255: kf50     */	ABSENT_STRING,
	/* 256: kf51     */	ABSENT_STRING,
	/* 257: kf52     */	ABSENT_STRING,
	/* 258: kf53     */	ABSENT_STRING,
	/* 259: kf54     */	ABSENT_STRING,
	/* 260: kf55     */	ABSENT_STRING,
	/* 261: kf56     */	ABSENT_STRING,
	/* 262: kf57     */	ABSENT_STRING,
	/* 263: kf58     */	ABSENT_STRING,
	/* 264: kf59     */	ABSENT_STRING,
	/* 265: kf60     */	ABSENT_STRING,
	/* 266: kf61     */	ABSENT_STRING,
	/* 267: kf62     */	ABSENT_STRING,
	/* 268: kf63     */	ABSENT_STRING,
	/* 269: el1      */	screen_linux_s_el1,
	/* 270: mgc      */	ABSENT_STRING,
	/* 271: smgl     */	ABSENT_STRING,
	/* 272: smgr     */	ABSENT_STRING,
	/* 273: fln      */	ABSENT_STRING,
	/* 274: sclk     */	ABSENT_STRING,
	/* 275: dclk     */	ABSENT_STRING,
	/* 276: rmclk    */	ABSENT_STRING,
	/* 277: cwin     */	ABSENT_STRING,
	/* 278: wingo    */	ABSENT_STRING,
	/* 279: hup      */	ABSENT_STRING,
	/* 280: dial     */	ABSENT_STRING,
	/* 281: qdial    */	ABSENT_STRING,
	/* 282: tone     */	ABSENT_STRING,
	/* 283: pulse    */	ABSENT_STRING,
	/* 284: hook     */	ABSENT_STRING,
	/* 285: pause    */	ABSENT_STRING,
	/* 286: wait     */	ABSENT_STRING,
	/* 287: u0       */	ABSENT_STRING,
	/* 288: u1       */	ABSENT_STRING,
	/* 289: u2       */	ABSENT_STRING,
	/* 290: u3       */	ABSENT_STRING,
	/* 291: u4       */	ABSENT_STRING,
	/* 292: u5       */	ABSENT_STRING,
	/* 293: u6       */	ABSENT_STRING,
	/* 294: u7       */	ABSENT_STRING,
	/* 295: u8       */	ABSENT_STRING,
	/* 296: u9       */	ABSENT_STRING,
	/* 297: op       */	screen_linux_s_op,
	/* 298: oc       */	ABSENT_STRING,
	/* 299: initc    */	ABSENT_STRING,
	/* 300: initp    */	ABSENT_STRING,
	/* 301: scp      */	ABSENT_STRING,
	/* 302: setf     */	ABSENT_STRING,
	/* 303: setb     */	ABSENT_STRING,
	/* 304: cpi      */	ABSENT_STRING,
	/* 305: lpi      */	ABSENT_STRING,
	/* 306: chr      */	ABSENT_STRING,
	/* 307: cvr      */	ABSENT_STRING,
	/* 308: defc     */	ABSENT_STRING,
	/* 309: swidm    */	ABSENT_STRING,
	/* 310: sdrfq    */	ABSENT_STRING,
	/* 311: sitm     */	ABSENT_STRING,
	/* 312: slm      */	ABSENT_STRING,
	/* 313: smicm    */	ABSENT_STRING,
	/* 314: snlq     */	ABSENT_STRING,
	/* 315: snrmq    */	ABSENT_STRING,
	/* 316: sshm     */	ABSENT_STRING,
	/* 317: ssubm    */	ABSENT_STRING,
	/* 318: ssupm    */	ABSENT_STRING,
	/* 319: sum      */	ABSENT_STRING,
	/* 320: rwidm    */	ABSENT_STRING,
	/* 321: ritm     */	ABSENT_STRING,
	/* 322: rlm      */	ABSENT_STRING,
	/* 323: rmicm    */	ABSENT_STRING,
	/* 324: rshm     */	ABSENT_STRING,
	/* 325: rsubm    */	ABSENT_STRING,
	/* 326: rsupm    */	ABSENT_STRING,
	/* 327: rum      */	ABSENT_STRING,
	/* 328: mhpa     */	ABSENT_STRING,
	/* 329: mcud1    */	ABSENT_STRING,
	/* 330: mcub1    */	ABSENT_STRING,
	/* 331: mcuf1    */	ABSENT_STRING,
	/* 332: mvpa     */	ABSENT_STRING,
	/* 333: mcuu1    */	ABSENT_STRING,
	/* 334: porder   */	ABSENT_STRING,
	/* 335: mcud     */	ABSENT_STRING,
	/* 336: mcub     */	ABSENT_STRING,
	/* 337: mcuf     */	ABSENT_STRING,
	/* 338: mcuu     */	ABSENT_STRING,
	/* 339: scs      */	ABSENT_STRING,
	/* 340: smgb     */	ABSENT_STRING,
	/* 341: smgbp    */	ABSENT_STRING,
	/* 342: smglp    */	ABSENT_STRING,
	/* 343: smgrp    */	ABSENT_STRING,
	/* 344: smgt     */	ABSENT_STRING,
	/* 345: smgtp    */	ABSENT_STRING,
	/* 346: sbim     */	ABSENT_STRING,
	/* 347: scsd     */	ABSENT_STRING,
	/* 348: rbim     */	ABSENT_STRING,
	/* 349: rcsd     */	ABSENT_STRING,
	/* 350: subcs    */	ABSENT_STRING,
	/* 351: supcs    */	ABSENT_STRING,
	/* 352: docr     */	ABSENT_STRING,
	/* 353: zerom    */	ABSENT_STRING,
	/* 354: csnm     */	ABSENT_STRING,
	/* 355: kmous    */	screen_linux_s_kmous,
	/* 356: minfo    */	ABSENT_STRING,
	/* 357: reqmp    */	ABSENT_STRING,
	/* 358: getm     */	ABSENT_STRING,
	/* 359: setaf    */	screen_linux_s_setaf,
	/* 360: setab    */	screen_linux_s_setab,
	/* 361: pfxl     */	ABSENT_STRING,
	/* 362: devt     */	ABSENT_STRING,
	/* 363: csin     */	ABSENT_STRING,
	/* 364: s0ds     */	ABSENT_STRING,
	/* 365: s1ds     */	ABSENT_STRING,
	/* 366: s2ds     */	ABSENT_STRING,
	/* 367: s3ds     */	ABSENT_STRING,
	/* 368: smglr    */	ABSENT_STRING,
	/* 369: smgtb    */	ABSENT_STRING,
	/* 370: birep    */	ABSENT_STRING,
	/* 371: binel    */	ABSENT_STRING,
	/* 372: bicr     */	ABSENT_STRING,
	/* 373: colornm  */	ABSENT_STRING,
	/* 374: defbi    */	ABSENT_STRING,
	/* 375: endbi    */	ABSENT_STRING,
	/* 376: setcolor */	ABSENT_STRING,
	/* 377: slines   */	ABSENT_STRING,
	/* 378: dispc    */	ABSENT_STRING,
	/* 379: smpch    */	ABSENT_STRING,
	/* 380: rmpch    */	ABSENT_STRING,
	/* 381: smsc     */	ABSENT_STRING,
	/* 382: rmsc     */	ABSENT_STRING,
	/* 383: pctrm    */	ABSENT_STRING,
	/* 384: scesc    */	ABSENT_STRING,
	/* 385: scesa    */	ABSENT_STRING,
	/* 386: ehhlm    */	ABSENT_STRING,
	/* 387: elhlm    */	ABSENT_STRING,
	/* 388: elohlm   */	ABSENT_STRING,
	/* 389: erhlm    */	ABSENT_STRING,
	/* 390: ethlm    */	ABSENT_STRING,
	/* 391: evhlm    */	ABSENT_STRING,
	/* 392: sgr1     */	ABSENT_STRING,
	/* 393: slength  */	ABSENT_STRING,
	/* 394: OTi2     */	ABSENT_STRING,
	/* 395: OTrs     */	ABSENT_STRING,
	/* 396: OTnl     */	ABSENT_STRING,
	/* 397: OTbc     */	ABSENT_STRING,
	/* 398: OTko     */	ABSENT_STRING,
	/* 399: OTma     */	ABSENT_STRING,
	/* 400: OTG2     */	ABSENT_STRING,
	/* 401: OTG3     */	ABSENT_STRING,
	/* 402: OTG1     */	ABSENT_STRING,
	/* 403: OTG4     */	ABSENT_STRING,
	/* 404: OTGR     */	ABSENT_STRING,
	/* 405: OTGL     */	ABSENT_STRING,
	/* 406: OTGU     */	ABSENT_STRING,
	/* 407: OTGD     */	ABSENT_STRING,
	/* 408: OTGH     */	ABSENT_STRING,
	/* 409: OTGV     */	ABSENT_STRING,
	/* 410: OTGC     */	ABSENT_STRING,
	/* 411: meml     */	ABSENT_STRING,
	/* 412: memu     */	ABSENT_STRING,
	/* 413: box1     */	ABSENT_STRING,
};
/* screen.rxvt */

static char screen_rxvt_alias_data[] = "screen.rxvt|screen in rxvt";

static char screen_rxvt_s_cbt   [] = "\033[Z";
static char screen_rxvt_s_bel   [] = "\007";
static char screen_rxvt_s_cr    [] = "\015";
static char screen_rxvt_s_csr   [] = "\033[%i%p1%d;%p2%dr";
static char screen_rxvt_s_tbc   [] = "\033[3g";
static char screen_rxvt_s_clear [] = "\033[H\033[J";
static char screen_rxvt_s_el    [] = "\033[K";
static char screen_rxvt_s_ed    [] = "\033[J";
static char screen_rxvt_s_hpa   [] = "\033[%i%p1%dG";
static char screen_rxvt_s_cup   [] = "\033[%i%p1%d;%p2%dH";
static char screen_rxvt_s_cud1  [] = "\012";
static char screen_rxvt_s_home  [] = "\033[H";
static char screen_rxvt_s_civis [] = "\033[?25l";
static char screen_rxvt_s_cub1  [] = "\010";
static char screen_rxvt_s_cnorm [] = "\033[34h\033[?25h";
static char screen_rxvt_s_cuf1  [] = "\033[C";
static char screen_rxvt_s_cuu1  [] = "\033M";
static char screen_rxvt_s_dch1  [] = "\033[P";
static char screen_rxvt_s_dl1   [] = "\033[M";
static char screen_rxvt_s_smacs [] = "\016";
static char screen_rxvt_s_blink [] = "\033[5m";
static char screen_rxvt_s_bold  [] = "\033[1m";
static char screen_rxvt_s_smcup [] = "\033[?1049h";
static char screen_rxvt_s_dim   [] = "\033[2m";
static char screen_rxvt_s_smir  [] = "\033[4h";
static char screen_rxvt_s_rev   [] = "\033[7m";
static char screen_rxvt_s_smso  [] = "\033[3m";
static char screen_rxvt_s_smul  [] = "\033[4m";
static char screen_rxvt_s_rmacs [] = "\017";
static char screen_rxvt_s_sgr0  [] = "\033[m\017";
static char screen_rxvt_s_rmcup [] = "\033[?1049l";
static char screen_rxvt_s_rmir  [] = "\033[4l";
static char screen_rxvt_s_rmso  [] = "\033[23m";
static char screen_rxvt_s_rmul  [] = "\033[24m";
static char screen_rxvt_s_is2   [] = "\033)0";
static char screen_rxvt_s_il1   [] = "\033[L";
static char screen_rxvt_s_kbs   [] = "\010";
static char screen_rxvt_s_kdch1 [] = "\033[3~";
static char screen_rxvt_s_kcud1 [] = "\033OB";
static char screen_rxvt_s_kel   [] = "\033[8^";
static char screen_rxvt_s_kf1   [] = "\033OP";
static char screen_rxvt_s_kf10  [] = "\033[21~";
static char screen_rxvt_s_kf2   [] = "\033OQ";
static char screen_rxvt_s_kf3   [] = "\033OR";
static char screen_rxvt_s_kf4   [] = "\033OS";
static char screen_rxvt_s_kf5   [] = "\033[15~";
static char screen_rxvt_s_kf6   [] = "\033[17~";
static char screen_rxvt_s_kf7   [] = "\033[18~";
static char screen_rxvt_s_kf8   [] = "\033[19~";
static char screen_rxvt_s_kf9   [] = "\033[20~";
static char screen_rxvt_s_khome [] = "\033[1~";
static char screen_rxvt_s_kich1 [] = "\033[2~";
static char screen_rxvt_s_kcub1 [] = "\033OD";
static char screen_rxvt_s_knp   [] = "\033[6~";
static char screen_rxvt_s_kpp   [] = "\033[5~";
static char screen_rxvt_s_kcuf1 [] = "\033OC";
static char screen_rxvt_s_kind  [] = "\033[a";
static char screen_rxvt_s_kri   [] = "\033[b";
static char screen_rxvt_s_kcuu1 [] = "\033OA";
static char screen_rxvt_s_rmkx  [] = "\033[?1l\033>";
static char screen_rxvt_s_smkx  [] = "\033[?1h\033=";
static char screen_rxvt_s_nel   [] = "\033E";
static char screen_rxvt_s_dch   [] = "\033[%p1%dP";
static char screen_rxvt_s_dl    [] = "\033[%p1%dM";
static char screen_rxvt_s_cud   [] = "\033[%p1%dB";
static char screen_rxvt_s_ich   [] = "\033[%p1%d@";
static char screen_rxvt_s_indn  [] = "\033[%p1%dS";
static char screen_rxvt_s_il    [] = "\033[%p1%dL";
static char screen_rxvt_s_cub   [] = "\033[%p1%dD";
static char screen_rxvt_s_cuf   [] = "\033[%p1%dC";
static char screen_rxvt_s_cuu   [] = "\033[%p1%dA";
static char screen_rxvt_s_rs2   [] = "\033c\033[?1000l\033[?25h";
static char screen_rxvt_s_rc    [] = "\0338";
static char screen_rxvt_s_vpa   [] = "\033[%i%p1%dd";
static char screen_rxvt_s_sc    [] = "\0337";
static char screen_rxvt_s_ind   [] = "\012";
static char screen_rxvt_s_ri    [] = "\033M";
static char screen_rxvt_s_sgr   [] = "\033[0%?%p6%t;1%;%?%p1%t;3%;%?%p2%t;4%;%?%p3%t;7%;%?%p4%t;5%;%?%p5%t;2%;m%?%p9%t\016%e\017%;";
static char screen_rxvt_s_hts   [] = "\033H";
static char screen_rxvt_s_ht    [] = "\011";
static char screen_rxvt_s_ka1   [] = "\033Ow";
static char screen_rxvt_s_ka3   [] = "\033Oy";
static char screen_rxvt_s_kb2   [] = "\033Ou";
static char screen_rxvt_s_kc1   [] = "\033Oq";
static char screen_rxvt_s_kc3   [] = "\033Os";
static char screen_rxvt_s_acsc  [] = "++,,--..00``aaffgghhiijjkkllmmnnooppqqrrssttuuvvwwxxyyzz{{||}}~~";
static char screen_rxvt_s_kcbt  [] = "\033[Z";
static char screen_rxvt_s_enacs [] = "\033(B\033)0";
static char screen_rxvt_s_kend  [] = "\033[4~";
static char screen_rxvt_s_kent  [] = "\033OM";
static char screen_rxvt_s_kDC   [] = "\033[3$";
static char screen_rxvt_s_kEND  [] = "\033[8$";
static char screen_rxvt_s_kHOM  [] = "\033[7$";
static char screen_rxvt_s_kIC   [] = "\033[2$";
static char screen_rxvt_s_kLFT  [] = "\033[d";
static char screen_rxvt_s_kNXT  [] = "\033[6$";
static char screen_rxvt_s_kPRV  [] = "\033[5$";
static char screen_rxvt_s_kRIT  [] = "\033[c";
static char screen_rxvt_s_kf11  [] = "\033[23~";
static char screen_rxvt_s_kf12  [] = "\033[24~";
static char screen_rxvt_s_kf13  [] = "\033[25~";
static char screen_rxvt_s_kf14  [] = "\033[26~";
static char screen_rxvt_s_kf15  [] = "\033[28~";
static char screen_rxvt_s_kf16  [] = "\033[29~";
static char screen_rxvt_s_kf17  [] = "\033[31~";
static char screen_rxvt_s_kf18  [] = "\033[32~";
static char screen_rxvt_s_kf19  [] = "\033[33~";
static char screen_rxvt_s_kf20  [] = "\033[34~";
static char screen_rxvt_s_kf21  [] = "\033[23$";
static char screen_rxvt_s_kf22  [] = "\033[24$";
static char screen_rxvt_s_kf23  [] = "\033[11^";
static char screen_rxvt_s_kf24  [] = "\033[12^";
static char screen_rxvt_s_kf25  [] = "\033[13^";
static char screen_rxvt_s_kf26  [] = "\033[14^";
static char screen_rxvt_s_kf27  [] = "\033[15^";
static char screen_rxvt_s_kf28  [] = "\033[17^";
static char screen_rxvt_s_kf29  [] = "\033[18^";
static char screen_rxvt_s_kf30  [] = "\033[19^";
static char screen_rxvt_s_kf31  [] = "\033[20^";
static char screen_rxvt_s_kf32  [] = "\033[21^";
static char screen_rxvt_s_kf33  [] = "\033[23^";
static char screen_rxvt_s_kf34  [] = "\033[24^";
static char screen_rxvt_s_kf35  [] = "\033[25^";
static char screen_rxvt_s_kf36  [] = "\033[26^";
static char screen_rxvt_s_kf37  [] = "\033[28^";
static char screen_rxvt_s_kf38  [] = "\033[29^";
static char screen_rxvt_s_kf39  [] = "\033[31^";
static char screen_rxvt_s_kf40  [] = "\033[32^";
static char screen_rxvt_s_kf41  [] = "\033[33^";
static char screen_rxvt_s_kf42  [] = "\033[34^";
static char screen_rxvt_s_kf43  [] = "\033[23@";
static char screen_rxvt_s_kf44  [] = "\033[24@";
static char screen_rxvt_s_el1   [] = "\033[1K";
static char screen_rxvt_s_u6    [] = "\033[%i%d;%dR";
static char screen_rxvt_s_u7    [] = "\033[6n";
static char screen_rxvt_s_u8    [] = "\033[?1;2c";
static char screen_rxvt_s_u9    [] = "\033[c";
static char screen_rxvt_s_op    [] = "\033[39;49m";
static char screen_rxvt_s_kmous [] = "\033[M";
static char screen_rxvt_s_setaf [] = "\033[3%p1%dm";
static char screen_rxvt_s_setab [] = "\033[4%p1%dm";

static char screen_rxvt_bool_data[] = {
	/*   0: bw       */	TRUE,
	/*   1: am       */	TRUE,
	/*   2: xsb      */	FALSE,
	/*   3: xhp      */	FALSE,
	/*   4: xenl     */	TRUE,
	/*   5: eo       */	FALSE,
	/*   6: gn       */	FALSE,
	/*   7: hc       */	FALSE,
	/*   8: km       */	TRUE,
	/*   9: hs       */	FALSE,
	/*  10: in       */	FALSE,
	/*  11: da       */	FALSE,
	/*  12: db       */	FALSE,
	/*  13: mir      */	TRUE,
	/*  14: msgr     */	TRUE,
	/*  15: os       */	FALSE,
	/*  16: eslok    */	FALSE,
	/*  17: xt       */	FALSE,
	/*  18: hz       */	FALSE,
	/*  19: ul       */	FALSE,
	/*  20: xon      */	FALSE,
	/*  21: nxon     */	FALSE,
	/*  22: mc5i     */	FALSE,
	/*  23: chts     */	FALSE,
	/*  24: nrrmc    */	FALSE,
	/*  25: npc      */	FALSE,
	/*  26: ndscr    */	FALSE,
	/*  27: ccc      */	FALSE,
	/*  28: bce      */	FALSE,
	/*  29: hls      */	FALSE,
	/*  30: xhpa     */	FALSE,
	/*  31: crxm     */	FALSE,
	/*  32: daisy    */	FALSE,
	/*  33: xvpa     */	FALSE,
	/*  34: sam      */	FALSE,
	/*  35: cpix     */	FALSE,
	/*  36: lpix     */	FALSE,
	/*  37: OTbs     */	TRUE,
	/*  38: OTns     */	FALSE,
	/*  39: OTnc     */	FALSE,
	/*  40: OTMT     */	FALSE,
	/*  41: OTNL     */	FALSE,
	/*  42: OTpt     */	TRUE,
	/*  43: OTxr     */	FALSE,
};
static NCURSES_INT2 screen_rxvt_number_data[] = {
	/*   0: cols     */	80,
	/*   1: it       */	8,
	/*   2: lines    */	24,
	/*   3: lm       */	ABSENT_NUMERIC,
	/*   4: xmc      */	ABSENT_NUMERIC,
	/*   5: pb       */	ABSENT_NUMERIC,
	/*   6: vt       */	ABSENT_NUMERIC,
	/*   7: wsl      */	ABSENT_NUMERIC,
	/*   8: nlab     */	ABSENT_NUMERIC,
	/*   9: lh       */	ABSENT_NUMERIC,
	/*  10: lw       */	ABSENT_NUMERIC,
	/*  11: ma       */	ABSENT_NUMERIC,
	/*  12: wnum     */	ABSENT_NUMERIC,
	/*  13: colors   */	8,
	/*  14: pairs    */	64,
	/*  15: ncv      */	ABSENT_NUMERIC,
	/*  16: bufsz    */	ABSENT_NUMERIC,
	/*  17: spinv    */	ABSENT_NUMERIC,
	/*  18: spinh    */	ABSENT_NUMERIC,
	/*  19: maddr    */	ABSENT_NUMERIC,
	/*  20: mjump    */	ABSENT_NUMERIC,
	/*  21: mcs      */	ABSENT_NUMERIC,
	/*  22: mls      */	ABSENT_NUMERIC,
	/*  23: npins    */	ABSENT_NUMERIC,
	/*  24: orc      */	ABSENT_NUMERIC,
	/*  25: orl      */	ABSENT_NUMERIC,
	/*  26: orhi     */	ABSENT_NUMERIC,
	/*  27: orvi     */	ABSENT_NUMERIC,
	/*  28: cps      */	ABSENT_NUMERIC,
	/*  29: widcs    */	ABSENT_NUMERIC,
	/*  30: btns     */	ABSENT_NUMERIC,
	/*  31: bitwin   */	ABSENT_NUMERIC,
	/*  32: bitype   */	ABSENT_NUMERIC,
	/*  33: OTug     */	ABSENT_NUMERIC,
	/*  34: OTdC     */	ABSENT_NUMERIC,
	/*  35: OTdN     */	ABSENT_NUMERIC,
	/*  36: OTdB     */	ABSENT_NUMERIC,
	/*  37: OTdT     */	ABSENT_NUMERIC,
	/*  38: OTkn     */	ABSENT_NUMERIC,
};
static char * screen_rxvt_string_data[] = {
	/*   0: cbt      */	screen_rxvt_s_cbt,
	/*   1: bel      */	screen_rxvt_s_bel,
	/*   2: cr       */	screen_rxvt_s_cr,
	/*   3: csr      */	screen_rxvt_s_csr,
	/*   4: tbc      */	screen_rxvt_s_tbc,
	/*   5: clear    */	screen_rxvt_s_clear,
	/*   6: el       */	screen_rxvt_s_el,
	/*   7: ed       */	screen_rxvt_s_ed,
	/*   8: hpa      */	screen_rxvt_s_hpa,
	/*   9: cmdch    */	ABSENT_STRING,
	/*  10: cup      */	screen_rxvt_s_cup,
	/*  11: cud1     */	screen_rxvt_s_cud1,
	/*  12: home     */	screen_rxvt_s_home,
	/*  13: civis    */	screen_rxvt_s_civis,
	/*  14: cub1     */	screen_rxvt_s_cub1,
	/*  15: mrcup    */	ABSENT_STRING,
	/*  16: cnorm    */	screen_rxvt_s_cnorm,
	/*  17: cuf1     */	screen_rxvt_s_cuf1,
	/*  18: ll       */	ABSENT_STRING,
	/*  19: cuu1     */	screen_rxvt_s_cuu1,
	/*  20: cvvis    */	CANCELLED_STRING,
	/*  21: dch1     */	screen_rxvt_s_dch1,
	/*  22: dl1      */	screen_rxvt_s_dl1,
	/*  23: dsl      */	ABSENT_STRING,
	/*  24: hd       */	ABSENT_STRING,
	/*  25: smacs    */	screen_rxvt_s_smacs,
	/*  26: blink    */	screen_rxvt_s_blink,
	/*  27: bold     */	screen_rxvt_s_bold,
	/*  28: smcup    */	screen_rxvt_s_smcup,
	/*  29: smdc     */	ABSENT_STRING,
	/*  30: dim      */	screen_rxvt_s_dim,
	/*  31: smir     */	screen_rxvt_s_smir,
	/*  32: invis    */	ABSENT_STRING,
	/*  33: prot     */	ABSENT_STRING,
	/*  34: rev      */	screen_rxvt_s_rev,
	/*  35: smso     */	screen_rxvt_s_smso,
	/*  36: smul     */	screen_rxvt_s_smul,
	/*  37: ech      */	ABSENT_STRING,
	/*  38: rmacs    */	screen_rxvt_s_rmacs,
	/*  39: sgr0     */	screen_rxvt_s_sgr0,
	/*  40: rmcup    */	screen_rxvt_s_rmcup,
	/*  41: rmdc     */	ABSENT_STRING,
	/*  42: rmir     */	screen_rxvt_s_rmir,
	/*  43: rmso     */	screen_rxvt_s_rmso,
	/*  44: rmul     */	screen_rxvt_s_rmul,
	/*  45: flash    */	CANCELLED_STRING,
	/*  46: ff       */	ABSENT_STRING,
	/*  47: fsl      */	ABSENT_STRING,
	/*  48: is1      */	ABSENT_STRING,
	/*  49: is2      */	screen_rxvt_s_is2,
	/*  50: is3      */	ABSENT_STRING,
	/*  51: if       */	ABSENT_STRING,
	/*  52: ich1     */	ABSENT_STRING,
	/*  53: il1      */	screen_rxvt_s_il1,
	/*  54: ip       */	ABSENT_STRING,
	/*  55: kbs      */	screen_rxvt_s_kbs,
	/*  56: ktbc     */	ABSENT_STRING,
	/*  57: kclr     */	ABSENT_STRING,
	/*  58: kctab    */	ABSENT_STRING,
	/*  59: kdch1    */	screen_rxvt_s_kdch1,
	/*  60: kdl1     */	ABSENT_STRING,
	/*  61: kcud1    */	screen_rxvt_s_kcud1,
	/*  62: krmir    */	ABSENT_STRING,
	/*  63: kel      */	screen_rxvt_s_kel,
	/*  64: ked      */	ABSENT_STRING,
	/*  65: kf0      */	ABSENT_STRING,
	/*  66: kf1      */	screen_rxvt_s_kf1,
	/*  67: kf10     */	screen_rxvt_s_kf10,
	/*  68: kf2      */	screen_rxvt_s_kf2,
	/*  69: kf3      */	screen_rxvt_s_kf3,
	/*  70: kf4      */	screen_rxvt_s_kf4,
	/*  71: kf5      */	screen_rxvt_s_kf5,
	/*  72: kf6      */	screen_rxvt_s_kf6,
	/*  73: kf7      */	screen_rxvt_s_kf7,
	/*  74: kf8      */	screen_rxvt_s_kf8,
	/*  75: kf9      */	screen_rxvt_s_kf9,
	/*  76: khome    */	screen_rxvt_s_khome,
	/*  77: kich1    */	screen_rxvt_s_kich1,
	/*  78: kil1     */	ABSENT_STRING,
	/*  79: kcub1    */	screen_rxvt_s_kcub1,
	/*  80: kll      */	ABSENT_STRING,
	/*  81: knp      */	screen_rxvt_s_knp,
	/*  82: kpp      */	screen_rxvt_s_kpp,
	/*  83: kcuf1    */	screen_rxvt_s_kcuf1,
	/*  84: kind     */	screen_rxvt_s_kind,
	/*  85: kri      */	screen_rxvt_s_kri,
	/*  86: khts     */	ABSENT_STRING,
	/*  87: kcuu1    */	screen_rxvt_s_kcuu1,
	/*  88: rmkx     */	screen_rxvt_s_rmkx,
	/*  89: smkx     */	screen_rxvt_s_smkx,
	/*  90: lf0      */	ABSENT_STRING,
	/*  91: lf1      */	ABSENT_STRING,
	/*  92: lf10     */	ABSENT_STRING,
	/*  93: lf2      */	ABSENT_STRING,
	/*  94: lf3      */	ABSENT_STRING,
	/*  95: lf4      */	ABSENT_STRING,
	/*  96: lf5      */	ABSENT_STRING,
	/*  97: lf6      */	ABSENT_STRING,
	/*  98: lf7      */	ABSENT_STRING,
	/*  99: lf8      */	ABSENT_STRING,
	/* 100: lf9      */	ABSENT_STRING,
	/* 101: rmm      */	ABSENT_STRING,
	/* 102: smm      */	ABSENT_STRING,
	/* 103: nel      */	screen_rxvt_s_nel,
	/* 104: pad      */	ABSENT_STRING,
	/* 105: dch      */	screen_rxvt_s_dch,
	/* 106: dl       */	screen_rxvt_s_dl,
	/* 107: cud      */	screen_rxvt_s_cud,
	/* 108: ich      */	screen_rxvt_s_ich,
	/* 109: indn     */	screen_rxvt_s_indn,
	/* 110: il       */	screen_rxvt_s_il,
	/* 111: cub      */	screen_rxvt_s_cub,
	/* 112: cuf      */	screen_rxvt_s_cuf,
	/* 113: rin      */	ABSENT_STRING,
	/* 114: cuu      */	screen_rxvt_s_cuu,
	/* 115: pfkey    */	ABSENT_STRING,
	/* 116: pfloc    */	ABSENT_STRING,
	/* 117: pfx      */	ABSENT_STRING,
	/* 118: mc0      */	ABSENT_STRING,
	/* 119: mc4      */	ABSENT_STRING,
	/* 120: mc5      */	ABSENT_STRING,
	/* 121: rep      */	ABSENT_STRING,
	/* 122: rs1      */	ABSENT_STRING,
	/* 123: rs2      */	screen_rxvt_s_rs2,
	/* 124: rs3      */	ABSENT_STRING,
	/* 125: rf       */	ABSENT_STRING,
	/* 126: rc       */	screen_rxvt_s_rc,
	/* 127: vpa      */	screen_rxvt_s_vpa,
	/* 128: sc       */	screen_rxvt_s_sc,
	/* 129: ind      */	screen_rxvt_s_ind,
	/* 130: ri       */	screen_rxvt_s_ri,
	/* 131: sgr      */	screen_rxvt_s_sgr,
	/* 132: hts      */	screen_rxvt_s_hts,
	/* 133: wind     */	ABSENT_STRING,
	/* 134: ht       */	screen_rxvt_s_ht,
	/* 135: tsl      */	ABSENT_STRING,
	/* 136: uc       */	ABSENT_STRING,
	/* 137: hu       */	ABSENT_STRING,
	/* 138: iprog    */	ABSENT_STRING,
	/* 139: ka1      */	screen_rxvt_s_ka1,
	/* 140: ka3      */	screen_rxvt_s_ka3,
	/* 141: kb2      */	screen_rxvt_s_kb2,
	/* 142: kc1      */	screen_rxvt_s_kc1,
	/* 143: kc3      */	screen_rxvt_s_kc3,
	/* 144: mc5p     */	ABSENT_STRING,
	/* 145: rmp      */	ABSENT_STRING,
	/* 146: acsc     */	screen_rxvt_s_acsc,
	/* 147: pln      */	ABSENT_STRING,
	/* 148: kcbt     */	screen_rxvt_s_kcbt,
	/* 149: smxon    */	ABSENT_STRING,
	/* 150: rmxon    */	ABSENT_STRING,
	/* 151: smam     */	ABSENT_STRING,
	/* 152: rmam     */	ABSENT_STRING,
	/* 153: xonc     */	ABSENT_STRING,
	/* 154: xoffc    */	ABSENT_STRING,
	/* 155: enacs    */	screen_rxvt_s_enacs,
	/* 156: smln     */	ABSENT_STRING,
	/* 157: rmln     */	ABSENT_STRING,
	/* 158: kbeg     */	ABSENT_STRING,
	/* 159: kcan     */	ABSENT_STRING,
	/* 160: kclo     */	ABSENT_STRING,
	/* 161: kcmd     */	ABSENT_STRING,
	/* 162: kcpy     */	ABSENT_STRING,
	/* 163: kcrt     */	ABSENT_STRING,
	/* 164: kend     */	screen_rxvt_s_kend,
	/* 165: kent     */	screen_rxvt_s_kent,
	/* 166: kext     */	ABSENT_STRING,
	/* 167: kfnd     */	ABSENT_STRING,
	/* 168: khlp     */	ABSENT_STRING,
	/* 169: kmrk     */	ABSENT_STRING,
	/* 170: kmsg     */	ABSENT_STRING,
	/* 171: kmov     */	ABSENT_STRING,
	/* 172: knxt     */	ABSENT_STRING,
	/* 173: kopn     */	ABSENT_STRING,
	/* 174: kopt     */	ABSENT_STRING,
	/* 175: kprv     */	ABSENT_STRING,
	/* 176: kprt     */	ABSENT_STRING,
	/* 177: krdo     */	ABSENT_STRING,
	/* 178: kref     */	ABSENT_STRING,
	/* 179: krfr     */	ABSENT_STRING,
	/* 180: krpl     */	ABSENT_STRING,
	/* 181: krst     */	ABSENT_STRING,
	/* 182: kres     */	ABSENT_STRING,
	/* 183: ksav     */	ABSENT_STRING,
	/* 184: kspd     */	ABSENT_STRING,
	/* 185: kund     */	ABSENT_STRING,
	/* 186: kBEG     */	ABSENT_STRING,
	/* 187: kCAN     */	ABSENT_STRING,
	/* 188: kCMD     */	ABSENT_STRING,
	/* 189: kCPY     */	ABSENT_STRING,
	/* 190: kCRT     */	ABSENT_STRING,
	/* 191: kDC      */	screen_rxvt_s_kDC,
	/* 192: kDL      */	ABSENT_STRING,
	/* 193: kslt     */	ABSENT_STRING,
	/* 194: kEND     */	screen_rxvt_s_kEND,
	/* 195: kEOL     */	ABSENT_STRING,
	/* 196: kEXT     */	ABSENT_STRING,
	/* 197: kFND     */	ABSENT_STRING,
	/* 198: kHLP     */	ABSENT_STRING,
	/* 199: kHOM     */	screen_rxvt_s_kHOM,
	/* 200: kIC      */	screen_rxvt_s_kIC,
	/* 201: kLFT     */	screen_rxvt_s_kLFT,
	/* 202: kMSG     */	ABSENT_STRING,
	/* 203: kMOV     */	ABSENT_STRING,
	/* 204: kNXT     */	screen_rxvt_s_kNXT,
	/* 205: kOPT     */	ABSENT_STRING,
	/* 206: kPRV     */	screen_rxvt_s_kPRV,
	/* 207: kPRT     */	ABSENT_STRING,
	/* 208: kRDO     */	ABSENT_STRING,
	/* 209: kRPL     */	ABSENT_STRING,
	/* 210: kRIT     */	screen_rxvt_s_kRIT,
	/* 211: kRES     */	ABSENT_STRING,
	/* 212: kSAV     */	ABSENT_STRING,
	/* 213: kSPD     */	ABSENT_STRING,
	/* 214: kUND     */	ABSENT_STRING,
	/* 215: rfi      */	ABSENT_STRING,
	/* 216: kf11     */	screen_rxvt_s_kf11,
	/* 217: kf12     */	screen_rxvt_s_kf12,
	/* 218: kf13     */	screen_rxvt_s_kf13,
	/* 219: kf14     */	screen_rxvt_s_kf14,
	/* 220: kf15     */	screen_rxvt_s_kf15,
	/* 221: kf16     */	screen_rxvt_s_kf16,
	/* 222: kf17     */	screen_rxvt_s_kf17,
	/* 223: kf18     */	screen_rxvt_s_kf18,
	/* 224: kf19     */	screen_rxvt_s_kf19,
	/* 225: kf20     */	screen_rxvt_s_kf20,
	/* 226: kf21     */	screen_rxvt_s_kf21,
	/* 227: kf22     */	screen_rxvt_s_kf22,
	/* 228: kf23     */	screen_rxvt_s_kf23,
	/* 229: kf24     */	screen_rxvt_s_kf24,
	/* 230: kf25     */	screen_rxvt_s_kf25,
	/* 231: kf26     */	screen_rxvt_s_kf26,
	/* 232: kf27     */	screen_rxvt_s_kf27,
	/* 233: kf28     */	screen_rxvt_s_kf28,
	/* 234: kf29     */	screen_rxvt_s_kf29,
	/* 235: kf30     */	screen_rxvt_s_kf30,
	/* 236: kf31     */	screen_rxvt_s_kf31,
	/* 237: kf32     */	screen_rxvt_s_kf32,
	/* 238: kf33     */	screen_rxvt_s_kf33,
	/* 239: kf34     */	screen_rxvt_s_kf34,
	/* 240: kf35     */	screen_rxvt_s_kf35,
	/* 241: kf36     */	screen_rxvt_s_kf36,
	/* 242: kf37     */	screen_rxvt_s_kf37,
	/* 243: kf38     */	screen_rxvt_s_kf38,
	/* 244: kf39     */	screen_rxvt_s_kf39,
	/* 245: kf40     */	screen_rxvt_s_kf40,
	/* 246: kf41     */	screen_rxvt_s_kf41,
	/* 247: kf42     */	screen_rxvt_s_kf42,
	/* 248: kf43     */	screen_rxvt_s_kf43,
	/* 249: kf44     */	screen_rxvt_s_kf44,
	/* 250: kf45     */	ABSENT_STRING,
	/* 251: kf46     */	ABSENT_STRING,
	/* 252: kf47     */	ABSENT_STRING,
	/* 253: kf48     */	ABSENT_STRING,
	/* 254: kf49     */	ABSENT_STRING,
	/* 255: kf50     */	ABSENT_STRING,
	/* 256: kf51     */	ABSENT_STRING,
	/* 257: kf52     */	ABSENT_STRING,
	/* 258: kf53     */	ABSENT_STRING,
	/* 259: kf54     */	ABSENT_STRING,
	/* 260: kf55     */	ABSENT_STRING,
	/* 261: kf56     */	ABSENT_STRING,
	/* 262: kf57     */	ABSENT_STRING,
	/* 263: kf58     */	ABSENT_STRING,
	/* 264: kf59     */	ABSENT_STRING,
	/* 265: kf60     */	ABSENT_STRING,
	/* 266: kf61     */	ABSENT_STRING,
	/* 267: kf62     */	ABSENT_STRING,
	/* 268: kf63     */	ABSENT_STRING,
	/* 269: el1      */	screen_rxvt_s_el1,
	/* 270: mgc      */	ABSENT_STRING,
	/* 271: smgl     */	ABSENT_STRING,
	/* 272: smgr     */	ABSENT_STRING,
	/* 273: fln      */	ABSENT_STRING,
	/* 274: sclk     */	ABSENT_STRING,
	/* 275: dclk     */	ABSENT_STRING,
	/* 276: rmclk    */	ABSENT_STRING,
	/* 277: cwin     */	ABSENT_STRING,
	/* 278: wingo    */	ABSENT_STRING,
	/* 279: hup      */	ABSENT_STRING,
	/* 280: dial     */	ABSENT_STRING,
	/* 281: qdial    */	ABSENT_STRING,
	/* 282: tone     */	ABSENT_STRING,
	/* 283: pulse    */	ABSENT_STRING,
	/* 284: hook     */	ABSENT_STRING,
	/* 285: pause    */	ABSENT_STRING,
	/* 286: wait     */	ABSENT_STRING,
	/* 287: u0       */	ABSENT_STRING,
	/* 288: u1       */	ABSENT_STRING,
	/* 289: u2       */	ABSENT_STRING,
	/* 290: u3       */	ABSENT_STRING,
	/* 291: u4       */	ABSENT_STRING,
	/* 292: u5       */	ABSENT_STRING,
	/* 293: u6       */	screen_rxvt_s_u6,
	/* 294: u7       */	screen_rxvt_s_u7,
	/* 295: u8       */	screen_rxvt_s_u8,
	/* 296: u9       */	screen_rxvt_s_u9,
	/* 297: op       */	screen_rxvt_s_op,
	/* 298: oc       */	ABSENT_STRING,
	/* 299: initc    */	ABSENT_STRING,
	/* 300: initp    */	ABSENT_STRING,
	/* 301: scp      */	ABSENT_STRING,
	/* 302: setf     */	ABSENT_STRING,
	/* 303: setb     */	ABSENT_STRING,
	/* 304: cpi      */	ABSENT_STRING,
	/* 305: lpi      */	ABSENT_STRING,
	/* 306: chr      */	ABSENT_STRING,
	/* 307: cvr      */	ABSENT_STRING,
	/* 308: defc     */	ABSENT_STRING,
	/* 309: swidm    */	ABSENT_STRING,
	/* 310: sdrfq    */	ABSENT_STRING,
	/* 311: sitm     */	ABSENT_STRING,
	/* 312: slm      */	ABSENT_STRING,
	/* 313: smicm    */	ABSENT_STRING,
	/* 314: snlq     */	ABSENT_STRING,
	/* 315: snrmq    */	ABSENT_STRING,
	/* 316: sshm     */	ABSENT_STRING,
	/* 317: ssubm    */	ABSENT_STRING,
	/* 318: ssupm    */	ABSENT_STRING,
	/* 319: sum      */	ABSENT_STRING,
	/* 320: rwidm    */	ABSENT_STRING,
	/* 321: ritm     */	ABSENT_STRING,
	/* 322: rlm      */	ABSENT_STRING,
	/* 323: rmicm    */	ABSENT_STRING,
	/* 324: rshm     */	ABSENT_STRING,
	/* 325: rsubm    */	ABSENT_STRING,
	/* 326: rsupm    */	ABSENT_STRING,
	/* 327: rum      */	ABSENT_STRING,
	/* 328: mhpa     */	ABSENT_STRING,
	/* 329: mcud1    */	ABSENT_STRING,
	/* 330: mcub1    */	ABSENT_STRING,
	/* 331: mcuf1    */	ABSENT_STRING,
	/* 332: mvpa     */	ABSENT_STRING,
	/* 333: mcuu1    */	ABSENT_STRING,
	/* 334: porder   */	ABSENT_STRING,
	/* 335: mcud     */	ABSENT_STRING,
	/* 336: mcub     */	ABSENT_STRING,
	/* 337: mcuf     */	ABSENT_STRING,
	/* 338: mcuu     */	ABSENT_STRING,
	/* 339: scs      */	ABSENT_STRING,
	/* 340: smgb     */	ABSENT_STRING,
	/* 341: smgbp    */	ABSENT_STRING,
	/* 342: smglp    */	ABSENT_STRING,
	/* 343: smgrp    */	ABSENT_STRING,
	/* 344: smgt     */	ABSENT_STRING,
	/* 345: smgtp    */	ABSENT_STRING,
	/* 346: sbim     */	ABSENT_STRING,
	/* 347: scsd     */	ABSENT_STRING,
	/* 348: rbim     */	ABSENT_STRING,
	/* 349: rcsd     */	ABSENT_STRING,
	/* 350: subcs    */	ABSENT_STRING,
	/* 351: supcs    */	ABSENT_STRING,
	/* 352: docr     */	ABSENT_STRING,
	/* 353: zerom    */	ABSENT_STRING,
	/* 354: csnm     */	ABSENT_STRING,
	/* 355: kmous    */	screen_rxvt_s_kmous,
	/* 356: minfo    */	ABSENT_STRING,
	/* 357: reqmp    */	ABSENT_STRING,
	/* 358: getm     */	ABSENT_STRING,
	/* 359: setaf    */	screen_rxvt_s_setaf,
	/* 360: setab    */	screen_rxvt_s_setab,
	/* 361: pfxl     */	ABSENT_STRING,
	/* 362: devt     */	ABSENT_STRING,
	/* 363: csin     */	ABSENT_STRING,
	/* 364: s0ds     */	ABSENT_STRING,
	/* 365: s1ds     */	ABSENT_STRING,
	/* 366: s2ds     */	ABSENT_STRING,
	/* 367: s3ds     */	ABSENT_STRING,
	/* 368: smglr    */	ABSENT_STRING,
	/* 369: smgtb    */	ABSENT_STRING,
	/* 370: birep    */	ABSENT_STRING,
	/* 371: binel    */	ABSENT_STRING,
	/* 372: bicr     */	ABSENT_STRING,
	/* 373: colornm  */	ABSENT_STRING,
	/* 374: defbi    */	ABSENT_STRING,
	/* 375: endbi    */	ABSENT_STRING,
	/* 376: setcolor */	ABSENT_STRING,
	/* 377: slines   */	ABSENT_STRING,
	/* 378: dispc    */	ABSENT_STRING,
	/* 379: smpch    */	ABSENT_STRING,
	/* 380: rmpch    */	ABSENT_STRING,
	/* 381: smsc     */	ABSENT_STRING,
	/* 382: rmsc     */	ABSENT_STRING,
	/* 383: pctrm    */	ABSENT_STRING,
	/* 384: scesc    */	ABSENT_STRING,
	/* 385: scesa    */	ABSENT_STRING,
	/* 386: ehhlm    */	ABSENT_STRING,
	/* 387: elhlm    */	ABSENT_STRING,
	/* 388: elohlm   */	ABSENT_STRING,
	/* 389: erhlm    */	ABSENT_STRING,
	/* 390: ethlm    */	ABSENT_STRING,
	/* 391: evhlm    */	ABSENT_STRING,
	/* 392: sgr1     */	ABSENT_STRING,
	/* 393: slength  */	ABSENT_STRING,
	/* 394: OTi2     */	ABSENT_STRING,
	/* 395: OTrs     */	ABSENT_STRING,
	/* 396: OTnl     */	ABSENT_STRING,
	/* 397: OTbc     */	ABSENT_STRING,
	/* 398: OTko     */	ABSENT_STRING,
	/* 399: OTma     */	ABSENT_STRING,
	/* 400: OTG2     */	ABSENT_STRING,
	/* 401: OTG3     */	ABSENT_STRING,
	/* 402: OTG1     */	ABSENT_STRING,
	/* 403: OTG4     */	ABSENT_STRING,
	/* 404: OTGR     */	ABSENT_STRING,
	/* 405: OTGL     */	ABSENT_STRING,
	/* 406: OTGU     */	ABSENT_STRING,
	/* 407: OTGD     */	ABSENT_STRING,
	/* 408: OTGH     */	ABSENT_STRING,
	/* 409: OTGV     */	ABSENT_STRING,
	/* 410: OTGC     */	ABSENT_STRING,
	/* 411: meml     */	ABSENT_STRING,
	/* 412: memu     */	ABSENT_STRING,
	/* 413: box1     */	ABSENT_STRING,
};
/* screen.xterm-new */

static char screen_xterm_xfree86_alias_data[] = "screen.xterm-xfree86|screen.xterm-new|screen customized for modern xterm";

static char screen_xterm_xfree86_s_cbt[] = "\033[Z";
static char screen_xterm_xfree86_s_bel[] = "\007";
static char screen_xterm_xfree86_s_cr[] = "\015";
static char screen_xterm_xfree86_s_csr[] = "\033[%i%p1%d;%p2%dr";
static char screen_xterm_xfree86_s_tbc[] = "\033[3g";
static char screen_xterm_xfree86_s_clear[] = "\033[H\033[2J";
static char screen_xterm_xfree86_s_el[] = "\033[K";
static char screen_xterm_xfree86_s_ed[] = "\033[J";
static char screen_xterm_xfree86_s_hpa[] = "\033[%i%p1%dG";
static char screen_xterm_xfree86_s_cup[] = "\033[%i%p1%d;%p2%dH";
static char screen_xterm_xfree86_s_cud1[] = "\012";
static char screen_xterm_xfree86_s_home[] = "\033[H";
static char screen_xterm_xfree86_s_civis[] = "\033[?25l";
static char screen_xterm_xfree86_s_cub1[] = "\010";
static char screen_xterm_xfree86_s_cnorm[] = "\033[?12l\033[?25h";
static char screen_xterm_xfree86_s_cuf1[] = "\033[C";
static char screen_xterm_xfree86_s_cuu1[] = "\033[A";
static char screen_xterm_xfree86_s_cvvis[] = "\033[?12;25h";
static char screen_xterm_xfree86_s_dch1[] = "\033[P";
static char screen_xterm_xfree86_s_dl1[] = "\033[M";
static char screen_xterm_xfree86_s_smacs[] = "\033(0";
static char screen_xterm_xfree86_s_blink[] = "\033[5m";
static char screen_xterm_xfree86_s_bold[] = "\033[1m";
static char screen_xterm_xfree86_s_smcup[] = "\033[?1049h\033[22;0;0t";
static char screen_xterm_xfree86_s_dim[] = "\033[2m";
static char screen_xterm_xfree86_s_smir[] = "\033[4h";
static char screen_xterm_xfree86_s_rev[] = "\033[7m";
static char screen_xterm_xfree86_s_smso[] = "\033[7m";
static char screen_xterm_xfree86_s_smul[] = "\033[4m";
static char screen_xterm_xfree86_s_ech[] = "\033[%p1%dX";
static char screen_xterm_xfree86_s_rmacs[] = "\033(B";
static char screen_xterm_xfree86_s_sgr0[] = "\033(B\033[m";
static char screen_xterm_xfree86_s_rmcup[] = "\033[?1049l\033[23;0;0t";
static char screen_xterm_xfree86_s_rmir[] = "\033[4l";
static char screen_xterm_xfree86_s_rmso[] = "\033[27m";
static char screen_xterm_xfree86_s_rmul[] = "\033[24m";
static char screen_xterm_xfree86_s_flash[] = "\033[?5h$<100/>\033[?5l";
static char screen_xterm_xfree86_s_is2[] = "\033[!p\033[?3;4l\033[4l\033>";
static char screen_xterm_xfree86_s_il1[] = "\033[L";
static char screen_xterm_xfree86_s_kbs[] = "\010";
static char screen_xterm_xfree86_s_kdch1[] = "\033[3~";
static char screen_xterm_xfree86_s_kcud1[] = "\033OB";
static char screen_xterm_xfree86_s_kf1[] = "\033OP";
static char screen_xterm_xfree86_s_kf10[] = "\033[21~";
static char screen_xterm_xfree86_s_kf2[] = "\033OQ";
static char screen_xterm_xfree86_s_kf3[] = "\033OR";
static char screen_xterm_xfree86_s_kf4[] = "\033OS";
static char screen_xterm_xfree86_s_kf5[] = "\033[15~";
static char screen_xterm_xfree86_s_kf6[] = "\033[17~";
static char screen_xterm_xfree86_s_kf7[] = "\033[18~";
static char screen_xterm_xfree86_s_kf8[] = "\033[19~";
static char screen_xterm_xfree86_s_kf9[] = "\033[20~";
static char screen_xterm_xfree86_s_khome[] = "\033[1~";
static char screen_xterm_xfree86_s_kich1[] = "\033[2~";
static char screen_xterm_xfree86_s_kcub1[] = "\033OD";
static char screen_xterm_xfree86_s_knp[] = "\033[6~";
static char screen_xterm_xfree86_s_kpp[] = "\033[5~";
static char screen_xterm_xfree86_s_kcuf1[] = "\033OC";
static char screen_xterm_xfree86_s_kind[] = "\033[1;2B";
static char screen_xterm_xfree86_s_kri[] = "\033[1;2A";
static char screen_xterm_xfree86_s_kcuu1[] = "\033OA";
static char screen_xterm_xfree86_s_rmkx[] = "\033[?1l\033>";
static char screen_xterm_xfree86_s_smkx[] = "\033[?1h\033=";
static char screen_xterm_xfree86_s_rmm[] = "\033[?1034l";
static char screen_xterm_xfree86_s_smm[] = "\033[?1034h";
static char screen_xterm_xfree86_s_dch[] = "\033[%p1%dP";
static char screen_xterm_xfree86_s_dl[] = "\033[%p1%dM";
static char screen_xterm_xfree86_s_cud[] = "\033[%p1%dB";
static char screen_xterm_xfree86_s_ich[] = "\033[%p1%d@";
static char screen_xterm_xfree86_s_indn[] = "\033[%p1%dS";
static char screen_xterm_xfree86_s_il[] = "\033[%p1%dL";
static char screen_xterm_xfree86_s_cub[] = "\033[%p1%dD";
static char screen_xterm_xfree86_s_cuf[] = "\033[%p1%dC";
static char screen_xterm_xfree86_s_rin[] = "\033[%p1%dT";
static char screen_xterm_xfree86_s_cuu[] = "\033[%p1%dA";
static char screen_xterm_xfree86_s_mc0[] = "\033[i";
static char screen_xterm_xfree86_s_mc4[] = "\033[4i";
static char screen_xterm_xfree86_s_mc5[] = "\033[5i";
static char screen_xterm_xfree86_s_rs1[] = "\033c";
static char screen_xterm_xfree86_s_rs2[] = "\033[!p\033[?3;4l\033[4l\033>";
static char screen_xterm_xfree86_s_rc[] = "\0338";
static char screen_xterm_xfree86_s_vpa[] = "\033[%i%p1%dd";
static char screen_xterm_xfree86_s_sc[] = "\0337";
static char screen_xterm_xfree86_s_ind[] = "\012";
static char screen_xterm_xfree86_s_ri[] = "\033M";
static char screen_xterm_xfree86_s_sgr[] = "%?%p9%t\033(0%e\033(B%;\033[0%?%p6%t;1%;%?%p2%t;4%;%?%p1%p3%|%t;7%;%?%p4%t;5%;%?%p5%t;2%;m";
static char screen_xterm_xfree86_s_hts[] = "\033H";
static char screen_xterm_xfree86_s_ht[] = "\011";
static char screen_xterm_xfree86_s_kb2[] = "\033OE";
static char screen_xterm_xfree86_s_acsc[] = "``aaffggiijjkkllmmnnooppqqrrssttuuvvwwxxyyzz{{||}}~~";
static char screen_xterm_xfree86_s_kcbt[] = "\033[Z";
static char screen_xterm_xfree86_s_smam[] = "\033[?7h";
static char screen_xterm_xfree86_s_rmam[] = "\033[?7l";
static char screen_xterm_xfree86_s_kend[] = "\033[4~";
static char screen_xterm_xfree86_s_kent[] = "\033OM";
static char screen_xterm_xfree86_s_kDC[] = "\033[3;2~";
static char screen_xterm_xfree86_s_kEND[] = "\033[1;2F";
static char screen_xterm_xfree86_s_kHOM[] = "\033[1;2H";
static char screen_xterm_xfree86_s_kLFT[] = "\033[1;2D";
static char screen_xterm_xfree86_s_kRIT[] = "\033[1;2C";
static char screen_xterm_xfree86_s_kf11[] = "\033[23~";
static char screen_xterm_xfree86_s_kf12[] = "\033[24~";
static char screen_xterm_xfree86_s_kf13[] = "\033[1;2P";
static char screen_xterm_xfree86_s_kf14[] = "\033[1;2Q";
static char screen_xterm_xfree86_s_kf15[] = "\033[1;2R";
static char screen_xterm_xfree86_s_kf16[] = "\033[1;2S";
static char screen_xterm_xfree86_s_kf17[] = "\033[15;2~";
static char screen_xterm_xfree86_s_kf18[] = "\033[17;2~";
static char screen_xterm_xfree86_s_kf19[] = "\033[18;2~";
static char screen_xterm_xfree86_s_kf20[] = "\033[19;2~";
static char screen_xterm_xfree86_s_kf21[] = "\033[20;2~";
static char screen_xterm_xfree86_s_kf22[] = "\033[21;2~";
static char screen_xterm_xfree86_s_kf23[] = "\033[23;2~";
static char screen_xterm_xfree86_s_kf24[] = "\033[24;2~";
static char screen_xterm_xfree86_s_kf25[] = "\033[1;5P";
static char screen_xterm_xfree86_s_kf26[] = "\033[1;5Q";
static char screen_xterm_xfree86_s_kf27[] = "\033[1;5R";
static char screen_xterm_xfree86_s_kf28[] = "\033[1;5S";
static char screen_xterm_xfree86_s_kf29[] = "\033[15;5~";
static char screen_xterm_xfree86_s_kf30[] = "\033[17;5~";
static char screen_xterm_xfree86_s_kf31[] = "\033[18;5~";
static char screen_xterm_xfree86_s_kf32[] = "\033[19;5~";
static char screen_xterm_xfree86_s_kf33[] = "\033[20;5~";
static char screen_xterm_xfree86_s_kf34[] = "\033[21;5~";
static char screen_xterm_xfree86_s_kf35[] = "\033[23;5~";
static char screen_xterm_xfree86_s_kf36[] = "\033[24;5~";
static char screen_xterm_xfree86_s_kf37[] = "\033[1;6P";
static char screen_xterm_xfree86_s_kf38[] = "\033[1;6Q";
static char screen_xterm_xfree86_s_kf39[] = "\033[1;6R";
static char screen_xterm_xfree86_s_kf40[] = "\033[1;6S";
static char screen_xterm_xfree86_s_kf41[] = "\033[15;6~";
static char screen_xterm_xfree86_s_kf42[] = "\033[17;6~";
static char screen_xterm_xfree86_s_kf43[] = "\033[18;6~";
static char screen_xterm_xfree86_s_kf44[] = "\033[19;6~";
static char screen_xterm_xfree86_s_kf45[] = "\033[20;6~";
static char screen_xterm_xfree86_s_kf46[] = "\033[21;6~";
static char screen_xterm_xfree86_s_kf47[] = "\033[23;6~";
static char screen_xterm_xfree86_s_kf48[] = "\033[24;6~";
static char screen_xterm_xfree86_s_kf49[] = "\033[1;3P";
static char screen_xterm_xfree86_s_kf50[] = "\033[1;3Q";
static char screen_xterm_xfree86_s_kf51[] = "\033[1;3R";
static char screen_xterm_xfree86_s_kf52[] = "\033[1;3S";
static char screen_xterm_xfree86_s_kf53[] = "\033[15;3~";
static char screen_xterm_xfree86_s_kf54[] = "\033[17;3~";
static char screen_xterm_xfree86_s_kf55[] = "\033[18;3~";
static char screen_xterm_xfree86_s_kf56[] = "\033[19;3~";
static char screen_xterm_xfree86_s_kf57[] = "\033[20;3~";
static char screen_xterm_xfree86_s_kf58[] = "\033[21;3~";
static char screen_xterm_xfree86_s_kf59[] = "\033[23;3~";
static char screen_xterm_xfree86_s_kf60[] = "\033[24;3~";
static char screen_xterm_xfree86_s_kf61[] = "\033[1;4P";
static char screen_xterm_xfree86_s_kf62[] = "\033[1;4Q";
static char screen_xterm_xfree86_s_kf63[] = "\033[1;4R";
static char screen_xterm_xfree86_s_el1[] = "\033[1K";
static char screen_xterm_xfree86_s_u6[] = "\033[%i%d;%dR";
static char screen_xterm_xfree86_s_u7[] = "\033[6n";
static char screen_xterm_xfree86_s_u8[] = "\033[?%[;0123456789]c";
static char screen_xterm_xfree86_s_u9[] = "\033[c";
static char screen_xterm_xfree86_s_op[] = "\033[39;49m";
static char screen_xterm_xfree86_s_setf[] = "\033[3%?%p1%{1}%=%t4%e%p1%{3}%=%t6%e%p1%{4}%=%t1%e%p1%{6}%=%t3%e%p1%d%;m";
static char screen_xterm_xfree86_s_setb[] = "\033[4%?%p1%{1}%=%t4%e%p1%{3}%=%t6%e%p1%{4}%=%t1%e%p1%{6}%=%t3%e%p1%d%;m";
static char screen_xterm_xfree86_s_kmous[] = "\033[M";
static char screen_xterm_xfree86_s_setaf[] = "\033[3%p1%dm";
static char screen_xterm_xfree86_s_setab[] = "\033[4%p1%dm";

static char screen_xterm_xfree86_bool_data[] = {
	/*   0: bw       */	TRUE,
	/*   1: am       */	TRUE,
	/*   2: xsb      */	FALSE,
	/*   3: xhp      */	FALSE,
	/*   4: xenl     */	TRUE,
	/*   5: eo       */	FALSE,
	/*   6: gn       */	FALSE,
	/*   7: hc       */	FALSE,
	/*   8: km       */	TRUE,
	/*   9: hs       */	FALSE,
	/*  10: in       */	FALSE,
	/*  11: da       */	FALSE,
	/*  12: db       */	FALSE,
	/*  13: mir      */	TRUE,
	/*  14: msgr     */	TRUE,
	/*  15: os       */	FALSE,
	/*  16: eslok    */	FALSE,
	/*  17: xt       */	FALSE,
	/*  18: hz       */	FALSE,
	/*  19: ul       */	FALSE,
	/*  20: xon      */	FALSE,
	/*  21: nxon     */	FALSE,
	/*  22: mc5i     */	TRUE,
	/*  23: chts     */	FALSE,
	/*  24: nrrmc    */	FALSE,
	/*  25: npc      */	TRUE,
	/*  26: ndscr    */	FALSE,
	/*  27: ccc      */	FALSE,
	/*  28: bce      */	FALSE,
	/*  29: hls      */	FALSE,
	/*  30: xhpa     */	FALSE,
	/*  31: crxm     */	FALSE,
	/*  32: daisy    */	FALSE,
	/*  33: xvpa     */	FALSE,
	/*  34: sam      */	FALSE,
	/*  35: cpix     */	FALSE,
	/*  36: lpix     */	FALSE,
	/*  37: OTbs     */	TRUE,
	/*  38: OTns     */	FALSE,
	/*  39: OTnc     */	FALSE,
	/*  40: OTMT     */	FALSE,
	/*  41: OTNL     */	FALSE,
	/*  42: OTpt     */	FALSE,
	/*  43: OTxr     */	FALSE,
};
static NCURSES_INT2 screen_xterm_xfree86_number_data[] = {
	/*   0: cols     */	80,
	/*   1: it       */	8,
	/*   2: lines    */	24,
	/*   3: lm       */	ABSENT_NUMERIC,
	/*   4: xmc      */	ABSENT_NUMERIC,
	/*   5: pb       */	ABSENT_NUMERIC,
	/*   6: vt       */	ABSENT_NUMERIC,
	/*   7: wsl      */	ABSENT_NUMERIC,
	/*   8: nlab     */	ABSENT_NUMERIC,
	/*   9: lh       */	ABSENT_NUMERIC,
	/*  10: lw       */	ABSENT_NUMERIC,
	/*  11: ma       */	ABSENT_NUMERIC,
	/*  12: wnum     */	ABSENT_NUMERIC,
	/*  13: colors   */	8,
	/*  14: pairs    */	64,
	/*  15: ncv      */	ABSENT_NUMERIC,
	/*  16: bufsz    */	ABSENT_NUMERIC,
	/*  17: spinv    */	ABSENT_NUMERIC,
	/*  18: spinh    */	ABSENT_NUMERIC,
	/*  19: maddr    */	ABSENT_NUMERIC,
	/*  20: mjump    */	ABSENT_NUMERIC,
	/*  21: mcs      */	ABSENT_NUMERIC,
	/*  22: mls      */	ABSENT_NUMERIC,
	/*  23: npins    */	ABSENT_NUMERIC,
	/*  24: orc      */	ABSENT_NUMERIC,
	/*  25: orl      */	ABSENT_NUMERIC,
	/*  26: orhi     */	ABSENT_NUMERIC,
	/*  27: orvi     */	ABSENT_NUMERIC,
	/*  28: cps      */	ABSENT_NUMERIC,
	/*  29: widcs    */	ABSENT_NUMERIC,
	/*  30: btns     */	ABSENT_NUMERIC,
	/*  31: bitwin   */	ABSENT_NUMERIC,
	/*  32: bitype   */	ABSENT_NUMERIC,
	/*  33: OTug     */	ABSENT_NUMERIC,
	/*  34: OTdC     */	ABSENT_NUMERIC,
	/*  35: OTdN     */	ABSENT_NUMERIC,
	/*  36: OTdB     */	ABSENT_NUMERIC,
	/*  37: OTdT     */	ABSENT_NUMERIC,
	/*  38: OTkn     */	ABSENT_NUMERIC,
};
static char * screen_xterm_xfree86_string_data[] = {
	/*   0: cbt      */	screen_xterm_xfree86_s_cbt,
	/*   1: bel      */	screen_xterm_xfree86_s_bel,
	/*   2: cr       */	screen_xterm_xfree86_s_cr,
	/*   3: csr      */	screen_xterm_xfree86_s_csr,
	/*   4: tbc      */	screen_xterm_xfree86_s_tbc,
	/*   5: clear    */	screen_xterm_xfree86_s_clear,
	/*   6: el       */	screen_xterm_xfree86_s_el,
	/*   7: ed       */	screen_xterm_xfree86_s_ed,
	/*   8: hpa      */	screen_xterm_xfree86_s_hpa,
	/*   9: cmdch    */	ABSENT_STRING,
	/*  10: cup      */	screen_xterm_xfree86_s_cup,
	/*  11: cud1     */	screen_xterm_xfree86_s_cud1,
	/*  12: home     */	screen_xterm_xfree86_s_home,
	/*  13: civis    */	screen_xterm_xfree86_s_civis,
	/*  14: cub1     */	screen_xterm_xfree86_s_cub1,
	/*  15: mrcup    */	ABSENT_STRING,
	/*  16: cnorm    */	screen_xterm_xfree86_s_cnorm,
	/*  17: cuf1     */	screen_xterm_xfree86_s_cuf1,
	/*  18: ll       */	ABSENT_STRING,
	/*  19: cuu1     */	screen_xterm_xfree86_s_cuu1,
	/*  20: cvvis    */	screen_xterm_xfree86_s_cvvis,
	/*  21: dch1     */	screen_xterm_xfree86_s_dch1,
	/*  22: dl1      */	screen_xterm_xfree86_s_dl1,
	/*  23: dsl      */	ABSENT_STRING,
	/*  24: hd       */	ABSENT_STRING,
	/*  25: smacs    */	screen_xterm_xfree86_s_smacs,
	/*  26: blink    */	screen_xterm_xfree86_s_blink,
	/*  27: bold     */	screen_xterm_xfree86_s_bold,
	/*  28: smcup    */	screen_xterm_xfree86_s_smcup,
	/*  29: smdc     */	ABSENT_STRING,
	/*  30: dim      */	screen_xterm_xfree86_s_dim,
	/*  31: smir     */	screen_xterm_xfree86_s_smir,
	/*  32: invis    */	CANCELLED_STRING,
	/*  33: prot     */	ABSENT_STRING,
	/*  34: rev      */	screen_xterm_xfree86_s_rev,
	/*  35: smso     */	screen_xterm_xfree86_s_smso,
	/*  36: smul     */	screen_xterm_xfree86_s_smul,
	/*  37: ech      */	screen_xterm_xfree86_s_ech,
	/*  38: rmacs    */	screen_xterm_xfree86_s_rmacs,
	/*  39: sgr0     */	screen_xterm_xfree86_s_sgr0,
	/*  40: rmcup    */	screen_xterm_xfree86_s_rmcup,
	/*  41: rmdc     */	ABSENT_STRING,
	/*  42: rmir     */	screen_xterm_xfree86_s_rmir,
	/*  43: rmso     */	screen_xterm_xfree86_s_rmso,
	/*  44: rmul     */	screen_xterm_xfree86_s_rmul,
	/*  45: flash    */	screen_xterm_xfree86_s_flash,
	/*  46: ff       */	ABSENT_STRING,
	/*  47: fsl      */	ABSENT_STRING,
	/*  48: is1      */	ABSENT_STRING,
	/*  49: is2      */	screen_xterm_xfree86_s_is2,
	/*  50: is3      */	ABSENT_STRING,
	/*  51: if       */	ABSENT_STRING,
	/*  52: ich1     */	ABSENT_STRING,
	/*  53: il1      */	screen_xterm_xfree86_s_il1,
	/*  54: ip       */	ABSENT_STRING,
	/*  55: kbs      */	screen_xterm_xfree86_s_kbs,
	/*  56: ktbc     */	ABSENT_STRING,
	/*  57: kclr     */	ABSENT_STRING,
	/*  58: kctab    */	ABSENT_STRING,
	/*  59: kdch1    */	screen_xterm_xfree86_s_kdch1,
	/*  60: kdl1     */	ABSENT_STRING,
	/*  61: kcud1    */	screen_xterm_xfree86_s_kcud1,
	/*  62: krmir    */	ABSENT_STRING,
	/*  63: kel      */	ABSENT_STRING,
	/*  64: ked      */	ABSENT_STRING,
	/*  65: kf0      */	ABSENT_STRING,
	/*  66: kf1      */	screen_xterm_xfree86_s_kf1,
	/*  67: kf10     */	screen_xterm_xfree86_s_kf10,
	/*  68: kf2      */	screen_xterm_xfree86_s_kf2,
	/*  69: kf3      */	screen_xterm_xfree86_s_kf3,
	/*  70: kf4      */	screen_xterm_xfree86_s_kf4,
	/*  71: kf5      */	screen_xterm_xfree86_s_kf5,
	/*  72: kf6      */	screen_xterm_xfree86_s_kf6,
	/*  73: kf7      */	screen_xterm_xfree86_s_kf7,
	/*  74: kf8      */	screen_xterm_xfree86_s_kf8,
	/*  75: kf9      */	screen_xterm_xfree86_s_kf9,
	/*  76: khome    */	screen_xterm_xfree86_s_khome,
	/*  77: kich1    */	screen_xterm_xfree86_s_kich1,
	/*  78: kil1     */	ABSENT_STRING,
	/*  79: kcub1    */	screen_xterm_xfree86_s_kcub1,
	/*  80: kll      */	ABSENT_STRING,
	/*  81: knp      */	screen_xterm_xfree86_s_knp,
	/*  82: kpp      */	screen_xterm_xfree86_s_kpp,
	/*  83: kcuf1    */	screen_xterm_xfree86_s_kcuf1,
	/*  84: kind     */	screen_xterm_xfree86_s_kind,
	/*  85: kri      */	screen_xterm_xfree86_s_kri,
	/*  86: khts     */	ABSENT_STRING,
	/*  87: kcuu1    */	screen_xterm_xfree86_s_kcuu1,
	/*  88: rmkx     */	screen_xterm_xfree86_s_rmkx,
	/*  89: smkx     */	screen_xterm_xfree86_s_smkx,
	/*  90: lf0      */	ABSENT_STRING,
	/*  91: lf1      */	ABSENT_STRING,
	/*  92: lf10     */	ABSENT_STRING,
	/*  93: lf2      */	ABSENT_STRING,
	/*  94: lf3      */	ABSENT_STRING,
	/*  95: lf4      */	ABSENT_STRING,
	/*  96: lf5      */	ABSENT_STRING,
	/*  97: lf6      */	ABSENT_STRING,
	/*  98: lf7      */	ABSENT_STRING,
	/*  99: lf8      */	ABSENT_STRING,
	/* 100: lf9      */	ABSENT_STRING,
	/* 101: rmm      */	screen_xterm_xfree86_s_rmm,
	/* 102: smm      */	screen_xterm_xfree86_s_smm,
	/* 103: nel      */	ABSENT_STRING,
	/* 104: pad      */	ABSENT_STRING,
	/* 105: dch      */	screen_xterm_xfree86_s_dch,
	/* 106: dl       */	screen_xterm_xfree86_s_dl,
	/* 107: cud      */	screen_xterm_xfree86_s_cud,
	/* 108: ich      */	screen_xterm_xfree86_s_ich,
	/* 109: indn     */	screen_xterm_xfree86_s_indn,
	/* 110: il       */	screen_xterm_xfree86_s_il,
	/* 111: cub      */	screen_xterm_xfree86_s_cub,
	/* 112: cuf      */	screen_xterm_xfree86_s_cuf,
	/* 113: rin      */	screen_xterm_xfree86_s_rin,
	/* 114: cuu      */	screen_xterm_xfree86_s_cuu,
	/* 115: pfkey    */	ABSENT_STRING,
	/* 116: pfloc    */	ABSENT_STRING,
	/* 117: pfx      */	ABSENT_STRING,
	/* 118: mc0      */	screen_xterm_xfree86_s_mc0,
	/* 119: mc4      */	screen_xterm_xfree86_s_mc4,
	/* 120: mc5      */	screen_xterm_xfree86_s_mc5,
	/* 121: rep      */	CANCELLED_STRING,
	/* 122: rs1      */	screen_xterm_xfree86_s_rs1,
	/* 123: rs2      */	screen_xterm_xfree86_s_rs2,
	/* 124: rs3      */	ABSENT_STRING,
	/* 125: rf       */	ABSENT_STRING,
	/* 126: rc       */	screen_xterm_xfree86_s_rc,
	/* 127: vpa      */	screen_xterm_xfree86_s_vpa,
	/* 128: sc       */	screen_xterm_xfree86_s_sc,
	/* 129: ind      */	screen_xterm_xfree86_s_ind,
	/* 130: ri       */	screen_xterm_xfree86_s_ri,
	/* 131: sgr      */	screen_xterm_xfree86_s_sgr,
	/* 132: hts      */	screen_xterm_xfree86_s_hts,
	/* 133: wind     */	ABSENT_STRING,
	/* 134: ht       */	screen_xterm_xfree86_s_ht,
	/* 135: tsl      */	ABSENT_STRING,
	/* 136: uc       */	ABSENT_STRING,
	/* 137: hu       */	ABSENT_STRING,
	/* 138: iprog    */	ABSENT_STRING,
	/* 139: ka1      */	ABSENT_STRING,
	/* 140: ka3      */	ABSENT_STRING,
	/* 141: kb2      */	screen_xterm_xfree86_s_kb2,
	/* 142: kc1      */	ABSENT_STRING,
	/* 143: kc3      */	ABSENT_STRING,
	/* 144: mc5p     */	ABSENT_STRING,
	/* 145: rmp      */	ABSENT_STRING,
	/* 146: acsc     */	screen_xterm_xfree86_s_acsc,
	/* 147: pln      */	ABSENT_STRING,
	/* 148: kcbt     */	screen_xterm_xfree86_s_kcbt,
	/* 149: smxon    */	ABSENT_STRING,
	/* 150: rmxon    */	ABSENT_STRING,
	/* 151: smam     */	screen_xterm_xfree86_s_smam,
	/* 152: rmam     */	screen_xterm_xfree86_s_rmam,
	/* 153: xonc     */	ABSENT_STRING,
	/* 154: xoffc    */	ABSENT_STRING,
	/* 155: enacs    */	ABSENT_STRING,
	/* 156: smln     */	ABSENT_STRING,
	/* 157: rmln     */	ABSENT_STRING,
	/* 158: kbeg     */	ABSENT_STRING,
	/* 159: kcan     */	ABSENT_STRING,
	/* 160: kclo     */	ABSENT_STRING,
	/* 161: kcmd     */	ABSENT_STRING,
	/* 162: kcpy     */	ABSENT_STRING,
	/* 163: kcrt     */	ABSENT_STRING,
	/* 164: kend     */	screen_xterm_xfree86_s_kend,
	/* 165: kent     */	screen_xterm_xfree86_s_kent,
	/* 166: kext     */	ABSENT_STRING,
	/* 167: kfnd     */	ABSENT_STRING,
	/* 168: khlp     */	ABSENT_STRING,
	/* 169: kmrk     */	ABSENT_STRING,
	/* 170: kmsg     */	ABSENT_STRING,
	/* 171: kmov     */	ABSENT_STRING,
	/* 172: knxt     */	ABSENT_STRING,
	/* 173: kopn     */	ABSENT_STRING,
	/* 174: kopt     */	ABSENT_STRING,
	/* 175: kprv     */	ABSENT_STRING,
	/* 176: kprt     */	ABSENT_STRING,
	/* 177: krdo     */	ABSENT_STRING,
	/* 178: kref     */	ABSENT_STRING,
	/* 179: krfr     */	ABSENT_STRING,
	/* 180: krpl     */	ABSENT_STRING,
	/* 181: krst     */	ABSENT_STRING,
	/* 182: kres     */	ABSENT_STRING,
	/* 183: ksav     */	ABSENT_STRING,
	/* 184: kspd     */	ABSENT_STRING,
	/* 185: kund     */	ABSENT_STRING,
	/* 186: kBEG     */	ABSENT_STRING,
	/* 187: kCAN     */	ABSENT_STRING,
	/* 188: kCMD     */	ABSENT_STRING,
	/* 189: kCPY     */	ABSENT_STRING,
	/* 190: kCRT     */	ABSENT_STRING,
	/* 191: kDC      */	screen_xterm_xfree86_s_kDC,
	/* 192: kDL      */	ABSENT_STRING,
	/* 193: kslt     */	ABSENT_STRING,
	/* 194: kEND     */	screen_xterm_xfree86_s_kEND,
	/* 195: kEOL     */	ABSENT_STRING,
	/* 196: kEXT     */	ABSENT_STRING,
	/* 197: kFND     */	ABSENT_STRING,
	/* 198: kHLP     */	ABSENT_STRING,
	/* 199: kHOM     */	screen_xterm_xfree86_s_kHOM,
	/* 200: kIC      */	CANCELLED_STRING,
	/* 201: kLFT     */	screen_xterm_xfree86_s_kLFT,
	/* 202: kMSG     */	ABSENT_STRING,
	/* 203: kMOV     */	ABSENT_STRING,
	/* 204: kNXT     */	CANCELLED_STRING,
	/* 205: kOPT     */	ABSENT_STRING,
	/* 206: kPRV     */	CANCELLED_STRING,
	/* 207: kPRT     */	ABSENT_STRING,
	/* 208: kRDO     */	ABSENT_STRING,
	/* 209: kRPL     */	ABSENT_STRING,
	/* 210: kRIT     */	screen_xterm_xfree86_s_kRIT,
	/* 211: kRES     */	ABSENT_STRING,
	/* 212: kSAV     */	ABSENT_STRING,
	/* 213: kSPD     */	ABSENT_STRING,
	/* 214: kUND     */	ABSENT_STRING,
	/* 215: rfi      */	ABSENT_STRING,
	/* 216: kf11     */	screen_xterm_xfree86_s_kf11,
	/* 217: kf12     */	screen_xterm_xfree86_s_kf12,
	/* 218: kf13     */	screen_xterm_xfree86_s_kf13,
	/* 219: kf14     */	screen_xterm_xfree86_s_kf14,
	/* 220: kf15     */	screen_xterm_xfree86_s_kf15,
	/* 221: kf16     */	screen_xterm_xfree86_s_kf16,
	/* 222: kf17     */	screen_xterm_xfree86_s_kf17,
	/* 223: kf18     */	screen_xterm_xfree86_s_kf18,
	/* 224: kf19     */	screen_xterm_xfree86_s_kf19,
	/* 225: kf20     */	screen_xterm_xfree86_s_kf20,
	/* 226: kf21     */	screen_xterm_xfree86_s_kf21,
	/* 227: kf22     */	screen_xterm_xfree86_s_kf22,
	/* 228: kf23     */	screen_xterm_xfree86_s_kf23,
	/* 229: kf24     */	screen_xterm_xfree86_s_kf24,
	/* 230: kf25     */	screen_xterm_xfree86_s_kf25,
	/* 231: kf26     */	screen_xterm_xfree86_s_kf26,
	/* 232: kf27     */	screen_xterm_xfree86_s_kf27,
	/* 233: kf28     */	screen_xterm_xfree86_s_kf28,
	/* 234: kf29     */	screen_xterm_xfree86_s_kf29,
	/* 235: kf30     */	screen_xterm_xfree86_s_kf30,
	/* 236: kf31     */	screen_xterm_xfree86_s_kf31,
	/* 237: kf32     */	screen_xterm_xfree86_s_kf32,
	/* 238: kf33     */	screen_xterm_xfree86_s_kf33,
	/* 239: kf34     */	screen_xterm_xfree86_s_kf34,
	/* 240: kf35     */	screen_xterm_xfree86_s_kf35,
	/* 241: kf36     */	screen_xterm_xfree86_s_kf36,
	/* 242: kf37     */	screen_xterm_xfree86_s_kf37,
	/* 243: kf38     */	screen_xterm_xfree86_s_kf38,
	/* 244: kf39     */	screen_xterm_xfree86_s_kf39,
	/* 245: kf40     */	screen_xterm_xfree86_s_kf40,
	/* 246: kf41     */	screen_xterm_xfree86_s_kf41,
	/* 247: kf42     */	screen_xterm_xfree86_s_kf42,
	/* 248: kf43     */	screen_xterm_xfree86_s_kf43,
	/* 249: kf44     */	screen_xterm_xfree86_s_kf44,
	/* 250: kf45     */	screen_xterm_xfree86_s_kf45,
	/* 251: kf46     */	screen_xterm_xfree86_s_kf46,
	/* 252: kf47     */	screen_xterm_xfree86_s_kf47,
	/* 253: kf48     */	screen_xterm_xfree86_s_kf48,
	/* 254: kf49     */	screen_xterm_xfree86_s_kf49,
	/* 255: kf50     */	screen_xterm_xfree86_s_kf50,
	/* 256: kf51     */	screen_xterm_xfree86_s_kf51,
	/* 257: kf52     */	screen_xterm_xfree86_s_kf52,
	/* 258: kf53     */	screen_xterm_xfree86_s_kf53,
	/* 259: kf54     */	screen_xterm_xfree86_s_kf54,
	/* 260: kf55     */	screen_xterm_xfree86_s_kf55,
	/* 261: kf56     */	screen_xterm_xfree86_s_kf56,
	/* 262: kf57     */	screen_xterm_xfree86_s_kf57,
	/* 263: kf58     */	screen_xterm_xfree86_s_kf58,
	/* 264: kf59     */	screen_xterm_xfree86_s_kf59,
	/* 265: kf60     */	screen_xterm_xfree86_s_kf60,
	/* 266: kf61     */	screen_xterm_xfree86_s_kf61,
	/* 267: kf62     */	screen_xterm_xfree86_s_kf62,
	/* 268: kf63     */	screen_xterm_xfree86_s_kf63,
	/* 269: el1      */	screen_xterm_xfree86_s_el1,
	/* 270: mgc      */	ABSENT_STRING,
	/* 271: smgl     */	ABSENT_STRING,
	/* 272: smgr     */	ABSENT_STRING,
	/* 273: fln      */	ABSENT_STRING,
	/* 274: sclk     */	ABSENT_STRING,
	/* 275: dclk     */	ABSENT_STRING,
	/* 276: rmclk    */	ABSENT_STRING,
	/* 277: cwin     */	ABSENT_STRING,
	/* 278: wingo    */	ABSENT_STRING,
	/* 279: hup      */	ABSENT_STRING,
	/* 280: dial     */	ABSENT_STRING,
	/* 281: qdial    */	ABSENT_STRING,
	/* 282: tone     */	ABSENT_STRING,
	/* 283: pulse    */	ABSENT_STRING,
	/* 284: hook     */	ABSENT_STRING,
	/* 285: pause    */	ABSENT_STRING,
	/* 286: wait     */	ABSENT_STRING,
	/* 287: u0       */	ABSENT_STRING,
	/* 288: u1       */	ABSENT_STRING,
	/* 289: u2       */	ABSENT_STRING,
	/* 290: u3       */	ABSENT_STRING,
	/* 291: u4       */	ABSENT_STRING,
	/* 292: u5       */	ABSENT_STRING,
	/* 293: u6       */	screen_xterm_xfree86_s_u6,
	/* 294: u7       */	screen_xterm_xfree86_s_u7,
	/* 295: u8       */	screen_xterm_xfree86_s_u8,
	/* 296: u9       */	screen_xterm_xfree86_s_u9,
	/* 297: op       */	screen_xterm_xfree86_s_op,
	/* 298: oc       */	ABSENT_STRING,
	/* 299: initc    */	ABSENT_STRING,
	/* 300: initp    */	ABSENT_STRING,
	/* 301: scp      */	ABSENT_STRING,
	/* 302: setf     */	screen_xterm_xfree86_s_setf,
	/* 303: setb     */	screen_xterm_xfree86_s_setb,
	/* 304: cpi      */	ABSENT_STRING,
	/* 305: lpi      */	ABSENT_STRING,
	/* 306: chr      */	ABSENT_STRING,
	/* 307: cvr      */	ABSENT_STRING,
	/* 308: defc     */	ABSENT_STRING,
	/* 309: swidm    */	ABSENT_STRING,
	/* 310: sdrfq    */	ABSENT_STRING,
	/* 311: sitm     */	ABSENT_STRING,
	/* 312: slm      */	ABSENT_STRING,
	/* 313: smicm    */	ABSENT_STRING,
	/* 314: snlq     */	ABSENT_STRING,
	/* 315: snrmq    */	ABSENT_STRING,
	/* 316: sshm     */	ABSENT_STRING,
	/* 317: ssubm    */	ABSENT_STRING,
	/* 318: ssupm    */	ABSENT_STRING,
	/* 319: sum      */	ABSENT_STRING,
	/* 320: rwidm    */	ABSENT_STRING,
	/* 321: ritm     */	ABSENT_STRING,
	/* 322: rlm      */	ABSENT_STRING,
	/* 323: rmicm    */	ABSENT_STRING,
	/* 324: rshm     */	ABSENT_STRING,
	/* 325: rsubm    */	ABSENT_STRING,
	/* 326: rsupm    */	ABSENT_STRING,
	/* 327: rum      */	ABSENT_STRING,
	/* 328: mhpa     */	ABSENT_STRING,
	/* 329: mcud1    */	ABSENT_STRING,
	/* 330: mcub1    */	ABSENT_STRING,
	/* 331: mcuf1    */	ABSENT_STRING,
	/* 332: mvpa     */	ABSENT_STRING,
	/* 333: mcuu1    */	ABSENT_STRING,
	/* 334: porder   */	ABSENT_STRING,
	/* 335: mcud     */	ABSENT_STRING,
	/* 336: mcub     */	ABSENT_STRING,
	/* 337: mcuf     */	ABSENT_STRING,
	/* 338: mcuu     */	ABSENT_STRING,
	/* 339: scs      */	ABSENT_STRING,
	/* 340: smgb     */	ABSENT_STRING,
	/* 341: smgbp    */	ABSENT_STRING,
	/* 342: smglp    */	ABSENT_STRING,
	/* 343: smgrp    */	ABSENT_STRING,
	/* 344: smgt     */	ABSENT_STRING,
	/* 345: smgtp    */	ABSENT_STRING,
	/* 346: sbim     */	ABSENT_STRING,
	/* 347: scsd     */	ABSENT_STRING,
	/* 348: rbim     */	ABSENT_STRING,
	/* 349: rcsd     */	ABSENT_STRING,
	/* 350: subcs    */	ABSENT_STRING,
	/* 351: supcs    */	ABSENT_STRING,
	/* 352: docr     */	ABSENT_STRING,
	/* 353: zerom    */	ABSENT_STRING,
	/* 354: csnm     */	ABSENT_STRING,
	/* 355: kmous    */	screen_xterm_xfree86_s_kmous,
	/* 356: minfo    */	ABSENT_STRING,
	/* 357: reqmp    */	ABSENT_STRING,
	/* 358: getm     */	ABSENT_STRING,
	/* 359: setaf    */	screen_xterm_xfree86_s_setaf,
	/* 360: setab    */	screen_xterm_xfree86_s_setab,
	/* 361: pfxl     */	ABSENT_STRING,
	/* 362: devt     */	ABSENT_STRING,
	/* 363: csin     */	ABSENT_STRING,
	/* 364: s0ds     */	ABSENT_STRING,
	/* 365: s1ds     */	ABSENT_STRING,
	/* 366: s2ds     */	ABSENT_STRING,
	/* 367: s3ds     */	ABSENT_STRING,
	/* 368: smglr    */	ABSENT_STRING,
	/* 369: smgtb    */	ABSENT_STRING,
	/* 370: birep    */	ABSENT_STRING,
	/* 371: binel    */	ABSENT_STRING,
	/* 372: bicr     */	ABSENT_STRING,
	/* 373: colornm  */	ABSENT_STRING,
	/* 374: defbi    */	ABSENT_STRING,
	/* 375: endbi    */	ABSENT_STRING,
	/* 376: setcolor */	ABSENT_STRING,
	/* 377: slines   */	ABSENT_STRING,
	/* 378: dispc    */	ABSENT_STRING,
	/* 379: smpch    */	ABSENT_STRING,
	/* 380: rmpch    */	ABSENT_STRING,
	/* 381: smsc     */	ABSENT_STRING,
	/* 382: rmsc     */	ABSENT_STRING,
	/* 383: pctrm    */	ABSENT_STRING,
	/* 384: scesc    */	ABSENT_STRING,
	/* 385: scesa    */	ABSENT_STRING,
	/* 386: ehhlm    */	ABSENT_STRING,
	/* 387: elhlm    */	ABSENT_STRING,
	/* 388: elohlm   */	ABSENT_STRING,
	/* 389: erhlm    */	ABSENT_STRING,
	/* 390: ethlm    */	ABSENT_STRING,
	/* 391: evhlm    */	ABSENT_STRING,
	/* 392: sgr1     */	ABSENT_STRING,
	/* 393: slength  */	ABSENT_STRING,
	/* 394: OTi2     */	ABSENT_STRING,
	/* 395: OTrs     */	ABSENT_STRING,
	/* 396: OTnl     */	ABSENT_STRING,
	/* 397: OTbc     */	ABSENT_STRING,
	/* 398: OTko     */	ABSENT_STRING,
	/* 399: OTma     */	ABSENT_STRING,
	/* 400: OTG2     */	ABSENT_STRING,
	/* 401: OTG3     */	ABSENT_STRING,
	/* 402: OTG1     */	ABSENT_STRING,
	/* 403: OTG4     */	ABSENT_STRING,
	/* 404: OTGR     */	ABSENT_STRING,
	/* 405: OTGL     */	ABSENT_STRING,
	/* 406: OTGU     */	ABSENT_STRING,
	/* 407: OTGD     */	ABSENT_STRING,
	/* 408: OTGH     */	ABSENT_STRING,
	/* 409: OTGV     */	ABSENT_STRING,
	/* 410: OTGC     */	ABSENT_STRING,
	/* 411: meml     */	CANCELLED_STRING,
	/* 412: memu     */	CANCELLED_STRING,
	/* 413: box1     */	ABSENT_STRING,
};
/* screen.xterm-256color */

static char screen_xterm_256color_alias_data[] = "screen.xterm-256color|GNU Screen with xterm using 256 colors";

static char screen_xterm_256color_s_cbt[] = "\033[Z";
static char screen_xterm_256color_s_bel[] = "\007";
static char screen_xterm_256color_s_cr[] = "\015";
static char screen_xterm_256color_s_csr[] = "\033[%i%p1%d;%p2%dr";
static char screen_xterm_256color_s_tbc[] = "\033[3g";
static char screen_xterm_256color_s_clear[] = "\033[H\033[2J";
static char screen_xterm_256color_s_el[] = "\033[K";
static char screen_xterm_256color_s_ed[] = "\033[J";
static char screen_xterm_256color_s_hpa[] = "\033[%i%p1%dG";
static char screen_xterm_256color_s_cup[] = "\033[%i%p1%d;%p2%dH";
static char screen_xterm_256color_s_cud1[] = "\012";
static char screen_xterm_256color_s_home[] = "\033[H";
static char screen_xterm_256color_s_civis[] = "\033[?25l";
static char screen_xterm_256color_s_cub1[] = "\010";
static char screen_xterm_256color_s_cnorm[] = "\033[?12l\033[?25h";
static char screen_xterm_256color_s_cuf1[] = "\033[C";
static char screen_xterm_256color_s_cuu1[] = "\033[A";
static char screen_xterm_256color_s_cvvis[] = "\033[?12;25h";
static char screen_xterm_256color_s_dch1[] = "\033[P";
static char screen_xterm_256color_s_dl1[] = "\033[M";
static char screen_xterm_256color_s_smacs[] = "\033(0";
static char screen_xterm_256color_s_blink[] = "\033[5m";
static char screen_xterm_256color_s_bold[] = "\033[1m";
static char screen_xterm_256color_s_smcup[] = "\033[?1049h\033[22;0;0t";
static char screen_xterm_256color_s_dim[] = "\033[2m";
static char screen_xterm_256color_s_smir[] = "\033[4h";
static char screen_xterm_256color_s_rev[] = "\033[7m";
static char screen_xterm_256color_s_smso[] = "\033[7m";
static char screen_xterm_256color_s_smul[] = "\033[4m";
static char screen_xterm_256color_s_ech[] = "\033[%p1%dX";
static char screen_xterm_256color_s_rmacs[] = "\033(B";
static char screen_xterm_256color_s_sgr0[] = "\033(B\033[m";
static char screen_xterm_256color_s_rmcup[] = "\033[?1049l\033[23;0;0t";
static char screen_xterm_256color_s_rmir[] = "\033[4l";
static char screen_xterm_256color_s_rmso[] = "\033[27m";
static char screen_xterm_256color_s_rmul[] = "\033[24m";
static char screen_xterm_256color_s_flash[] = "\033[?5h$<100/>\033[?5l";
static char screen_xterm_256color_s_is2[] = "\033[!p\033[?3;4l\033[4l\033>";
static char screen_xterm_256color_s_il1[] = "\033[L";
static char screen_xterm_256color_s_kbs[] = "\010";
static char screen_xterm_256color_s_kdch1[] = "\033[3~";
static char screen_xterm_256color_s_kcud1[] = "\033OB";
static char screen_xterm_256color_s_kf1[] = "\033OP";
static char screen_xterm_256color_s_kf10[] = "\033[21~";
static char screen_xterm_256color_s_kf2[] = "\033OQ";
static char screen_xterm_256color_s_kf3[] = "\033OR";
static char screen_xterm_256color_s_kf4[] = "\033OS";
static char screen_xterm_256color_s_kf5[] = "\033[15~";
static char screen_xterm_256color_s_kf6[] = "\033[17~";
static char screen_xterm_256color_s_kf7[] = "\033[18~";
static char screen_xterm_256color_s_kf8[] = "\033[19~";
static char screen_xterm_256color_s_kf9[] = "\033[20~";
static char screen_xterm_256color_s_khome[] = "\033[1~";
static char screen_xterm_256color_s_kich1[] = "\033[2~";
static char screen_xterm_256color_s_kcub1[] = "\033OD";
static char screen_xterm_256color_s_knp[] = "\033[6~";
static char screen_xterm_256color_s_kpp[] = "\033[5~";
static char screen_xterm_256color_s_kcuf1[] = "\033OC";
static char screen_xterm_256color_s_kind[] = "\033[1;2B";
static char screen_xterm_256color_s_kri[] = "\033[1;2A";
static char screen_xterm_256color_s_kcuu1[] = "\033OA";
static char screen_xterm_256color_s_rmkx[] = "\033[?1l\033>";
static char screen_xterm_256color_s_smkx[] = "\033[?1h\033=";
static char screen_xterm_256color_s_rmm[] = "\033[?1034l";
static char screen_xterm_256color_s_smm[] = "\033[?1034h";
static char screen_xterm_256color_s_dch[] = "\033[%p1%dP";
static char screen_xterm_256color_s_dl[] = "\033[%p1%dM";
static char screen_xterm_256color_s_cud[] = "\033[%p1%dB";
static char screen_xterm_256color_s_ich[] = "\033[%p1%d@";
static char screen_xterm_256color_s_indn[] = "\033[%p1%dS";
static char screen_xterm_256color_s_il[] = "\033[%p1%dL";
static char screen_xterm_256color_s_cub[] = "\033[%p1%dD";
static char screen_xterm_256color_s_cuf[] = "\033[%p1%dC";
static char screen_xterm_256color_s_rin[] = "\033[%p1%dT";
static char screen_xterm_256color_s_cuu[] = "\033[%p1%dA";
static char screen_xterm_256color_s_mc0[] = "\033[i";
static char screen_xterm_256color_s_mc4[] = "\033[4i";
static char screen_xterm_256color_s_mc5[] = "\033[5i";
static char screen_xterm_256color_s_rs1[] = "\033c";
static char screen_xterm_256color_s_rs2[] = "\033[!p\033[?3;4l\033[4l\033>";
static char screen_xterm_256color_s_rc[] = "\0338";
static char screen_xterm_256color_s_vpa[] = "\033[%i%p1%dd";
static char screen_xterm_256color_s_sc[] = "\0337";
static char screen_xterm_256color_s_ind[] = "\012";
static char screen_xterm_256color_s_ri[] = "\033M";
static char screen_xterm_256color_s_sgr[] = "%?%p9%t\033(0%e\033(B%;\033[0%?%p6%t;1%;%?%p2%t;4%;%?%p1%p3%|%t;7%;%?%p4%t;5%;%?%p5%t;2%;m";
static char screen_xterm_256color_s_hts[] = "\033H";
static char screen_xterm_256color_s_ht[] = "\011";
static char screen_xterm_256color_s_kb2[] = "\033OE";
static char screen_xterm_256color_s_acsc[] = "``aaffggiijjkkllmmnnooppqqrrssttuuvvwwxxyyzz{{||}}~~";
static char screen_xterm_256color_s_kcbt[] = "\033[Z";
static char screen_xterm_256color_s_smam[] = "\033[?7h";
static char screen_xterm_256color_s_rmam[] = "\033[?7l";
static char screen_xterm_256color_s_kend[] = "\033[4~";
static char screen_xterm_256color_s_kent[] = "\033OM";
static char screen_xterm_256color_s_kDC[] = "\033[3;2~";
static char screen_xterm_256color_s_kEND[] = "\033[1;2F";
static char screen_xterm_256color_s_kHOM[] = "\033[1;2H";
static char screen_xterm_256color_s_kLFT[] = "\033[1;2D";
static char screen_xterm_256color_s_kRIT[] = "\033[1;2C";
static char screen_xterm_256color_s_kf11[] = "\033[23~";
static char screen_xterm_256color_s_kf12[] = "\033[24~";
static char screen_xterm_256color_s_kf13[] = "\033[1;2P";
static char screen_xterm_256color_s_kf14[] = "\033[1;2Q";
static char screen_xterm_256color_s_kf15[] = "\033[1;2R";
static char screen_xterm_256color_s_kf16[] = "\033[1;2S";
static char screen_xterm_256color_s_kf17[] = "\033[15;2~";
static char screen_xterm_256color_s_kf18[] = "\033[17;2~";
static char screen_xterm_256color_s_kf19[] = "\033[18;2~";
static char screen_xterm_256color_s_kf20[] = "\033[19;2~";
static char screen_xterm_256color_s_kf21[] = "\033[20;2~";
static char screen_xterm_256color_s_kf22[] = "\033[21;2~";
static char screen_xterm_256color_s_kf23[] = "\033[23;2~";
static char screen_xterm_256color_s_kf24[] = "\033[24;2~";
static char screen_xterm_256color_s_kf25[] = "\033[1;5P";
static char screen_xterm_256color_s_kf26[] = "\033[1;5Q";
static char screen_xterm_256color_s_kf27[] = "\033[1;5R";
static char screen_xterm_256color_s_kf28[] = "\033[1;5S";
static char screen_xterm_256color_s_kf29[] = "\033[15;5~";
static char screen_xterm_256color_s_kf30[] = "\033[17;5~";
static char screen_xterm_256color_s_kf31[] = "\033[18;5~";
static char screen_xterm_256color_s_kf32[] = "\033[19;5~";
static char screen_xterm_256color_s_kf33[] = "\033[20;5~";
static char screen_xterm_256color_s_kf34[] = "\033[21;5~";
static char screen_xterm_256color_s_kf35[] = "\033[23;5~";
static char screen_xterm_256color_s_kf36[] = "\033[24;5~";
static char screen_xterm_256color_s_kf37[] = "\033[1;6P";
static char screen_xterm_256color_s_kf38[] = "\033[1;6Q";
static char screen_xterm_256color_s_kf39[] = "\033[1;6R";
static char screen_xterm_256color_s_kf40[] = "\033[1;6S";
static char screen_xterm_256color_s_kf41[] = "\033[15;6~";
static char screen_xterm_256color_s_kf42[] = "\033[17;6~";
static char screen_xterm_256color_s_kf43[] = "\033[18;6~";
static char screen_xterm_256color_s_kf44[] = "\033[19;6~";
static char screen_xterm_256color_s_kf45[] = "\033[20;6~";
static char screen_xterm_256color_s_kf46[] = "\033[21;6~";
static char screen_xterm_256color_s_kf47[] = "\033[23;6~";
static char screen_xterm_256color_s_kf48[] = "\033[24;6~";
static char screen_xterm_256color_s_kf49[] = "\033[1;3P";
static char screen_xterm_256color_s_kf50[] = "\033[1;3Q";
static char screen_xterm_256color_s_kf51[] = "\033[1;3R";
static char screen_xterm_256color_s_kf52[] = "\033[1;3S";
static char screen_xterm_256color_s_kf53[] = "\033[15;3~";
static char screen_xterm_256color_s_kf54[] = "\033[17;3~";
static char screen_xterm_256color_s_kf55[] = "\033[18;3~";
static char screen_xterm_256color_s_kf56[] = "\033[19;3~";
static char screen_xterm_256color_s_kf57[] = "\033[20;3~";
static char screen_xterm_256color_s_kf58[] = "\033[21;3~";
static char screen_xterm_256color_s_kf59[] = "\033[23;3~";
static char screen_xterm_256color_s_kf60[] = "\033[24;3~";
static char screen_xterm_256color_s_kf61[] = "\033[1;4P";
static char screen_xterm_256color_s_kf62[] = "\033[1;4Q";
static char screen_xterm_256color_s_kf63[] = "\033[1;4R";
static char screen_xterm_256color_s_el1[] = "\033[1K";
static char screen_xterm_256color_s_u6[] = "\033[%i%d;%dR";
static char screen_xterm_256color_s_u7[] = "\033[6n";
static char screen_xterm_256color_s_u8[] = "\033[?%[;0123456789]c";
static char screen_xterm_256color_s_u9[] = "\033[c";
static char screen_xterm_256color_s_op[] = "\033[39;49m";
static char screen_xterm_256color_s_kmous[] = "\033[M";
static char screen_xterm_256color_s_setaf[] = "\033[%?%p1%{8}%<%t3%p1%d%e%p1%{16}%<%t9%p1%{8}%-%d%e38;5;%p1%d%;m";
static char screen_xterm_256color_s_setab[] = "\033[%?%p1%{8}%<%t4%p1%d%e%p1%{16}%<%t10%p1%{8}%-%d%e48;5;%p1%d%;m";

static char screen_xterm_256color_bool_data[] = {
	/*   0: bw       */	TRUE,
	/*   1: am       */	TRUE,
	/*   2: xsb      */	FALSE,
	/*   3: xhp      */	FALSE,
	/*   4: xenl     */	TRUE,
	/*   5: eo       */	FALSE,
	/*   6: gn       */	FALSE,
	/*   7: hc       */	FALSE,
	/*   8: km       */	TRUE,
	/*   9: hs       */	FALSE,
	/*  10: in       */	FALSE,
	/*  11: da       */	FALSE,
	/*  12: db       */	FALSE,
	/*  13: mir      */	TRUE,
	/*  14: msgr     */	TRUE,
	/*  15: os       */	FALSE,
	/*  16: eslok    */	FALSE,
	/*  17: xt       */	FALSE,
	/*  18: hz       */	FALSE,
	/*  19: ul       */	FALSE,
	/*  20: xon      */	FALSE,
	/*  21: nxon     */	FALSE,
	/*  22: mc5i     */	TRUE,
	/*  23: chts     */	FALSE,
	/*  24: nrrmc    */	FALSE,
	/*  25: npc      */	TRUE,
	/*  26: ndscr    */	FALSE,
	/*  27: ccc      */	FALSE,
	/*  28: bce      */	FALSE,
	/*  29: hls      */	FALSE,
	/*  30: xhpa     */	FALSE,
	/*  31: crxm     */	FALSE,
	/*  32: daisy    */	FALSE,
	/*  33: xvpa     */	FALSE,
	/*  34: sam      */	FALSE,
	/*  35: cpix     */	FALSE,
	/*  36: lpix     */	FALSE,
	/*  37: OTbs     */	TRUE,
	/*  38: OTns     */	FALSE,
	/*  39: OTnc     */	FALSE,
	/*  40: OTMT     */	FALSE,
	/*  41: OTNL     */	FALSE,
	/*  42: OTpt     */	FALSE,
	/*  43: OTxr     */	FALSE,
};
static NCURSES_INT2 screen_xterm_256color_number_data[] = {
	/*   0: cols     */	80,
	/*   1: it       */	8,
	/*   2: lines    */	24,
	/*   3: lm       */	ABSENT_NUMERIC,
	/*   4: xmc      */	ABSENT_NUMERIC,
	/*   5: pb       */	ABSENT_NUMERIC,
	/*   6: vt       */	ABSENT_NUMERIC,
	/*   7: wsl      */	ABSENT_NUMERIC,
	/*   8: nlab     */	ABSENT_NUMERIC,
	/*   9: lh       */	ABSENT_NUMERIC,
	/*  10: lw       */	ABSENT_NUMERIC,
	/*  11: ma       */	ABSENT_NUMERIC,
	/*  12: wnum     */	ABSENT_NUMERIC,
	/*  13: colors   */	256,
	/*  14: pairs    */	32767,
	/*  15: ncv      */	ABSENT_NUMERIC,
	/*  16: bufsz    */	ABSENT_NUMERIC,
	/*  17: spinv    */	ABSENT_NUMERIC,
	/*  18: spinh    */	ABSENT_NUMERIC,
	/*  19: maddr    */	ABSENT_NUMERIC,
	/*  20: mjump    */	ABSENT_NUMERIC,
	/*  21: mcs      */	ABSENT_NUMERIC,
	/*  22: mls      */	ABSENT_NUMERIC,
	/*  23: npins    */	ABSENT_NUMERIC,
	/*  24: orc      */	ABSENT_NUMERIC,
	/*  25: orl      */	ABSENT_NUMERIC,
	/*  26: orhi     */	ABSENT_NUMERIC,
	/*  27: orvi     */	ABSENT_NUMERIC,
	/*  28: cps      */	ABSENT_NUMERIC,
	/*  29: widcs    */	ABSENT_NUMERIC,
	/*  30: btns     */	ABSENT_NUMERIC,
	/*  31: bitwin   */	ABSENT_NUMERIC,
	/*  32: bitype   */	ABSENT_NUMERIC,
	/*  33: OTug     */	ABSENT_NUMERIC,
	/*  34: OTdC     */	ABSENT_NUMERIC,
	/*  35: OTdN     */	ABSENT_NUMERIC,
	/*  36: OTdB     */	ABSENT_NUMERIC,
	/*  37: OTdT     */	ABSENT_NUMERIC,
	/*  38: OTkn     */	ABSENT_NUMERIC,
};
static char * screen_xterm_256color_string_data[] = {
	/*   0: cbt      */	screen_xterm_256color_s_cbt,
	/*   1: bel      */	screen_xterm_256color_s_bel,
	/*   2: cr       */	screen_xterm_256color_s_cr,
	/*   3: csr      */	screen_xterm_256color_s_csr,
	/*   4: tbc      */	screen_xterm_256color_s_tbc,
	/*   5: clear    */	screen_xterm_256color_s_clear,
	/*   6: el       */	screen_xterm_256color_s_el,
	/*   7: ed       */	screen_xterm_256color_s_ed,
	/*   8: hpa      */	screen_xterm_256color_s_hpa,
	/*   9: cmdch    */	ABSENT_STRING,
	/*  10: cup      */	screen_xterm_256color_s_cup,
	/*  11: cud1     */	screen_xterm_256color_s_cud1,
	/*  12: home     */	screen_xterm_256color_s_home,
	/*  13: civis    */	screen_xterm_256color_s_civis,
	/*  14: cub1     */	screen_xterm_256color_s_cub1,
	/*  15: mrcup    */	ABSENT_STRING,
	/*  16: cnorm    */	screen_xterm_256color_s_cnorm,
	/*  17: cuf1     */	screen_xterm_256color_s_cuf1,
	/*  18: ll       */	ABSENT_STRING,
	/*  19: cuu1     */	screen_xterm_256color_s_cuu1,
	/*  20: cvvis    */	screen_xterm_256color_s_cvvis,
	/*  21: dch1     */	screen_xterm_256color_s_dch1,
	/*  22: dl1      */	screen_xterm_256color_s_dl1,
	/*  23: dsl      */	ABSENT_STRING,
	/*  24: hd       */	ABSENT_STRING,
	/*  25: smacs    */	screen_xterm_256color_s_smacs,
	/*  26: blink    */	screen_xterm_256color_s_blink,
	/*  27: bold     */	screen_xterm_256color_s_bold,
	/*  28: smcup    */	screen_xterm_256color_s_smcup,
	/*  29: smdc     */	ABSENT_STRING,
	/*  30: dim      */	screen_xterm_256color_s_dim,
	/*  31: smir     */	screen_xterm_256color_s_smir,
	/*  32: invis    */	ABSENT_STRING,
	/*  33: prot     */	ABSENT_STRING,
	/*  34: rev      */	screen_xterm_256color_s_rev,
	/*  35: smso     */	screen_xterm_256color_s_smso,
	/*  36: smul     */	screen_xterm_256color_s_smul,
	/*  37: ech      */	screen_xterm_256color_s_ech,
	/*  38: rmacs    */	screen_xterm_256color_s_rmacs,
	/*  39: sgr0     */	screen_xterm_256color_s_sgr0,
	/*  40: rmcup    */	screen_xterm_256color_s_rmcup,
	/*  41: rmdc     */	ABSENT_STRING,
	/*  42: rmir     */	screen_xterm_256color_s_rmir,
	/*  43: rmso     */	screen_xterm_256color_s_rmso,
	/*  44: rmul     */	screen_xterm_256color_s_rmul,
	/*  45: flash    */	screen_xterm_256color_s_flash,
	/*  46: ff       */	ABSENT_STRING,
	/*  47: fsl      */	ABSENT_STRING,
	/*  48: is1      */	ABSENT_STRING,
	/*  49: is2      */	screen_xterm_256color_s_is2,
	/*  50: is3      */	ABSENT_STRING,
	/*  51: if       */	ABSENT_STRING,
	/*  52: ich1     */	ABSENT_STRING,
	/*  53: il1      */	screen_xterm_256color_s_il1,
	/*  54: ip       */	ABSENT_STRING,
	/*  55: kbs      */	screen_xterm_256color_s_kbs,
	/*  56: ktbc     */	ABSENT_STRING,
	/*  57: kclr     */	ABSENT_STRING,
	/*  58: kctab    */	ABSENT_STRING,
	/*  59: kdch1    */	screen_xterm_256color_s_kdch1,
	/*  60: kdl1     */	ABSENT_STRING,
	/*  61: kcud1    */	screen_xterm_256color_s_kcud1,
	/*  62: krmir    */	ABSENT_STRING,
	/*  63: kel      */	ABSENT_STRING,
	/*  64: ked      */	ABSENT_STRING,
	/*  65: kf0      */	ABSENT_STRING,
	/*  66: kf1      */	screen_xterm_256color_s_kf1,
	/*  67: kf10     */	screen_xterm_256color_s_kf10,
	/*  68: kf2      */	screen_xterm_256color_s_kf2,
	/*  69: kf3      */	screen_xterm_256color_s_kf3,
	/*  70: kf4      */	screen_xterm_256color_s_kf4,
	/*  71: kf5      */	screen_xterm_256color_s_kf5,
	/*  72: kf6      */	screen_xterm_256color_s_kf6,
	/*  73: kf7      */	screen_xterm_256color_s_kf7,
	/*  74: kf8      */	screen_xterm_256color_s_kf8,
	/*  75: kf9      */	screen_xterm_256color_s_kf9,
	/*  76: khome    */	screen_xterm_256color_s_khome,
	/*  77: kich1    */	screen_xterm_256color_s_kich1,
	/*  78: kil1     */	ABSENT_STRING,
	/*  79: kcub1    */	screen_xterm_256color_s_kcub1,
	/*  80: kll      */	ABSENT_STRING,
	/*  81: knp      */	screen_xterm_256color_s_knp,
	/*  82: kpp      */	screen_xterm_256color_s_kpp,
	/*  83: kcuf1    */	screen_xterm_256color_s_kcuf1,
	/*  84: kind     */	screen_xterm_256color_s_kind,
	/*  85: kri      */	screen_xterm_256color_s_kri,
	/*  86: khts     */	ABSENT_STRING,
	/*  87: kcuu1    */	screen_xterm_256color_s_kcuu1,
	/*  88: rmkx     */	screen_xterm_256color_s_rmkx,
	/*  89: smkx     */	screen_xterm_256color_s_smkx,
	/*  90: lf0      */	ABSENT_STRING,
	/*  91: lf1      */	ABSENT_STRING,
	/*  92: lf10     */	ABSENT_STRING,
	/*  93: lf2      */	ABSENT_STRING,
	/*  94: lf3      */	ABSENT_STRING,
	/*  95: lf4      */	ABSENT_STRING,
	/*  96: lf5      */	ABSENT_STRING,
	/*  97: lf6      */	ABSENT_STRING,
	/*  98: lf7      */	ABSENT_STRING,
	/*  99: lf8      */	ABSENT_STRING,
	/* 100: lf9      */	ABSENT_STRING,
	/* 101: rmm      */	screen_xterm_256color_s_rmm,
	/* 102: smm      */	screen_xterm_256color_s_smm,
	/* 103: nel      */	ABSENT_STRING,
	/* 104: pad      */	ABSENT_STRING,
	/* 105: dch      */	screen_xterm_256color_s_dch,
	/* 106: dl       */	screen_xterm_256color_s_dl,
	/* 107: cud      */	screen_xterm_256color_s_cud,
	/* 108: ich      */	screen_xterm_256color_s_ich,
	/* 109: indn     */	screen_xterm_256color_s_indn,
	/* 110: il       */	screen_xterm_256color_s_il,
	/* 111: cub      */	screen_xterm_256color_s_cub,
	/* 112: cuf      */	screen_xterm_256color_s_cuf,
	/* 113: rin      */	screen_xterm_256color_s_rin,
	/* 114: cuu      */	screen_xterm_256color_s_cuu,
	/* 115: pfkey    */	ABSENT_STRING,
	/* 116: pfloc    */	ABSENT_STRING,
	/* 117: pfx      */	ABSENT_STRING,
	/* 118: mc0      */	screen_xterm_256color_s_mc0,
	/* 119: mc4      */	screen_xterm_256color_s_mc4,
	/* 120: mc5      */	screen_xterm_256color_s_mc5,
	/* 121: rep      */	ABSENT_STRING,
	/* 122: rs1      */	screen_xterm_256color_s_rs1,
	/* 123: rs2      */	screen_xterm_256color_s_rs2,
	/* 124: rs3      */	ABSENT_STRING,
	/* 125: rf       */	ABSENT_STRING,
	/* 126: rc       */	screen_xterm_256color_s_rc,
	/* 127: vpa      */	screen_xterm_256color_s_vpa,
	/* 128: sc       */	screen_xterm_256color_s_sc,
	/* 129: ind      */	screen_xterm_256color_s_ind,
	/* 130: ri       */	screen_xterm_256color_s_ri,
	/* 131: sgr      */	screen_xterm_256color_s_sgr,
	/* 132: hts      */	screen_xterm_256color_s_hts,
	/* 133: wind     */	ABSENT_STRING,
	/* 134: ht       */	screen_xterm_256color_s_ht,
	/* 135: tsl      */	ABSENT_STRING,
	/* 136: uc       */	ABSENT_STRING,
	/* 137: hu       */	ABSENT_STRING,
	/* 138: iprog    */	ABSENT_STRING,
	/* 139: ka1      */	ABSENT_STRING,
	/* 140: ka3      */	ABSENT_STRING,
	/* 141: kb2      */	screen_xterm_256color_s_kb2,
	/* 142: kc1      */	ABSENT_STRING,
	/* 143: kc3      */	ABSENT_STRING,
	/* 144: mc5p     */	ABSENT_STRING,
	/* 145: rmp      */	ABSENT_STRING,
	/* 146: acsc     */	screen_xterm_256color_s_acsc,
	/* 147: pln      */	ABSENT_STRING,
	/* 148: kcbt     */	screen_xterm_256color_s_kcbt,
	/* 149: smxon    */	ABSENT_STRING,
	/* 150: rmxon    */	ABSENT_STRING,
	/* 151: smam     */	screen_xterm_256color_s_smam,
	/* 152: rmam     */	screen_xterm_256color_s_rmam,
	/* 153: xonc     */	ABSENT_STRING,
	/* 154: xoffc    */	ABSENT_STRING,
	/* 155: enacs    */	ABSENT_STRING,
	/* 156: smln     */	ABSENT_STRING,
	/* 157: rmln     */	ABSENT_STRING,
	/* 158: kbeg     */	ABSENT_STRING,
	/* 159: kcan     */	ABSENT_STRING,
	/* 160: kclo     */	ABSENT_STRING,
	/* 161: kcmd     */	ABSENT_STRING,
	/* 162: kcpy     */	ABSENT_STRING,
	/* 163: kcrt     */	ABSENT_STRING,
	/* 164: kend     */	screen_xterm_256color_s_kend,
	/* 165: kent     */	screen_xterm_256color_s_kent,
	/* 166: kext     */	ABSENT_STRING,
	/* 167: kfnd     */	ABSENT_STRING,
	/* 168: khlp     */	ABSENT_STRING,
	/* 169: kmrk     */	ABSENT_STRING,
	/* 170: kmsg     */	ABSENT_STRING,
	/* 171: kmov     */	ABSENT_STRING,
	/* 172: knxt     */	ABSENT_STRING,
	/* 173: kopn     */	ABSENT_STRING,
	/* 174: kopt     */	ABSENT_STRING,
	/* 175: kprv     */	ABSENT_STRING,
	/* 176: kprt     */	ABSENT_STRING,
	/* 177: krdo     */	ABSENT_STRING,
	/* 178: kref     */	ABSENT_STRING,
	/* 179: krfr     */	ABSENT_STRING,
	/* 180: krpl     */	ABSENT_STRING,
	/* 181: krst     */	ABSENT_STRING,
	/* 182: kres     */	ABSENT_STRING,
	/* 183: ksav     */	ABSENT_STRING,
	/* 184: kspd     */	ABSENT_STRING,
	/* 185: kund     */	ABSENT_STRING,
	/* 186: kBEG     */	ABSENT_STRING,
	/* 187: kCAN     */	ABSENT_STRING,
	/* 188: kCMD     */	ABSENT_STRING,
	/* 189: kCPY     */	ABSENT_STRING,
	/* 190: kCRT     */	ABSENT_STRING,
	/* 191: kDC      */	screen_xterm_256color_s_kDC,
	/* 192: kDL      */	ABSENT_STRING,
	/* 193: kslt     */	ABSENT_STRING,
	/* 194: kEND     */	screen_xterm_256color_s_kEND,
	/* 195: kEOL     */	ABSENT_STRING,
	/* 196: kEXT     */	ABSENT_STRING,
	/* 197: kFND     */	ABSENT_STRING,
	/* 198: kHLP     */	ABSENT_STRING,
	/* 199: kHOM     */	screen_xterm_256color_s_kHOM,
	/* 200: kIC      */	ABSENT_STRING,
	/* 201: kLFT     */	screen_xterm_256color_s_kLFT,
	/* 202: kMSG     */	ABSENT_STRING,
	/* 203: kMOV     */	ABSENT_STRING,
	/* 204: kNXT     */	ABSENT_STRING,
	/* 205: kOPT     */	ABSENT_STRING,
	/* 206: kPRV     */	ABSENT_STRING,
	/* 207: kPRT     */	ABSENT_STRING,
	/* 208: kRDO     */	ABSENT_STRING,
	/* 209: kRPL     */	ABSENT_STRING,
	/* 210: kRIT     */	screen_xterm_256color_s_kRIT,
	/* 211: kRES     */	ABSENT_STRING,
	/* 212: kSAV     */	ABSENT_STRING,
	/* 213: kSPD     */	ABSENT_STRING,
	/* 214: kUND     */	ABSENT_STRING,
	/* 215: rfi      */	ABSENT_STRING,
	/* 216: kf11     */	screen_xterm_256color_s_kf11,
	/* 217: kf12     */	screen_xterm_256color_s_kf12,
	/* 218: kf13     */	screen_xterm_256color_s_kf13,
	/* 219: kf14     */	screen_xterm_256color_s_kf14,
	/* 220: kf15     */	screen_xterm_256color_s_kf15,
	/* 221: kf16     */	screen_xterm_256color_s_kf16,
	/* 222: kf17     */	screen_xterm_256color_s_kf17,
	/* 223: kf18     */	screen_xterm_256color_s_kf18,
	/* 224: kf19     */	screen_xterm_256color_s_kf19,
	/* 225: kf20     */	screen_xterm_256color_s_kf20,
	/* 226: kf21     */	screen_xterm_256color_s_kf21,
	/* 227: kf22     */	screen_xterm_256color_s_kf22,
	/* 228: kf23     */	screen_xterm_256color_s_kf23,
	/* 229: kf24     */	screen_xterm_256color_s_kf24,
	/* 230: kf25     */	screen_xterm_256color_s_kf25,
	/* 231: kf26     */	screen_xterm_256color_s_kf26,
	/* 232: kf27     */	screen_xterm_256color_s_kf27,
	/* 233: kf28     */	screen_xterm_256color_s_kf28,
	/* 234: kf29     */	screen_xterm_256color_s_kf29,
	/* 235: kf30     */	screen_xterm_256color_s_kf30,
	/* 236: kf31     */	screen_xterm_256color_s_kf31,
	/* 237: kf32     */	screen_xterm_256color_s_kf32,
	/* 238: kf33     */	screen_xterm_256color_s_kf33,
	/* 239: kf34     */	screen_xterm_256color_s_kf34,
	/* 240: kf35     */	screen_xterm_256color_s_kf35,
	/* 241: kf36     */	screen_xterm_256color_s_kf36,
	/* 242: kf37     */	screen_xterm_256color_s_kf37,
	/* 243: kf38     */	screen_xterm_256color_s_kf38,
	/* 244: kf39     */	screen_xterm_256color_s_kf39,
	/* 245: kf40     */	screen_xterm_256color_s_kf40,
	/* 246: kf41     */	screen_xterm_256color_s_kf41,
	/* 247: kf42     */	screen_xterm_256color_s_kf42,
	/* 248: kf43     */	screen_xterm_256color_s_kf43,
	/* 249: kf44     */	screen_xterm_256color_s_kf44,
	/* 250: kf45     */	screen_xterm_256color_s_kf45,
	/* 251: kf46     */	screen_xterm_256color_s_kf46,
	/* 252: kf47     */	screen_xterm_256color_s_kf47,
	/* 253: kf48     */	screen_xterm_256color_s_kf48,
	/* 254: kf49     */	screen_xterm_256color_s_kf49,
	/* 255: kf50     */	screen_xterm_256color_s_kf50,
	/* 256: kf51     */	screen_xterm_256color_s_kf51,
	/* 257: kf52     */	screen_xterm_256color_s_kf52,
	/* 258: kf53     */	screen_xterm_256color_s_kf53,
	/* 259: kf54     */	screen_xterm_256color_s_kf54,
	/* 260: kf55     */	screen_xterm_256color_s_kf55,
	/* 261: kf56     */	screen_xterm_256color_s_kf56,
	/* 262: kf57     */	screen_xterm_256color_s_kf57,
	/* 263: kf58     */	screen_xterm_256color_s_kf58,
	/* 264: kf59     */	screen_xterm_256color_s_kf59,
	/* 265: kf60     */	screen_xterm_256color_s_kf60,
	/* 266: kf61     */	screen_xterm_256color_s_kf61,
	/* 267: kf62     */	screen_xterm_256color_s_kf62,
	/* 268: kf63     */	screen_xterm_256color_s_kf63,
	/* 269: el1      */	screen_xterm_256color_s_el1,
	/* 270: mgc      */	ABSENT_STRING,
	/* 271: smgl     */	ABSENT_STRING,
	/* 272: smgr     */	ABSENT_STRING,
	/* 273: fln      */	ABSENT_STRING,
	/* 274: sclk     */	ABSENT_STRING,
	/* 275: dclk     */	ABSENT_STRING,
	/* 276: rmclk    */	ABSENT_STRING,
	/* 277: cwin     */	ABSENT_STRING,
	/* 278: wingo    */	ABSENT_STRING,
	/* 279: hup      */	ABSENT_STRING,
	/* 280: dial     */	ABSENT_STRING,
	/* 281: qdial    */	ABSENT_STRING,
	/* 282: tone     */	ABSENT_STRING,
	/* 283: pulse    */	ABSENT_STRING,
	/* 284: hook     */	ABSENT_STRING,
	/* 285: pause    */	ABSENT_STRING,
	/* 286: wait     */	ABSENT_STRING,
	/* 287: u0       */	ABSENT_STRING,
	/* 288: u1       */	ABSENT_STRING,
	/* 289: u2       */	ABSENT_STRING,
	/* 290: u3       */	ABSENT_STRING,
	/* 291: u4       */	ABSENT_STRING,
	/* 292: u5       */	ABSENT_STRING,
	/* 293: u6       */	screen_xterm_256color_s_u6,
	/* 294: u7       */	screen_xterm_256color_s_u7,
	/* 295: u8       */	screen_xterm_256color_s_u8,
	/* 296: u9       */	screen_xterm_256color_s_u9,
	/* 297: op       */	screen_xterm_256color_s_op,
	/* 298: oc       */	ABSENT_STRING,
	/* 299: initc    */	ABSENT_STRING,
	/* 300: initp    */	ABSENT_STRING,
	/* 301: scp      */	ABSENT_STRING,
	/* 302: setf     */	ABSENT_STRING,
	/* 303: setb     */	ABSENT_STRING,
	/* 304: cpi      */	ABSENT_STRING,
	/* 305: lpi      */	ABSENT_STRING,
	/* 306: chr      */	ABSENT_STRING,
	/* 307: cvr      */	ABSENT_STRING,
	/* 308: defc     */	ABSENT_STRING,
	/* 309: swidm    */	ABSENT_STRING,
	/* 310: sdrfq    */	ABSENT_STRING,
	/* 311: sitm     */	ABSENT_STRING,
	/* 312: slm      */	ABSENT_STRING,
	/* 313: smicm    */	ABSENT_STRING,
	/* 314: snlq     */	ABSENT_STRING,
	/* 315: snrmq    */	ABSENT_STRING,
	/* 316: sshm     */	ABSENT_STRING,
	/* 317: ssubm    */	ABSENT_STRING,
	/* 318: ssupm    */	ABSENT_STRING,
	/* 319: sum      */	ABSENT_STRING,
	/* 320: rwidm    */	ABSENT_STRING,
	/* 321: ritm     */	ABSENT_STRING,
	/* 322: rlm      */	ABSENT_STRING,
	/* 323: rmicm    */	ABSENT_STRING,
	/* 324: rshm     */	ABSENT_STRING,
	/* 325: rsubm    */	ABSENT_STRING,
	/* 326: rsupm    */	ABSENT_STRING,
	/* 327: rum      */	ABSENT_STRING,
	/* 328: mhpa     */	ABSENT_STRING,
	/* 329: mcud1    */	ABSENT_STRING,
	/* 330: mcub1    */	ABSENT_STRING,
	/* 331: mcuf1    */	ABSENT_STRING,
	/* 332: mvpa     */	ABSENT_STRING,
	/* 333: mcuu1    */	ABSENT_STRING,
	/* 334: porder   */	ABSENT_STRING,
	/* 335: mcud     */	ABSENT_STRING,
	/* 336: mcub     */	ABSENT_STRING,
	/* 337: mcuf     */	ABSENT_STRING,
	/* 338: mcuu     */	ABSENT_STRING,
	/* 339: scs      */	ABSENT_STRING,
	/* 340: smgb     */	ABSENT_STRING,
	/* 341: smgbp    */	ABSENT_STRING,
	/* 342: smglp    */	ABSENT_STRING,
	/* 343: smgrp    */	ABSENT_STRING,
	/* 344: smgt     */	ABSENT_STRING,
	/* 345: smgtp    */	ABSENT_STRING,
	/* 346: sbim     */	ABSENT_STRING,
	/* 347: scsd     */	ABSENT_STRING,
	/* 348: rbim     */	ABSENT_STRING,
	/* 349: rcsd     */	ABSENT_STRING,
	/* 350: subcs    */	ABSENT_STRING,
	/* 351: supcs    */	ABSENT_STRING,
	/* 352: docr     */	ABSENT_STRING,
	/* 353: zerom    */	ABSENT_STRING,
	/* 354: csnm     */	ABSENT_STRING,
	/* 355: kmous    */	screen_xterm_256color_s_kmous,
	/* 356: minfo    */	ABSENT_STRING,
	/* 357: reqmp    */	ABSENT_STRING,
	/* 358: getm     */	ABSENT_STRING,
	/* 359: setaf    */	screen_xterm_256color_s_setaf,
	/* 360: setab    */	screen_xterm_256color_s_setab,
	/* 361: pfxl     */	ABSENT_STRING,
	/* 362: devt     */	ABSENT_STRING,
	/* 363: csin     */	ABSENT_STRING,
	/* 364: s0ds     */	ABSENT_STRING,
	/* 365: s1ds     */	ABSENT_STRING,
	/* 366: s2ds     */	ABSENT_STRING,
	/* 367: s3ds     */	ABSENT_STRING,
	/* 368: smglr    */	ABSENT_STRING,
	/* 369: smgtb    */	ABSENT_STRING,
	/* 370: birep    */	ABSENT_STRING,
	/* 371: binel    */	ABSENT_STRING,
	/* 372: bicr     */	ABSENT_STRING,
	/* 373: colornm  */	ABSENT_STRING,
	/* 374: defbi    */	ABSENT_STRING,
	/* 375: endbi    */	ABSENT_STRING,
	/* 376: setcolor */	ABSENT_STRING,
	/* 377: slines   */	ABSENT_STRING,
	/* 378: dispc    */	ABSENT_STRING,
	/* 379: smpch    */	ABSENT_STRING,
	/* 380: rmpch    */	ABSENT_STRING,
	/* 381: smsc     */	ABSENT_STRING,
	/* 382: rmsc     */	ABSENT_STRING,
	/* 383: pctrm    */	ABSENT_STRING,
	/* 384: scesc    */	ABSENT_STRING,
	/* 385: scesa    */	ABSENT_STRING,
	/* 386: ehhlm    */	ABSENT_STRING,
	/* 387: elhlm    */	ABSENT_STRING,
	/* 388: elohlm   */	ABSENT_STRING,
	/* 389: erhlm    */	ABSENT_STRING,
	/* 390: ethlm    */	ABSENT_STRING,
	/* 391: evhlm    */	ABSENT_STRING,
	/* 392: sgr1     */	ABSENT_STRING,
	/* 393: slength  */	ABSENT_STRING,
	/* 394: OTi2     */	ABSENT_STRING,
	/* 395: OTrs     */	ABSENT_STRING,
	/* 396: OTnl     */	ABSENT_STRING,
	/* 397: OTbc     */	ABSENT_STRING,
	/* 398: OTko     */	ABSENT_STRING,
	/* 399: OTma     */	ABSENT_STRING,
	/* 400: OTG2     */	ABSENT_STRING,
	/* 401: OTG3     */	ABSENT_STRING,
	/* 402: OTG1     */	ABSENT_STRING,
	/* 403: OTG4     */	ABSENT_STRING,
	/* 404: OTGR     */	ABSENT_STRING,
	/* 405: OTGL     */	ABSENT_STRING,
	/* 406: OTGU     */	ABSENT_STRING,
	/* 407: OTGD     */	ABSENT_STRING,
	/* 408: OTGH     */	ABSENT_STRING,
	/* 409: OTGV     */	ABSENT_STRING,
	/* 410: OTGC     */	ABSENT_STRING,
	/* 411: meml     */	ABSENT_STRING,
	/* 412: memu     */	ABSENT_STRING,
	/* 413: box1     */	ABSENT_STRING,
};
static const TERMTYPE2 fallbacks[10] =
{
 /* linux */
	{
		linux_alias_data,
		(char *)0,	/* pointer to string table */
		linux_bool_data,
		linux_number_data,
		linux_string_data,
#if NCURSES_XNAMES
		(char *)0,	/* pointer to extended string table */
		(char **)0,	/* ...corresponding names */
		44,		/* count total Booleans */
		39,		/* count total Numbers */
		414,		/* count total Strings */
		0,		/* count extensions to Booleans */
		0,		/* count extensions to Numbers */
		0,		/* count extensions to Strings */
#endif /* NCURSES_XNAMES */
	}
, /* rxvt */
	{
		rxvt_alias_data,
		(char *)0,	/* pointer to string table */
		rxvt_bool_data,
		rxvt_number_data,
		rxvt_string_data,
#if NCURSES_XNAMES
		(char *)0,	/* pointer to extended string table */
		(char **)0,	/* ...corresponding names */
		44,		/* count total Booleans */
		39,		/* count total Numbers */
		414,		/* count total Strings */
		0,		/* count extensions to Booleans */
		0,		/* count extensions to Numbers */
		0,		/* count extensions to Strings */
#endif /* NCURSES_XNAMES */
	}
, /* vt100 */
	{
		vt100_alias_data,
		(char *)0,	/* pointer to string table */
		vt100_bool_data,
		vt100_number_data,
		vt100_string_data,
#if NCURSES_XNAMES
		(char *)0,	/* pointer to extended string table */
		(char **)0,	/* ...corresponding names */
		44,		/* count total Booleans */
		39,		/* count total Numbers */
		414,		/* count total Strings */
		0,		/* count extensions to Booleans */
		0,		/* count extensions to Numbers */
		0,		/* count extensions to Strings */
#endif /* NCURSES_XNAMES */
	}
, /* xterm */
	{
		xterm_alias_data,
		(char *)0,	/* pointer to string table */
		xterm_bool_data,
		xterm_number_data,
		xterm_string_data,
#if NCURSES_XNAMES
		(char *)0,	/* pointer to extended string table */
		(char **)0,	/* ...corresponding names */
		44,		/* count total Booleans */
		39,		/* count total Numbers */
		414,		/* count total Strings */
		0,		/* count extensions to Booleans */
		0,		/* count extensions to Numbers */
		0,		/* count extensions to Strings */
#endif /* NCURSES_XNAMES */
	}
, /* xterm-256color */
	{
		xterm_256color_alias_data,
		(char *)0,	/* pointer to string table */
		xterm_256color_bool_data,
		xterm_256color_number_data,
		xterm_256color_string_data,
#if NCURSES_XNAMES
		(char *)0,	/* pointer to extended string table */
		(char **)0,	/* ...corresponding names */
		44,		/* count total Booleans */
		39,		/* count total Numbers */
		414,		/* count total Strings */
		0,		/* count extensions to Booleans */
		0,		/* count extensions to Numbers */
		0,		/* count extensions to Strings */
#endif /* NCURSES_XNAMES */
	}
, /* screen */
	{
		screen_alias_data,
		(char *)0,	/* pointer to string table */
		screen_bool_data,
		screen_number_data,
		screen_string_data,
#if NCURSES_XNAMES
		(char *)0,	/* pointer to extended string table */
		(char **)0,	/* ...corresponding names */
		44,		/* count total Booleans */
		39,		/* count total Numbers */
		414,		/* count total Strings */
		0,		/* count extensions to Booleans */
		0,		/* count extensions to Numbers */
		0,		/* count extensions to Strings */
#endif /* NCURSES_XNAMES */
	}
, /* screen.linux */
	{
		screen_linux_alias_data,
		(char *)0,	/* pointer to string table */
		screen_linux_bool_data,
		screen_linux_number_data,
		screen_linux_string_data,
#if NCURSES_XNAMES
		(char *)0,	/* pointer to extended string table */
		(char **)0,	/* ...corresponding names */
		44,		/* count total Booleans */
		39,		/* count total Numbers */
		414,		/* count total Strings */
		0,		/* count extensions to Booleans */
		0,		/* count extensions to Numbers */
		0,		/* count extensions to Strings */
#endif /* NCURSES_XNAMES */
	}
, /* screen.rxvt */
	{
		screen_rxvt_alias_data,
		(char *)0,	/* pointer to string table */
		screen_rxvt_bool_data,
		screen_rxvt_number_data,
		screen_rxvt_string_data,
#if NCURSES_XNAMES
		(char *)0,	/* pointer to extended string table */
		(char **)0,	/* ...corresponding names */
		44,		/* count total Booleans */
		39,		/* count total Numbers */
		414,		/* count total Strings */
		0,		/* count extensions to Booleans */
		0,		/* count extensions to Numbers */
		0,		/* count extensions to Strings */
#endif /* NCURSES_XNAMES */
	}
, /* screen.xterm-new */
	{
		screen_xterm_xfree86_alias_data,
		(char *)0,	/* pointer to string table */
		screen_xterm_xfree86_bool_data,
		screen_xterm_xfree86_number_data,
		screen_xterm_xfree86_string_data,
#if NCURSES_XNAMES
		(char *)0,	/* pointer to extended string table */
		(char **)0,	/* ...corresponding names */
		44,		/* count total Booleans */
		39,		/* count total Numbers */
		414,		/* count total Strings */
		0,		/* count extensions to Booleans */
		0,		/* count extensions to Numbers */
		0,		/* count extensions to Strings */
#endif /* NCURSES_XNAMES */
	}
, /* screen.xterm-256color */
	{
		screen_xterm_256color_alias_data,
		(char *)0,	/* pointer to string table */
		screen_xterm_256color_bool_data,
		screen_xterm_256color_number_data,
		screen_xterm_256color_string_data,
#if NCURSES_XNAMES
		(char *)0,	/* pointer to extended string table */
		(char **)0,	/* ...corresponding names */
		44,		/* count total Booleans */
		39,		/* count total Numbers */
		414,		/* count total Strings */
		0,		/* count extensions to Booleans */
		0,		/* count extensions to Numbers */
		0,		/* count extensions to Strings */
#endif /* NCURSES_XNAMES */
	}
};

NCURSES_EXPORT(const TERMTYPE2 *)
_nc_fallback2 (const char *name GCC_UNUSED)
{
    const TERMTYPE2	*tp;

    for (tp = fallbacks;
	 tp < fallbacks + sizeof(fallbacks)/sizeof(TERMTYPE2);
	 tp++) {
	if (_nc_name_match(tp->term_names, name, "|")) {
	    return(tp);
	}
    }
    return((const TERMTYPE2 *)0);
}

#if NCURSES_EXT_NUMBERS
#undef _nc_fallback

/*
 * This entrypoint is used by tack.
 */
NCURSES_EXPORT(const TERMTYPE *)
_nc_fallback (const char *name)
{
    const TERMTYPE2 *tp = _nc_fallback2(name);
    const TERMTYPE *result = 0;
    if (tp != 0) {
	static TERMTYPE temp;
	_nc_export_termtype2(&temp, tp);
	result = &temp;
    }
    return result;
}
#endif
